// cmarray.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Array definition.
// -----------------------------------------------------------------

#ifndef _CMARRAY_H
#define _CMARRAY_H

#include <cm/include/cmcont.h>

class CmArrayIterator;                          // Iterator class stub.

class CmArray : public CmContainer {            // Array class definition.
public:
  CmArray(unsigned = 0, unsigned = 0);          // Default array constructor.
  CmArray(const CmArray&);                      // Array copy constructor.
  CmArray(const CmContainer&);                  // Construct from container.
 ~CmArray();                                    // Array destructor.

  CmArray& operator=(const CmArray&);           // Assignment operator.

  void     delta  (unsigned);                   // Set delta value.
  unsigned delta  () const;                     // Get delta value.
  unsigned total  () const;                     // Return number of objects.
  bool     resize (unsigned);                   // Resize the array.
  bool     isEmpty() const;                     // Is array empty?

  bool add      (CmObject*);                    // Append object to array.
  bool insertAt (int, CmObject*);               // Insert object at index.
  bool replaceAt(int, CmObject*);               // Replace object at index.
  bool remove   (const CmObject*);              // Remove equal object.
  bool remove   (const CmIterator&);            // Remove iterator object.
  bool removeAt (int);                          // Remove object at index.
  void removeAll();                             // Remove all objects.

  int index(const CmObject*  ) const;           // Get index of equal object.
  int index(const CmIterator&) const;           // Get index of iter object.

  void quickSort();                             // Quick sort the array.

  CmObject* at         (int)             const; // Get object at index.
  CmObject* lookup     (const CmObject*) const; // Look for equal object.
  bool      contains   (const CmObject*) const; // Object is in array?
  unsigned  occurrences(const CmObject*) const; // How many occurrences?

  CmIterator* newIterator() const;              // Get array iterator.
  
  bool write(CmIOFile&) const;                  // Write to file.
  bool read (CmIOFile&);                        // Read from file.

  CMOBJECT_DEFINE(CmArray)                      // Define object funcs.

private:
  static int cmpObjs(const void*, const void*); // Quick sort compare func.

  unsigned   _delta;                            // Delta value.
  unsigned   _total;                            // Number of objects.
  CmObject **_entries;                          // Array of object pointers.

  friend CmArrayIterator;                       // Iterator can access,
};

class CmArrayIterator : public CmIterator {     // Iterator definition.
public:
  CmArrayIterator(const CmArray& arr)           // Iterator constructor.
                 : CmIterator(arr), _index(0) {}

  bool      done    () const;                   // Check if done iterating.
  CmObject* next    ();                         // Return and advance.
  CmObject* previous();                         // Return and backup.
  CmObject* current () const;                   // Return current object.
  void      first   ();                         // Move to first object.
  void      last    ();                         // Move to last object.

  const CmArray& array() const;                 // Return array reference.

  CMOBJECT_DEFINE(CmArrayIterator)              // Define object funcs.

private:
  int _index;                                   // Current array index.
};

// "delta" sets a new delta value for automatic growing.
inline void CmArray::delta(unsigned dt)
{ _delta = dt; }

// "delta" returns the current delta value.
inline unsigned CmArray::delta() const
{ return _delta; }

// "total" returns the number of objects added to this array.
inline unsigned CmArray::total() const
{ return _total; }

// "at" returns the object at the specified index.
inline CmObject* CmArray::at(int idx) const
{ return (idx >= 0 && idx < _total) ? _entries[idx] : NULL; }

// "contains" checks if the array contains an object equal to the input.
inline bool CmArray::contains(const CmObject* pObj) const
{ return (index(pObj) > -1) ? true : false; }

// "array" returns a reference to the array being iterated.
inline const CmArray& CmArrayIterator::array() const
{ return (const CmArray&) _container; }

// "done" checks to see if we can iterate any further.
inline bool CmArrayIterator::done() const
{ return (_index >= 0 && _index < array()._total) ? false : true; }

// "next" returns the current object and advances the iterator.
inline CmObject* CmArrayIterator::next()
{ return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index++] : NULL; }

// "previous" returns the current object and decrements the iterator.
inline CmObject* CmArrayIterator::previous()
{ return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index--] : NULL; }

// "current" returns the current object pointed to by the iterator.
inline CmObject* CmArrayIterator::current() const
{ return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index] : NULL; }

// "first" moves the iterator to the first object.
inline void CmArrayIterator::first()
{ _index = 0; }

// "last" moves the iterator to the last object.
inline void CmArrayIterator::last()
{ _index = array()._total - 1; }

#endif
