// cmbintr.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Binary tree definition.
// -----------------------------------------------------------------

#ifndef _CMBINTR_H
#define _CMBINTR_H

#include <cm/include/cmcont.h>
#include <cm/include/cmstack.h>
#include <cm/include/cmqueue.h>

class CmBinaryTreeNode;                          // Tree node class stub.
class CmBinaryTreeIterator;                      // Iterator class stub.

class CmBinaryTree : public CmContainer {        // Tree class definition.
public:
  CmBinaryTree() : _root(NULL) {}                // Default tree constructor.
  CmBinaryTree(const CmBinaryTree&);             // Tree copy constructor.
  CmBinaryTree(const CmContainer&);              // Construct from container.
 ~CmBinaryTree();                                // Tree destructor.

  CmBinaryTree& operator=(const CmBinaryTree&);  // Assignment operator.

  bool add      (CmObject*);                     // Add object to tree.
  bool remove   (const CmObject*);               // Remove equal object.
  bool remove   (const CmIterator&);             // Remove iterator object.
  bool removeAt (int);                           // Remove at index.
  void removeAll();                              // Clear tree contents.

  CmObject* root   () const;                     // Return root object.
  void      balance();                           // Balance the tree.

  CmObject* lookup     (const CmObject*) const;  // Locate equal object.
  bool      contains   (const CmObject*) const;  // Tree has object?
  unsigned  occurrences(const CmObject*) const;  // Count occurrences of.

  CmIterator* newIterator() const;               // Get tree iterator.

  bool write(CmIOFile&) const;                   // Write to file.
  bool read (CmIOFile&);                         // Read from file.

  CMOBJECT_DEFINE(CmBinaryTree)                  // Define object funcs.

private:
  bool removeNode (CmBinaryTreeNode*, CmBinaryTreeNode*); // Remove node.
  void balanceHalf(const CmQueue&, int, int);             // Balance tree.

  static void killNode(CmBinaryTreeNode*, bool); // Destroy sub-nodes.

  CmBinaryTreeNode *_root;                       // Tree root node.

  friend CmBinaryTreeIterator;                   // Iterator can access.
};

class CmBinaryTreeIterator : public CmIterator { // Iterator definition.
public:
  CmBinaryTreeIterator(const CmBinaryTree&);     // Constructor.

  bool      done    () const;                    // Check if at end.
  CmObject* next    ();                          // Return and advance.
  CmObject* previous();                          // Return and backup.
  CmObject* current () const;                    // Get current object.
  void      first   ();                          // Move to first object.
  void      last    ();                          // Move to last object.

  const CmBinaryTree& tree() const;              // Return tree reference.

  CMOBJECT_DEFINE(CmBinaryTreeIterator)          // Define object funcs.

private:
  CmBinaryTreeNode *_node;                       // Current tree node.
  CmStack           _stack;                      // Node stack.
};

class CmBinaryTreeNode : public CmObject {       // Tree node class.
private:
  CmBinaryTreeNode(CmObject *O)                  // Node constructor.
                  : _left(NULL), _data(O), _right(NULL) {}

  CMOBJECT_DEFINE(CmBinaryTreeNode)              // Define base.

  CmBinaryTreeNode *_left;                       // Left subtree pointer.
  CmObject         *_data;                       // Object pointer.
  CmBinaryTreeNode *_right;                      // Right subtree pointer.

  friend CmBinaryTree;                           // Tree can access.
  friend CmBinaryTreeIterator;                   // Iterator can access.
};

// "root" returns the root object of the tree.
inline CmObject* CmBinaryTree::root() const
{ return (_root) ? _root->_data : NULL; }

// "tree" returns a reference to the tree being iterated.
inline const CmBinaryTree& CmBinaryTreeIterator::tree() const
{ return (const CmBinaryTree&) _container; }

// "done" returns true if the iterator can proceed no further.
inline bool CmBinaryTreeIterator::done() const
{ return (!_node) ? true : false; }

// "current" returns the current tree object.
inline CmObject* CmBinaryTreeIterator::current() const
{ return (_node) ? _node->_data : NULL; }

#endif
