// cmbtree.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// BTree tree definition.
// -----------------------------------------------------------------

#ifndef _CMBTREE_H
#define _CMBTREE_H

#include <cm/include/cmcont.h>

class CmBTreeNode;                              // Tree node class stub.
class CmBTreeIterator;                          // Tree iterator class stub.

class CmBTree : public CmContainer {            // BTree class definition.
public:
  CmBTree(unsigned = 3);                        // Default tree constructor.
  CmBTree(const CmBTree&);                      // Tree copy constructor.
  CmBTree(const CmContainer&);                  // Construct from container.
 ~CmBTree();                                    // Tree destructor.

  CmBTree& operator=(const CmBTree&);           // Assignment operator.

  unsigned order () const;                      // Return tree order.
  unsigned height() const;                      // Return tree height.

  bool add      (CmObject*);                    // Add object to tree.
  bool remove   (const CmObject*);              // Remove equal object.
  bool remove   (const CmIterator&);            // Remove iterator object.
  bool removeAt (int);                          // Remove at index.
  void removeAll();                             // Remove all objects.

  CmObject* lookup     (const CmObject*) const; // Find equal object in tree.
  bool      contains   (const CmObject*) const; // See if tree contains equal.
  unsigned  occurrences(const CmObject*) const; // Count occurrences of equal.

  CmIterator* newIterator() const;              // Create tree iterator.

  CMOBJECT_DEFINE(CmBTree)                      // Define object funcs.

private:
  CmBTreeNode* nodeSearch(const CmObject*) const; // Search for insertion node.
  CmBTreeNode* lookupNode(const CmObject*) const; // Get node where object is.
  CmBTreeNode* lookupAbs (const CmObject*) const; // Get node where object is.
  bool         remove    (CmBTreeNode*, int);     // Remove object in node.

  static void removeNodes(CmBTreeNode*, bool);  // Remove node children.
  static unsigned minKeys(unsigned);            // Get minimum keys allowed.

  unsigned     _order;                          // Tree order.
  CmBTreeNode *_root;                           // Root tree node.

  friend CmBTreeIterator;                       // Iterator class can access.
};

class CmBTreeNode {                             // Tree node definition.
private:
  CmBTreeNode(CmBTreeNode*, unsigned);          // Node constructor.
 ~CmBTreeNode();                                // Node destructor.

  int  shouldGo(const CmObject*) const;         // Where should object insert.
  int  index   (const CmObject*) const;         // Index of equal object.
  int  indexAbs(const CmObject*) const;         // Index of absolute object.
  int  index   (CmBTreeNode*) const;            // Index where child is located.
  bool addAt   (int, CmObject*, CmBTreeNode*, int); // Add object at index.
  bool removeAt(int, int);                      // Remove object at index.
  void clear   (int);                           // Clear arrays.

  CmBTreeNode* split(unsigned);                 // Split the node.

  unsigned      _total;                         // Number of objects in node.
  CmBTreeNode  *_parent;                        // Parent node.
  CmObject    **_data;                          // (order  ) object pointers.
  CmBTreeNode **_children;                      // (order+1) child nodes.

  friend CmBTree;                               // Tree class can access.
  friend CmBTreeIterator;                       // Iterator class can access.
};

class CmBTreeIterator : public CmIterator {     // Iterator class definition.
public:
  CmBTreeIterator(const CmBTree&);              // Iterator constructor.

  bool      done    () const;                   // Check if at end.
  CmObject* next    ();                         // Return and advance.
  CmObject* previous();                         // Return and backup.
  CmObject* current () const;                   // Get current object.
  void      first   ();                         // Move to first object.
  void      last    ();                         // Move to last object.

  const CmBTree& tree() const;                  // Return tree reference.

  CMOBJECT_DEFINE(CmBTreeIterator)              // Define object funcs.

private:
  CmBTreeNode *_node;                           // Current tree node.
  int          _index;                          // Current node object.
};

// "order" returns the current tree order.
inline unsigned CmBTree::order () const
{ return _order; }

// "minKeys" returns the minimum number of keys allowed in a non-
// root node of the tree.
inline unsigned CmBTree::minKeys(unsigned ordr)
{ return (((ordr / 2) * 2) == ordr) ? (ordr / 2) - 1 : (ordr / 2); }

// "tree" returns a reference to the tree being iterated.
inline const CmBTree& CmBTreeIterator::tree() const
{ return (const CmBTree&) _container; }

// "done" returns whether or not we are done iterating.
inline bool CmBTreeIterator::done() const
{ return (!_node) ? true : false; }

// "current" returns the current object pointed to by this iterator.
inline CmObject* CmBTreeIterator::current() const
{ return (_node && _node->_data[_index]) ? _node->_data[_index] : NULL; }

#endif
