// cmcont.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Abstract container definition.
// -----------------------------------------------------------------

#ifndef _CMCONT_H
#define _CMCONT_H

#include <cm/include/cmiter.h>

typedef bool (*CmQueryFunc)(CmObject*, void*);     // Query function type.
typedef bool (CmObject::*CmQueryMemFunc)(void*);   // Query member function.

class CmContainer : public CmObject {              // Container definition.
public:
  CmContainer() : _size(0), _ownsObjects(true) {}  // Constructor.

  virtual ~CmContainer() {}                        // Virtual destructor.

  CmObject* operator[](int) const;                 // Get object at index.

  virtual unsigned size   () const;                // Return size.
  virtual unsigned total  () const;                // Return total objects.
  virtual bool     isEmpty() const;                // Is container empty?

  virtual bool add      (CmObject*)         = 0;   // Add object.
  virtual bool remove   (const CmObject*)   = 0;   // Remove equal object.
  virtual bool remove   (const CmIterator&) = 0;   // Remove iterator object.
  virtual bool removeAt (int)               = 0;   // Remove at index.
  virtual void removeAll()                  = 0;   // Remove all objects.

  virtual CmObject* at         (int)             const; // Get at index.
  virtual CmObject* lookup     (const CmObject*) const; // Find equal obj.
  virtual bool      contains   (const CmObject*) const; // Contains equal?
  virtual unsigned  occurrences(const CmObject*) const; // How many equals?

  virtual CmIterator* newIterator() const = 0;     // Get new iterator.

  CmContainer* lookupAll (const CmObject*) const;  // Find all equal objs.
  bool         addAllFrom(const CmContainer*);     // Combine containers.

  void ownsObjects(bool);                          // Set ownership policy.
  bool ownsObjects() const;                        // Get ownership policy.

  void printOn(ostream&) const;                    // Print to stream.
  bool write  (CmIOFile&) const;                   // Write to reserve.
  bool read   (CmIOFile&);                         // Read from reserve.

  // Call the specified function for every object in the container
  // until the end is reached or the function returns false.
  void forEach(CmQueryFunc,    void* = NULL);
  void forEach(CmQueryMemFunc, void* = NULL);

  // Return the first object to satisfy the conditional test performed
  // in the specified function.
  CmObject* firstThat(CmQueryFunc,    void* = NULL) const;
  CmObject* firstThat(CmQueryMemFunc, void* = NULL) const;

  // Return the last object to satisfy the conditional test performed
  // in the specified function.
  CmObject* lastThat(CmQueryFunc,    void* = NULL) const;
  CmObject* lastThat(CmQueryMemFunc, void* = NULL) const;

  // Call the specified function for every object in the container
  // adding all objects for which the function returns true to an
  // output container.
  CmContainer* query(CmQueryFunc,    void* = NULL) const;
  CmContainer* query(CmQueryMemFunc, void* = NULL) const;

  CMOBJECT_ABSTRACT(CmContainer)                   // Define object funcs.

protected:
  virtual void copy(const CmContainer&);           // Copy method.

  bool     writeBase(CmIOFile&) const;             // Write base container.
  unsigned readBase (CmIOFile&);                   // Read base container.

  unsigned _size;                                  // Current size.
  bool     _ownsObjects;                           // Ownership flag.
};

// "[]" operator returns the object at the specified index.
inline CmObject* CmContainer::operator[](int idx) const
{ return at(idx); }

// "size" returns the container size.
inline unsigned CmContainer::size() const
{ return _size; }

// "total" returns the total number of objects.
inline unsigned CmContainer::total() const
{ return _size; }

// "isEmpty" checks if the container is empty or not.
inline bool CmContainer::isEmpty() const
{ return (_size == 0) ? true : false; }

// "ownsObjects" sets a new ownership policy for this container.
inline void CmContainer::ownsObjects(bool owns)
{ _ownsObjects = owns; }

// "ownsObjects" returns the current ownership policy for this container.
inline bool CmContainer::ownsObjects() const
{ return _ownsObjects; }

// "CmForEach" and "CmEndFor" define a foreach loop where the object
// pointer and collection are passed in and the object pointer will
// point to the next object in the collection for each loop iteration.
//
#define CmForEach(obj, cltn) {                           \
        CmIterator *___cm_iterator = cltn.newIterator(); \
        while (!___cm_iterator->done()) {                \
          obj = ___cm_iterator->next();
#define CmEndFor } delete ___cm_iterator; }

// "CmRepeat" and "CmUntil" define a pascal style repeat/until loop
// where the object pointer and collection are passed in and the object
// pointer will point to the next object in the collection for each loop
// iteration.
//
#define CmRepeat(obj, cltn) {                            \
        CmIterator *___cm_iterator = cltn.newIterator(); \
        do {                                             \
          obj = ___cm_iterator->next();
#define CmUntil(condition) } while (!(condition)); delete ___cm_iterator; }

// "CmDo" and "CmWhile" define a typical do/while loop where the object
// pointer and collection are passed in and the object pointer will point
// to the next object in the collection for each loop iteration.
//
#define CmDo(obj, cltn) {                                \
        CmIterator *___cm_iterator = cltn.newIterator(); \
        do {                                             \
          obj = ___cm_iterator->next();
#define CmWhile(condition) } while (condition); delete ___cm_iterator; }

#endif
