// cmdate.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Date class definition.
// -----------------------------------------------------------------

#ifndef _CMDATE_H
#define _CMDATE_H

#include <cm/include/cmobject.h>

class CmDate : public CmObject {                  // Date class definition.
public:
  CmDate();                                       // Default date constructor.
  CmDate(unsigned, unsigned);                     // Construct from d, y.
  CmDate(unsigned, const char*, unsigned);        // Construct from d, m, y.
  CmDate(unsigned, unsigned, unsigned);           // Construct from d, m, y.
  CmDate(unsigned long j) : _julnum(j) {}         // Construct from jul num.

  unsigned    day            () const;            // Get day of year number.
  unsigned    dayOfMonth     () const;            // Get day of month number.
  unsigned    firstDayOfMonth() const;            // Get day num first of month.
  unsigned    firstDayOfMonth(unsigned) const;    // Get day num first of month.
  unsigned    month          () const;            // Get date's month number.
  const char* nameOfDay      () const;            // Get date's day name.
  const char* nameOfMonth    () const;            // Get date's month name.
  unsigned    weekDay        () const;            // Get date's weekday number.
  unsigned    year           () const;            // Get date's year number.
  bool        leap           () const;            // Check if leap year.
  CmDate      previous       (const char*) const; // Get previous named day.
  CmDate      previous       (unsigned) const;    // Get previous numbered day.
  void        makeCurrent    ();                  // Make this date current.

  bool isEqual(const CmObject*) const;            // Compare dates.
  int  compare(const CmObject*) const;            // Compare dates.

  unsigned hash(unsigned) const;                  // Hash date.

  void printOn(ostream&) const;                   // Print date to stream.

  bool write(CmIOFile&) const;                    // Write date to file.
  bool read (CmIOFile&);                          // Read date from file.

  CMOBJECT_DEFINE(CmDate)                         // Define object funcs.

  unsigned long julNum() const;                   // Return julian number.

  bool operator< (const CmDate&) const;           // Check if date <  this.
  bool operator<=(const CmDate&) const;           // Check if date <= this.
  bool operator> (const CmDate&) const;           // Check if date >  this.
  bool operator>=(const CmDate&) const;           // Check if date >= this.
  bool operator==(const CmDate&) const;           // Check if date == this.
  bool operator!=(const CmDate&) const;           // Check if date != this.

  CmDate operator+(const CmDate&) const;          // Add dates.
  CmDate operator+(int)           const;          // Add julnum to date.
  CmDate operator-(const CmDate&) const;          // Subtract dates.
  CmDate operator-(int)           const;          // Sub julnum from date.

  static unsigned long julDay   (unsigned, unsigned, unsigned); // Get julnum.
  static unsigned dayOfWeek     (const char*);    // Return day of week num.
  static bool     dayWithinMonth(unsigned, unsigned, unsigned); // Get day num.
  static bool     leapYear      (unsigned);       // Check if leap year.
  static unsigned indexOfMonth  (const char*);    // Get month number.

  static const char* dayName  (unsigned);         // Get day name from number.
  static const char* monthName(unsigned);         // Get month name from num.

private:
  void mdy(unsigned&, unsigned&, unsigned&) const; // Get date from julnum.

  unsigned long _julnum;                          // Julian day number.
  static   int  _displayStyle;                    // Date display style.
};

class CmDateFormat {                               // Date format definition.
private:
  friend ostream& dateNormal (ostream&);           // Normal format manipulator.
  friend ostream& dateNumbers(ostream&);           // Number format manipulator.
  static int      _index;                          // xalloc index from ios.
  friend CmDate;                                   // Date class can access.
};

// "julNum" returns the julian day number.
inline unsigned long CmDate::julNum() const
{ return _julnum; }

// "<" checks if the specified date is less than this.
inline bool CmDate::operator<(const CmDate& D) const
{ return (_julnum < D._julnum); }

// "<=" checks if the specified date is less than or equal to this.
inline bool CmDate::operator<=(const CmDate& D) const
{ return (_julnum <= D._julnum); }

// ">" checks if the specified date is greater than this.
inline bool CmDate::operator>(const CmDate& D) const
{ return (_julnum > D._julnum); }

// ">=" checks if the specified date is greater than or equal to this.
inline bool CmDate::operator>=(const CmDate& D) const
{ return (_julnum >= D._julnum); }

// "==" checks if the specified date is equal to this.
inline bool CmDate::operator==(const CmDate& D) const
{ return (_julnum == D._julnum); }

// "!=" checks if the specified date is not equal to this.
inline bool CmDate::operator!=(const CmDate& D) const
{ return (_julnum != D._julnum); }

// "+" adds two dates returning the resulting date.
inline CmDate CmDate::operator+(const CmDate& D) const
{ return CmDate(_julnum + D._julnum); }

// "+" adds a julian day number to this date returning the new date.
inline CmDate CmDate::operator+(int d) const
{ return CmDate(_julnum + d); }

// "-" subtracts the input date from this returning the resulting date.
inline CmDate CmDate::operator-(const CmDate& D) const
{ return CmDate(_julnum - D._julnum); }

// "-" subtracts a julian day number from this returning the new date.
inline CmDate CmDate::operator-(int d) const
{ return CmDate(_julnum - d); }

#endif
