// cmdlist.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Doubly linked list definition.
// -----------------------------------------------------------------

#ifndef _CMDLIST_H
#define _CMDLIST_H

#include <cm/include/cmcont.h>

class CmDLink;                                     // List link class stub.
class CmDLinkedListIterator;                       // Iterator class stub.

class CmDLinkedList : public CmContainer {         // Linked list definition.
public:
  CmDLinkedList() : _first(NULL), _last(NULL) {}   // Construct list.
  CmDLinkedList(const CmDLinkedList&);             // List copy constructor.
  CmDLinkedList(const CmContainer&);               // Construct from container.
 ~CmDLinkedList();                                 // List destructor.

  CmDLinkedList& operator=(const CmDLinkedList&);  // Assignment operator.

  bool add      (CmObject*);                       // Add to end of list.
  bool append   (CmObject*);                       // Add to end of list.
  bool prepend  (CmObject*);                       // Add to start of list.
  bool insertAt (int, CmObject*);                  // Insert object at index.
  bool replaceAt(int, CmObject*);                  // Replace object at index.
  bool remove   (const CmObject*);                 // Remove object from list.
  bool remove   (const CmIterator&);               // Remove iterator object.
  bool removeAt (int);                             // Remove at index.
  void removeAll();                                // Clear the list.

  CmObject* removeFirst();                         // Remove first object.
  CmObject* removeLast ();                         // Remove last object.

  bool insertAfter (const CmObject*,   CmObject*); // Insert after another.
  bool insertAfter (const CmIterator&, CmObject*); // Insert after iter obj.
  bool insertBefore(const CmObject*,   CmObject*); // Insert before another.
  bool insertBefore(const CmIterator&, CmObject*); // Insert before iter obj.

  int index(const CmObject*  ) const;              // Get index of equal obj.
  int index(const CmIterator&) const;              // Get index of iter obj.

  CmObject* first() const;                         // Get first object in list.
  CmObject* last () const;                         // Get last object in list.

  CmObject* at         (int)             const;    // Get object at index.
  CmObject* lookup     (const CmObject*) const;    // Find equal object.
  bool      contains   (const CmObject*) const;    // See if list contains.
  unsigned  occurrences(const CmObject*) const;    // Count occurrences of obj.

  CmIterator* newIterator() const;                 // Create list iterator.

  CMOBJECT_DEFINE(CmDLinkedList)                   // Define object funcs.

private:
  bool removeAbs(CmObject*);                       // Remove specific object.

  CmDLink *_first;                                 // First link in list.
  CmDLink *_last;                                  // Last link in list.

  friend CmDLinkedListIterator;                    // Iterator can access.
};

class CmDLinkedListIterator : public CmIterator {  // Iterator definition.
public:
  CmDLinkedListIterator(const CmDLinkedList& L)    // Iterator construct.
                      : CmIterator(L), _link(L._first) {}

  bool      done    () const;                      // Check if end.
  CmObject* next    ();                            // Return and advance.
  CmObject* previous();                            // Return and backup.
  CmObject* current () const;                      // Get current object.
  void      first   ();                            // Move to first object.
  void      last    ();                            // Move to last object.

  const CmDLinkedList& list() const;               // Return list reference.

  CMOBJECT_DEFINE(CmDLinkedListIterator)           // Define object funcs.

private:
  CmDLink *_link;                                  // Current list link.
};

class CmDLink {                                    // List link definition.
private:
  CmDLink(CmObject *O)                             // Link constructor.
         : _next(NULL), _data(O), _prev(NULL) {}

  CmDLink  *_next;                                 // Pointer to next link.
  CmObject *_data;                                 // Pointer to object.
  CmDLink  *_prev;                                 // Pointer to prev link.

  friend CmDLinkedList;                            // List class can access.
  friend CmDLinkedListIterator;                    // Iterator class can access.
};

// "first" returns the first object in the list.
inline CmObject* CmDLinkedList::first() const
{ return (_first) ? _first->_data : NULL; }

// "last" returns the last object in the list.
inline CmObject* CmDLinkedList::last() const
{ return (_last) ? _last->_data : NULL; }

// "list" returns a reference to the list being iterated.
inline const CmDLinkedList& CmDLinkedListIterator::list() const
{ return (const CmDLinkedList&) _container; }

// "done" checks if the iterator can advance any further.
inline bool CmDLinkedListIterator::done() const
{ return _link == NULL; }

// "current" returns the current object in the list.
inline CmObject* CmDLinkedListIterator::current() const
{ return (_link) ? _link->_data : NULL; }

// "first" resets the iterator to the start of the list.
inline void CmDLinkedListIterator::first()
{ _link = list()._first; }

// "last" moves the iterator to the last object.
inline void CmDLinkedListIterator::last()
{ _link = list()._last; }

#endif
