// cmhash.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Hash table definition.
// -----------------------------------------------------------------

#ifndef _CMHASH_H
#define _CMHASH_H

#include <cm/include/cmcont.h>

class CmHashBucket;                             // Hash bucket class stub.
class CmHashNode;                               // Hash node class stub.
class CmHashTableIterator;                      // Iterator class stub.

class CmHashTable : public CmContainer {        // Hash table definition.
public:
  CmHashTable(unsigned = 10);                   // Default table constructor.
  CmHashTable(const CmHashTable&);              // Table copy constructor.
  CmHashTable(const CmContainer&);              // Construct from container.
 ~CmHashTable();                                // Hash table destructor.

  CmHashTable& operator=(const CmHashTable&);   // Assignment operator.

  unsigned total  () const;                     // Number of object in table.
  bool     resize (unsigned);                   // Resize the hash table.
  bool     isEmpty() const;                     // Is table empty?

  bool add      (CmObject*);                    // Add object to container.
  bool remove   (const CmObject*);              // Remove equal object.
  bool remove   (const CmIterator&);            // Remove iterator object.
  bool removeAt (int);                          // Remove at index.
  void removeAll();                             // Remove all objects.

  CmObject* lookup     (const CmObject*) const; // Find equal object.
  bool      contains   (const CmObject*) const; // Table has object?
  unsigned  occurrences(const CmObject*) const; // Count occurrences of object.

  CmIterator* newIterator() const;              // Get hash table iterator.

  bool write(CmIOFile&) const;                  // Write to file.
  bool read (CmIOFile&);                        // Read from file.

  CMOBJECT_DEFINE(CmHashTable)                  // Define object funcs.

private:
  unsigned      _total;                         // Number of objects in table.
  CmHashBucket *_buckets;                       // Array of hash buckets.

  friend CmHashTableIterator;                   // Iterator can access.
};

class CmHashBucket {                            // Bucket class definition.
private:
  CmHashBucket() : _size(0), _first(NULL), _last(NULL) {}  // Bucket construct.
 ~CmHashBucket() { removeAll(false); }          // Bucket destructor.

  unsigned  size     () const;                  // Return list size.
  bool      append   (CmObject*);               // Add object to end of list.
  bool      remove   (const CmObject*, bool);   // Remove equal object.
  bool      removeAbs(const CmObject*, bool);   // Remove exact object.
  void      removeAll(bool);                    // Remove all objects.
  CmObject* lookup   (const CmObject*) const;   // Find equal object.
  bool      contains (const CmObject*) const;   // See if contains equal obj.

  unsigned    _size;                            // List size.
  CmHashNode *_first;                           // First object in list.
  CmHashNode *_last;                            // Last object in list.

  friend CmHashTable;                           // Table can access.
  friend CmHashTableIterator;                   // Iterator can access.
};

class CmHashNode {                              // Hash bucket node definition.
private:
  CmHashNode(CmObject* O)                       // Node constructor.
            : _next(NULL), _data(O) {}

  CmHashNode *_next;                            // Next node in list.
  CmObject   *_data;                            // Object pointer.

  friend CmHashBucket;                          // Bucket can access.
  friend CmHashTableIterator;                   // Iterator can access.
};

class CmHashTableIterator : public CmIterator { // Iterator definition.
public:
  CmHashTableIterator(const CmHashTable&);      // Iterator constructor.

  bool      done    () const;                   // Check if done iterating.
  CmObject* next    ();                         // Return and advance.
  CmObject* previous();                         // Return and backup.
  CmObject* current () const;                   // Return current object.
  void      first   ();                         // Move to first object.
  void      last    ();                         // Move to last object.

  const CmHashTable& table() const;             // Return table reference.

  CMOBJECT_DEFINE(CmHashTableIterator)          // Define object funcs.

private:
  int         _bucket;                          // Current bucket index.
  CmHashNode *_node;                            // Current bucket node.
};

// "total" returns the number of object contained in the table.
inline unsigned CmHashTable::total() const
{ return _total; }

// "isEmpty" returns true if the hash table contains no objects.
inline bool CmHashTable::isEmpty() const
{ return (_total == 0) ? true : false; }

// "size" returns the hash bucket size.
inline unsigned CmHashBucket::size() const
{ return _size; }

// "table" returns a reference to the table being iterated.
inline const CmHashTable& CmHashTableIterator::table() const
{ return (const CmHashTable&) _container; }

// "done" checks to see if the iterator can move any farther.
inline bool CmHashTableIterator::done() const
{ return (_node) ? false : true; }

// "current" returns the current object pointed to by the iterator.
inline CmObject* CmHashTableIterator::current() const
{ return (_node) ? _node->_data : NULL; }

#endif
