// cmlist.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Singly linked list definition.
// -----------------------------------------------------------------

#ifndef _CMLIST_H
#define _CMLIST_H

#include <cm/include/cmcont.h>

class CmLink;                                      // List link class stub.
class CmLinkedListIterator;                        // Iterator class stub.

class CmLinkedList : public CmContainer {          // Linked list definition.
public:
  CmLinkedList() : _first(NULL), _last(NULL) {}    // Construct list.
  CmLinkedList(const CmLinkedList&);               // List copy constructor.
  CmLinkedList(const CmContainer&);                // Construct from container.
 ~CmLinkedList();                                  // List destructor.

  CmLinkedList& operator=(const CmLinkedList&);    // Assignment operator.

  bool add      (CmObject*);                       // Add to end of list.
  bool append   (CmObject*);                       // Add to end of list.
  bool prepend  (CmObject*);                       // Add to start of list.
  bool insertAt (int, CmObject*);                  // Insert object at index.
  bool replaceAt(int, CmObject*);                  // Replace object at index.
  bool remove   (const CmObject*);                 // Remove object from list.
  bool remove   (const CmIterator&);               // Remove iterator object.
  bool removeAt (int);                             // Remove at index.
  void removeAll();                                // Clear the list.

  CmObject* removeFirst();                         // Remove first object.
  CmObject* removeLast ();                         // Remove last object.

  bool insertAfter (const CmObject*,   CmObject*); // Insert after another.
  bool insertAfter (const CmIterator&, CmObject*); // Insert after iter obj.
  bool insertBefore(const CmObject*,   CmObject*); // Insert before another.
  bool insertBefore(const CmIterator&, CmObject*); // Insert before iter obj.

  int index(const CmObject*  ) const;              // Get index of equal obj.
  int index(const CmIterator&) const;              // Get index of iter obj.

  CmObject* first() const;                         // Get first object in list.
  CmObject* last () const;                         // Get last object in list.

  CmObject* at         (int)             const;    // Get object at index.
  CmObject* lookup     (const CmObject*) const;    // Find equal object.
  bool      contains   (const CmObject*) const;    // See if list contains.
  unsigned  occurrences(const CmObject*) const;    // Count occurrences of obj.

  CmIterator* newIterator() const;                 // Create list iterator.

  CMOBJECT_DEFINE(CmLinkedList)                    // Define object funcs.

private:
  bool removeAbs(CmObject*);                       // Remove specific object.

  CmLink *_first;                                  // First link in list.
  CmLink *_last;                                   // Last link in list.

  friend CmLinkedListIterator;                     // Iterator can access.
};

class CmLinkedListIterator : public CmIterator {   // Iterator definition.
public:
  CmLinkedListIterator(const CmLinkedList& L)      // Iterator construct.
                      : CmIterator(L), _link(L._first) {}

  bool      done    () const;                      // Check if end.
  CmObject* next    ();                            // Return and advance.
  CmObject* previous();                            // Return and backup.
  CmObject* current () const;                      // Get current object.
  void      first   ();                            // Move to first object.
  void      last    ();                            // Move to last object.

  const CmLinkedList& list() const;                // Return list reference.

  CMOBJECT_DEFINE(CmLinkedListIterator)            // Define object funcs.

private:
  CmLink *_link;                                   // Current list link.
};

class CmLink {                                     // List link definition.
private:
  CmLink(CmObject *O) : _next(NULL), _data(O) {}   // Link constructor.

  CmLink   *_next;                                 // Pointer to next link.
  CmObject *_data;                                 // Pointer to object.

  friend CmLinkedList;                             // List class can access.
  friend CmLinkedListIterator;                     // Iterator class can access.
};

// "first" returns the first object in the list.
inline CmObject* CmLinkedList::first() const
{ return (_first) ? _first->_data : NULL; }

// "last" returns the last object in the list.
inline CmObject* CmLinkedList::last() const
{ return (_last) ? _last->_data : NULL; }

// "list" returns a reference to the list being iterated.
inline const CmLinkedList& CmLinkedListIterator::list() const
{ return (const CmLinkedList&) _container; }

// "done" checks if the iterator can advance any further.
inline bool CmLinkedListIterator::done() const
{ return _link == NULL; }

// "current" returns the current object in the list.
inline CmObject* CmLinkedListIterator::current() const
{ return (_link) ? _link->_data : NULL; }

// "first" resets the iterator to the start of the list.
inline void CmLinkedListIterator::first()
{ _link = list()._first; }

// "last" moves the iterator to the last object.
inline void CmLinkedListIterator::last()
{ _link = list()._last; }

#endif
