// cmobject.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Abstract object definition.
// -----------------------------------------------------------------

#ifndef _CMOBJECT_H
#define _CMOBJECT_H

#include <string.h>
#include <iostream.h>
#include <cm/include/cmdefs.h>
#include <cm/include/cmtypinf.h>
#include <cm/include/cmiofile.h>

class CmObject {                                  // Object definition.
public:
  CmObject() {}                                   // Default constructor.
 
  virtual ~CmObject() {}                          // Virtual destructor.

  virtual CmObject* newCopy() const = 0;          // Virtual copy function.

  virtual bool isEqual(const CmObject*) const;    // Compare objects.
  virtual int  compare(const CmObject*) const;    // Compare objects.

  virtual unsigned hash(unsigned) const;          // Hash function.

  virtual void printOn (ostream&) const {}        // Print to stream.
  virtual void readFrom(istream&)       {}        // Read from stream.

  virtual bool write(CmIOFile&) const;            // Write to file.
  virtual bool read (CmIOFile&);                  // Read from file.

  CMTYPEID_DEFINE                                 // Define type id stuff.
};

// "isEqual" function used by containers for object lookup.
inline bool CmObject::isEqual(const CmObject* pObj) const
{ return (pObj == this) ? true : false; }

// "compare" used by containers for object compare (binary tree, etc.).
inline int CmObject::compare(const CmObject* pObj) const
{ return ((this == pObj) ? 0 : ((this > pObj) ? 1 : -1)); }

// "hash" hashing function used by hash table classes.
inline unsigned CmObject::hash(unsigned tsize) const
{ return (unsigned)((long)this % (long)tsize); }

// "write" writes objects to the specified file.
inline bool CmObject::write(CmIOFile&) const
{ return true; }

// "read" reads objects from the specified file.
inline bool CmObject::read(CmIOFile&)
{ return true; }

// "<<" global output stream operator calls the object's print function.
inline ostream& operator<<(ostream& strm, const CmObject& obj)
{ obj.printOn(strm); return strm; }

// ">>" global input stream operator calls the object's read function.
inline istream& operator>>(istream& strm, CmObject& obj)
{ obj.readFrom(strm); return strm; }

// "<<" global output operator calls the object's write function.
inline CmIOFile& operator<<(CmIOFile& file, CmObject& obj)
{ obj.write(file); return file; }

// "<<" global input operator calls the object's read function.
inline CmIOFile& operator>>(CmIOFile& file, CmObject& obj)
{ obj.read(file); return file; }

// Define object create function type.
typedef CmObject* (*CmCreateFunc)();

// Prototype object registration functions.
extern void CmRegisterClass(const char*, CmCreateFunc);
extern void CmUnRegister   (const char*);

// Macro defines the static type info and pure virtual functions
// required for any class deriving either directly or indirectly
// from the class "CmObject".
//
// NOTE: This macro will not work if the derived class still contains
//       it's own pure virtual functions.  See the macro definition
//       immediately following this one.
//
#define CMOBJECT_DEFINE(name)                                 \
        CMTYPEID_DEFINE                                       \
        CmObject* newCopy() const { return new name(*this); }

// Macro defines the static type info and pure virtual functions
// for any class, deriving either directly or indirectly from the
// class "CmObject", which still contains pure virtual functions.
//
#define CMOBJECT_ABSTRACT(name)                \
        CMTYPEID_DEFINE                        \
        virtual CmObject* newCopy() const = 0;

// Macro initializes the static type info for any class that singly
// derives (directly or indirectly) from the class "CmObject".
//
#define CMOBJECT_INIT(name,parent) \
        CMTYPEID_INIT(name,parent)

// Macro begins the initialization of the static type info for any
// class that derives from more than one class containing run time
// type information.
//
#define BEGIN_CMOBJECT_INIT(name) \
        BEGIN_CMTYPEID_INIT(name)

// Macro is used during the initialization of the static type info
// to add another parent to the list of parents being derived from
// using run time type information.
//
#define CMOBJECT_PARENT(parent) \
        CMTYPEID_PARENT(parent)

// Macro ends the initialization of the static type info for any
// class that derives from more than one class containing run time
// type information.
//
#define END_CMOBJECT_INIT(name) \
        END_CMTYPEID_INIT(name)

// Macro registers the class for binary I/O operations and should be
// used in the source file of any class, deriving either directly
// or indirectly from "CmObject", which is to be input/output to
// a binary file.
//
#define CMOBJECT_REGISTER(name)                           \
        CmObject* CmCreate ## name() { return new name; } \
        struct CmRegister ## name                         \
        { CmRegister ## name();                           \
         ~CmRegister ## name(); };                        \
        CmRegister ## name::CmRegister ## name()          \
        { CmRegisterClass(#name, CmCreate ## name); }     \
        CmRegister ## name::~CmRegister ## name()         \
        { CmUnRegister(#name); }                          \
        CmRegister ## name register ## name ## Struct;

#endif
