// cmordarr.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Ordered array definition.
// -----------------------------------------------------------------

#ifndef _CMORDARR_H
#define _CMORDARR_H

#include <cm/include/cmcont.h>

class CmOrderedArrayIterator;                   // Iterator class stub.

class CmOrderedArray : public CmContainer {     // Ordered array definition.
public:
  CmOrderedArray(unsigned = 0, unsigned = 0);   // Default array constructor.
  CmOrderedArray(const CmOrderedArray&);        // Array copy constructor.
  CmOrderedArray(const CmContainer&);           // Construct from container.
 ~CmOrderedArray();                             // Array destructor.

  CmOrderedArray& operator=(const CmOrderedArray&);  // Assignment operator.

  void     delta  (unsigned);                   // Set delta value.
  unsigned delta  () const;                     // Get delta value.
  unsigned total  () const;                     // Return number of objects.
  bool     resize (unsigned);                   // Resize the array.
  bool     isEmpty() const;                     // Is array empty?

  bool add      (CmObject*);                    // Add object to array in order.
  bool remove   (const CmObject*);              // Remove equal object.
  bool remove   (const CmIterator&);            // Remove iterator object.
  bool removeAt (int);                          // Remove object at index.
  void removeAll();                             // Remove all objects.

  int index   (const CmObject*  ) const;        // Get index of equal object.
  int index   (const CmIterator&) const;        // Get index of iter object.
  int shouldGo(const CmObject*  ) const;        // Get index for potential obj.

  CmObject* at         (int)             const; // Get object at index.
  CmObject* lookup     (const CmObject*) const; // Look for equal object.
  bool      contains   (const CmObject*) const; // Object is in array?
  unsigned  occurrences(const CmObject*) const; // How many occurrences?

  CmIterator* newIterator() const;              // Get array iterator.

  bool write(CmIOFile&) const;                  // Write to file.
  bool read (CmIOFile&);                        // Read from file.

  CMOBJECT_DEFINE(CmOrderedArray)               // Define object funcs.

private:
  unsigned   _delta;                            // Delta value.
  unsigned   _total;                            // Number of objects.
  CmObject **_entries;                          // Array of object pointers.

  friend CmOrderedArrayIterator;                // Iterator can access,
};

class CmOrderedArrayIterator : public CmIterator {  // Iterator definition.
public:
  CmOrderedArrayIterator(const CmOrderedArray& arr) // Iterator constructor.
                        : CmIterator(arr), _index(0) {}

  bool      done    () const;                   // Check if done iterating.
  CmObject* next    ();                         // Return and advance.
  CmObject* previous();                         // Return and backup.
  CmObject* current () const;                   // Return current object.
  void      first   ();                         // Move to first object.
  void      last    ();                         // Move to last object.

  const CmOrderedArray& array() const;          // Return array reference.

  CMOBJECT_DEFINE(CmOrderedArrayIterator)       // Define object funcs.

private:
  int _index;                                   // Current array index.
};

// "delta" sets a new delta value for automatic growing.
inline void CmOrderedArray::delta(unsigned dt)
{ _delta = dt; }

// "delta" returns the current delta value.
inline unsigned CmOrderedArray::delta() const
{ return _delta; }

// "total" returns the number of objects added to this array.
inline unsigned CmOrderedArray::total() const
{ return _total; }

// "at" returns the object at the specified index.
inline CmObject* CmOrderedArray::at(int idx) const
{ return (idx >= 0 && idx < _total) ? _entries[idx] : NULL; }

// "contains" checks if the array contains an object equal to the input.
inline bool CmOrderedArray::contains(const CmObject* pObj) const
{ return (index(pObj) > -1) ? true : false; }

// "array" returns a reference to the array being iterated.
inline const CmOrderedArray& CmOrderedArrayIterator::array() const
{ return (const CmOrderedArray&) _container; }

// "done" checks to see if we can iterate any further.
inline bool CmOrderedArrayIterator::done() const
{ return (_index >= 0 && _index < array()._total) ? false : true; }

// "next" returns the current object and advances the iterator.
inline CmObject* CmOrderedArrayIterator::next()
{ return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index++] : NULL; }

// "previous" returns the current object and decrements the iterator.
inline CmObject* CmOrderedArrayIterator::previous()
{ return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index--] : NULL; }

// "current" returns the current object pointed to by the iterator.
inline CmObject* CmOrderedArrayIterator::current() const
{ return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index] : NULL; }

// "first" moves the iterator to the first object.
inline void CmOrderedArrayIterator::first()
{ _index = 0; }

// "last" moves the iterator to the last object.
inline void CmOrderedArrayIterator::last()
{ _index = array()._total - 1; }

#endif
