// cmstring.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Character string definition.
// -----------------------------------------------------------------

#ifndef _CMSTRING_H
#define _CMSTRING_H

#include <cm/include/cmobject.h>

class CmSubString;                                  // Substring class stub.

class CmString : public CmObject {                  // String definition.
public:
  CmString(const char* = NULL, int = -1);           // Default constructor.
  CmString(int, char);                              // Construct w/ char.
  CmString(const CmString&, int = -1);              // Copy constructor.
  CmString(const CmSubString&);                     // Construct /w substr.
 ~CmString();                                       // Destructor.

  CmString& operator=(const CmString&);             // Assign string.
  CmString& operator=(const char*);                 // Assign string.
  CmString& operator=(char);                        // Assign char.

  CmString& operator+=(const CmString&);            // Append string.
  CmString& operator+=(const char*);                // Append string.
  CmString& operator+=(char);                       // Append char.

  char  operator[](int) const;                      // Get char at index.
  char& operator[](int);                            // Get char at index.

        CmSubString operator()(int, int);           // Substring operator.
  const CmSubString operator()(int, int) const;     // SubString operator.

  operator const char*() const;                     // Type conversion op.

  const char* text() const;                         // Return char pointer.

  unsigned length() const;                          // Return string length.

  CmString& assign(const CmString&, int = -1);      // Assign string.
  CmString& assign(const char*,     int = -1);      // Assign string.
  CmString& assign(char,            int =  1);      // Assign char(s).

  CmString& append(const CmString&, int = -1);      // Append string.
  CmString& append(const char*,     int = -1);      // Append string.
  CmString& append(char,            int =  1);      // Append char(s).

  CmString& insert(int, const CmString&, int = -1); // Insert string at pos.
  CmString& insert(int, const char*,     int = -1); // Insert string at pos.
  CmString& insert(int, char,            int =  1); // Insert char(s) at pos.

  CmString& replace(int, int, const CmString&, int = -1); // Replace char(s).
  CmString& replace(int, int, const char*,     int = -1); // Replace char(s).
  CmString& replace(int, int, char,            int =  1); // Replace char(s).

  CmString& remove(int = 0, int = -1);              // Remove character(s).

  int index(const CmString&, int = -1) const;       // Find string in string.
  int index(const char*,     int = -1) const;       // Find string in string.
  int index(char)                      const;       // Find char in string.

  CmString substr(int, int) const;                  // Return a substring.

        CmSubString subString(int, int);            // Return a substring.
  const CmSubString subString(int, int) const;      // Return a substring.

  int compare(const CmString&, int = -1) const;     // Compare strings.
  int compare(const char*,     int = -1) const;     // Compare strings.

  void toLower     ();                              // Lower case the string.
  void toUpper     ();                              // Upper case the string.
  void leftJustify ();                              // Left justify string.
  void rightJustify();                              // Right justify string.
  void trim        ();                              // Trim lead/trail blanks.

  int copy(char*, int, int) const;                  // Copy into char array.

  bool isEqual(const CmObject*) const;              // Compare objects.
  int  compare(const CmObject*) const;              // Compare objects.

  unsigned hash(unsigned) const;                    // String hash function.

  void printOn (ostream&) const;                    // Print string to stream.
  void readFrom(istream&);                          // Read string from stream.

  bool write(CmIOFile&) const;                      // Write string to file.
  bool read (CmIOFile&);                            // Read string from file.

  CMOBJECT_DEFINE(CmString)                         // Define base functions.

  static void caseSensitive(bool);                  // Set case sensitive.
  static bool caseSensitive();                      // Get case sensitive.

private:
  bool resize(unsigned);                            // Resize string capacity.

  char* _text;                                      // Character array.

  static bool _caseSensitive;                       // Case sensitive compare.

  friend CmSubString;                               // Substring can access.
};

class CmSubString {                                 // Substring definition.
public:
  CmString& operator=(const CmString&);             // Assignment from string.
  CmString& operator=(const CmSubString&);          // Assignment from sub-str.
  CmString& operator=(const char*);                 // Assignment from char*.

  CmString& assign(const CmString&,    int = -1);   // Assign string.
  CmString& assign(const CmSubString&, int = -1);   // Assign substring.
  CmString& assign(const char*,        int = -1);   // Assign string.

  int compare(const CmString&)    const;            // Compare w/ string.
  int compare(const CmSubString&) const;            // Compare w/ substring.
  int compare(const char*)        const;            // Compare w/ string.

  friend ostream& operator<<(ostream&, const CmSubString&); // Stream output.

private:
  CmSubString(const CmString&, int, int);           // Construct w/ string.
  CmSubString(const CmSubString&);                  // Copy constructor.

  void replace(const char*, int);                   // Replace contents.

  char     *_sp;                                    // Sub string text.
  int       _sl;                                    // Length of sub-str.
  CmString *_st;                                    // String this is sub of.

  friend CmString;                                  // String can access.
};

// String concatenation functions.
CmString operator+(const CmString&, const CmString&);
CmString operator+(const CmString&, const char*);
CmString operator+(const char*, const CmString&);
CmString operator+(const CmString&, char);
CmString operator+(char, const CmString&);

// "=" assigns the specified string to this string.
inline CmString& CmString::operator=(const CmString& str)
{ return assign(str); }

// "=" assigns the specified string to this string.
inline CmString& CmString::operator=(const char* str)
{ return assign(str); }

// "=" assigns the specified character to this string.
inline CmString& CmString::operator=(char c)
{ return assign(c); }

// "+=" appends the specified string onto this string.
inline CmString& CmString::operator+=(const CmString& str)
{ return append(str); }

// "+=" appends the specified string onto this string.
inline CmString& CmString::operator+=(const char* str)
{ return append(str); }

// "+=" appends the specified character onto this string.
inline CmString& CmString::operator+=(char c)
{ return append(c); }

// "[]" returns the character at the specified index.
inline char CmString::operator[](int idx) const
{ return _text[idx]; }

// "[]" returns a reference to the character at the specified index.
inline char& CmString::operator[](int idx)
{ return _text[idx]; }

// "text" returns the internal char pointer.
inline const char* CmString::text() const
{ return _text; }

// "length" returns the current string length.
inline unsigned CmString::length() const
{ return (_text) ? strlen(_text) : 0; }

// "const char*" operator returns a pointer to the char array.
inline CmString::operator const char*() const
{ return _text; }

// "()" returns a substring containing n chars at idx from this string.
inline CmSubString CmString::operator()(int idx, int n)
{ return subString(idx, n); }

// "()" returns a substring containing n chars at idx from this string.
inline const CmSubString CmString::operator()(int idx, int n) const
{ return subString(idx, n); }

// "<" checks if the first string is less than the second.
inline bool operator<(const CmString& s1, const CmString& s2)
{ return (s1.compare(s2) < 0) ? true : false; }

// "<=" checks if the first string is less than or equal to the second.
inline bool operator<=(const CmString& s1, const CmString& s2)
{ return (s1.compare(s2) <= 0) ? true : false; }

// ">" checks if the first string is greater than the second.
inline bool operator>(const CmString& s1, const CmString& s2)
{ return (s1.compare(s2) > 0) ? true : false; }

// ">=" checks if the first string is greater than or equal to the second.
inline bool operator>=(const CmString& s1, const CmString& s2)
{ return (s1.compare(s2) >= 0) ? true : false; }

// "==" checks if the first string is equal to the second.
inline bool operator==(const CmString& s1, const CmString& s2)
{ return (s1.compare(s2) == 0) ? true : false; }

// "!=" checks if the first string is not equal to the second.
inline bool operator!=(const CmString& s1, const CmString& s2)
{ return (s1.compare(s2) != 0) ? true : false; }

// "<" checks if the first string is less than the second.
inline bool operator<(const CmString& s1, const char* s2)
{ return (s1.compare(s2) < 0) ? true : false; }

// "<=" checks if the first string is less than or equal to the second.
inline bool operator<=(const CmString& s1, const char* s2)
{ return (s1.compare(s2) <= 0) ? true : false; }

// ">" checks if the first string is greater than the second.
inline bool operator>(const CmString& s1, const char* s2)
{ return (s1.compare(s2) > 0) ? true : false; }

// ">=" checks if the first string is greater than or equal to the second.
inline bool operator>=(const CmString& s1, const char* s2)
{ return (s1.compare(s2) >= 0) ? true : false; }

// "==" checks if the first string is equal to the second.
inline bool operator==(const CmString& s1, const char* s2)
{ return (s1.compare(s2) == 0) ? true : false; }

// "!=" checks if the first string is not equal to the second.
inline bool operator!=(const CmString& s1, const char* s2)
{ return (s1.compare(s2) != 0) ? true : false; }

// "<" checks if the first string is less than the second.
inline bool operator<(const char* s1, const CmString& s2)
{ return (s2.compare(s1) > 0) ? true : false; }

// "<=" checks if the first string is less than or equal to the second.
inline bool operator<=(const char* s1, const CmString& s2)
{ return (s2.compare(s1) >= 0) ? true : false; }

// ">" checks if the first string is greater than the second.
inline bool operator>(const char* s1, const CmString& s2)
{ return (s2.compare(s1) < 0) ? true : false; }

// ">=" checks if the first string is greater than or equal to the second.
inline bool operator>=(const char* s1, const CmString& s2)
{ return (s2.compare(s1) <= 0) ? true : false; }

// "==" checks if the first string is equal to the second.
inline bool operator==(const char* s1, const CmString& s2)
{ return (s2.compare(s1) == 0) ? true : false; }

// "!=" checks if the first string is not equal to the second.
inline bool operator!=(const char* s1, const CmString& s2)
{ return (s2.compare(s1) != 0) ? true : false; }

// "=" assigns the specified string to this substring.
inline CmString& CmSubString::operator=(const CmString& str)
{ return assign(str); }

// "=" assigns the specified string to this substring.
inline CmString& CmSubString::operator=(const CmSubString& str)
{ return assign(str); }

// "=" assigns the specified string to this substring.
inline CmString& CmSubString::operator=(const char* str)
{ return assign(str); }

// "<" checks if the first string is less than the second.
inline bool operator<(const CmSubString& s1, const CmString& s2)
{ return (s1.compare(s2) < 0) ? true : false; }

// "<=" checks if the first string is less than or equal to the second.
inline bool operator<=(const CmSubString& s1, const CmString& s2) 
{ return (s1.compare(s2) <= 0) ? true : false; }

// ">" checks if the first string is greater than the second.
inline bool operator>(const CmSubString& s1, const CmString& s2) 
{ return (s1.compare(s2) > 0) ? true : false; }

// ">=" checks if the first string is greater than or equal to the second.
inline bool operator>=(const CmSubString& s1, const CmString& s2) 
{ return (s1.compare(s2) >= 0) ? true : false; }

// "==" checks if the first string is equal to the second.
inline bool operator==(const CmSubString& s1, const CmString& s2) 
{ return (s1.compare(s2) == 0) ? true : false; }

// "!=" checks if the first string is not equal to the second.
inline bool operator!=(const CmSubString& s1, const CmString& s2) 
{ return (s1.compare(s2) != 0) ? true : false; }

// "<" checks if the first string is less than the second.
inline bool operator<(const CmSubString& s1, const CmSubString& s2)
{ return (s1.compare(s2) < 0) ? true : false; }

// "<=" checks if the first string is less than or equal to the second.
inline bool operator<=(const CmSubString& s1, const CmSubString& s2)
{ return (s1.compare(s2) <= 0) ? true : false; }

// ">" checks if the first string is greater than the second.
inline bool operator>(const CmSubString& s1, const CmSubString& s2)
{ return (s1.compare(s2) > 0) ? true : false; }

// ">=" checks if the first string is greater than or equal to the second.
inline bool operator>=(const CmSubString& s1, const CmSubString& s2)
{ return (s1.compare(s2) >= 0) ? true : false; }

// "==" checks if the first string is equal to the second.
inline bool operator==(const CmSubString& s1, const CmSubString& s2)
{ return (s1.compare(s2) == 0) ? true : false; }

// "!=" checks if the first string is not equal to the second.
inline bool operator!=(const CmSubString& s1, const CmSubString& s2)
{ return (s1.compare(s2) != 0) ? true : false; }

// "<" checks if the first string is less than the second.
inline bool operator<(const CmSubString& s1, const char* s2)
{ return (s1.compare(s2) < 0) ? true : false; }

// "<=" checks if the first string is less than or equal to the second.
inline bool operator<=(const CmSubString& s1, const char* s2)
{ return (s1.compare(s2) <= 0) ? true : false; }

// ">" checks if the first string is greater than the second.
inline bool operator>(const CmSubString& s1, const char* s2)
{ return (s1.compare(s2) > 0) ? true : false; }

// ">=" checks if the first string is greater than or equal to the second.
inline bool operator>=(const CmSubString& s1, const char* s2)
{ return (s1.compare(s2) >= 0) ? true : false; }

// "==" checks if the first string is equal to the second.
inline bool operator==(const CmSubString& s1, const char* s2)
{ return (s1.compare(s2) == 0) ? true : false; }

// "!=" checks if the first string is not equal to the second.
inline bool operator!=(const CmSubString& s1, const char* s2)
{ return (s1.compare(s2) != 0) ? true : false; }

// "<" checks if the first string is less than the second.
inline bool operator<(const char* s1, const CmSubString& s2)
{ return (s2.compare(s1) > 0) ? true : false; }

// "<=" checks if the first string is less than or equal to the second.
inline bool operator<=(const char* s1, const CmSubString& s2)
{ return (s2.compare(s1) >= 0) ? true : false; }

// ">" checks if the first string is greater than the second.
inline bool operator>(const char* s1, const CmSubString& s2)
{ return (s2.compare(s1) < 0) ? true : false; }

// ">=" checks if the first string is greater than or equal to the second.
inline bool operator>=(const char* s1, const CmSubString& s2)
{ return (s2.compare(s1) <= 0) ? true : false; }

// "==" checks if the first string is equal to the second.
inline bool operator==(const char* s1, const CmSubString& s2)
{ return (s2.compare(s1) == 0) ? true : false; }

// "!=" checks if the first string is not equal to the second.
inline bool operator!=(const char* s1, const CmSubString& s2)
{ return (s2.compare(s1) != 0) ? true : false; }

#endif
