// cmtbintr.cc
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Binary tree template implementation.
// -----------------------------------------------------------------


// "CmTBinaryTree" is the default tree constructor.
//
template <class T> CmTBinaryTree<T>::CmTBinaryTree()
                                   : _root(NULL)
{}


// "CmTBinaryTree" is the tree copy constructor.
//
template <class T>
CmTBinaryTree<T>::CmTBinaryTree(const CmTBinaryTree<T>& tree)
                : CmTContainer<T>(tree), _root(NULL)
{
  copy   (tree);
  balance();
}


// "CmTBinaryTree" constructs a binary tree from the specified
// container.
//
template <class T>
CmTBinaryTree<T>::CmTBinaryTree(const CmTContainer<T>& cont)
                : _root(NULL)
{
  copy   (cont);
  balance();
}


// "~CmTBinaryTree" is the tree destructor.
//
template <class T> CmTBinaryTree<T>::~CmTBinaryTree()
{
  removeAll();
}


// "=" assignment operator copies the specified tree into this tree.
//
template <class T>
CmTBinaryTree<T>& CmTBinaryTree<T>::operator=(const CmTBinaryTree<T>& tree)
{
  if (&tree != this)
  {
    removeAll();
    copy     (tree);
    balance  ();
  }
  return *this;
}


// "add" adds a new item to the tree.
//
template <class T> bool CmTBinaryTree<T>::add(const T& rObj)
{
  CmTBinaryTreeNode<T> *newnode = new CmTBinaryTreeNode<T>(rObj);
  if (!newnode) return false;

  if (!_root)                                // Tree is empty.  Add at root.
    _root = newnode;

  else                                       // Tree is not empty.
  {
    CmTBinaryTreeNode<T> *parent = _root;
    CmTBinaryTreeNode<T> *rover;
    rover = (rObj < _root->_data) ? _root->_left : _root->_right;

    while (rover)                            // Search for vacant slot.
    {
      parent = rover;
      rover  = (rObj < rover->_data) ? rover->_left : rover->_right;
    }

    if (rObj < parent->_data) parent->_left  = newnode;  // Add new node.
    else                      parent->_right = newnode;
  }
  _size++;
  return true;
}


// "remove" removes the specified item from the tree.
//
template <class T> bool CmTBinaryTree<T>::remove(const T& rObj)
{
  if (!_root) return false;                    // Exit if empty tree.

  CmTBinaryTreeNode<T> *rover   = _root;       // Init for search.
  CmTBinaryTreeNode<T> *parent  = NULL;
  bool                  done    = false;
  bool                  success = false;

  while (rover && !done)                       // Search for item.
  {
    if (rObj != rover->_data)                  // Advance to next node.
    {
      parent = rover;
      rover  = (rObj < rover->_data) ? rover->_left : rover->_right;
    }

    else                                       // Item was found.
    {
      success = removeNode(parent, rover);
      done    = true;
    }
  }
  return success;
}


// "remove" removes the item pointed to by the specified iterator.
//
template <class T> bool CmTBinaryTree<T>::remove(const CmTIterator<T>& iter)
{
  return remove(iter.current());
}


// "removeAt" removes the item at the specified tree index.
//
template <class T> bool CmTBinaryTree<T>::removeAt(int idx)
{
  if (!_root) return false;
  CmTStack<CmTBinaryTreeNode<T>*> stack;
  CmTBinaryTreeNode<T>* node = _root;
  while (node->_left)
  {
    stack.push(node);
    node = node->_left;
  }

  int ix = 0;
  while (node && ix < idx)
  {
    if (node->_right)
    {
      stack.push(node);
      node = node->_right;
      while (node->_left)
      {
        stack.push(node);
        node = node->_left;
      }
    }
    else
    {
      CmTBinaryTreeNode<T> *temp;
      do
      {
        temp = node;
        node = (stack.size() == 0) ? NULL : stack.pop();
      } while (node && node->_right == temp);
    }
    ix++;
  }
  CmTBinaryTreeNode<T> *par = (stack.size() > 0) ? stack.pop() : NULL;
  return (ix == idx) ? removeNode(par, node) : false;
}


// "removeAll" removes all items from this tree.
//
template <class T> void CmTBinaryTree<T>::removeAll()
{
  if (_root)
  {
    killNode(_root);
    delete _root;
    _root = NULL;
    _size = 0;
  }
}


// "root" returns the item at the root of this tree.
//
template <class T> const T& CmTBinaryTree<T>::root() const
{
  return (_root) ? _root->_data : _error;
}


// "balance" balances the current contents of the tree.
//
template <class T> void CmTBinaryTree<T>::balance()
{
  CmTQueue<T> que;
  CmTBinaryTreeIterator<T> iter(*this);
  while (!iter.done()) que.push(iter.next());
  removeAll();
  balanceHalf(que, 0, que.size() - 1);
}


// "lookup" returns the item in the tree equal to the specified item.
//
template <class T> const T& CmTBinaryTree<T>::lookup(const T& rObj) const
{
  if (!_root) return _error;

  CmTBinaryTreeNode<T> *rover = _root;
  CmTBinaryTreeNode<T> *temp  = NULL;

  while (rover && !temp)
  {
    if (rObj == rover->_data)
      temp  = rover;
    else
      rover = (rObj < rover->_data) ? rover->_left : rover->_right;
  }
  return (temp) ? temp->_data : _error;
}


// "contains" checks if the specified item is in this tree.
//
template <class T> bool CmTBinaryTree<T>::contains(const T& rObj) const
{
  if (!_root) return NULL;

  CmTBinaryTreeNode<T> *rover = _root;
  bool                  found = false;

  while (rover && !found)
  {
    if (rObj == rover->_data)
      found = true;
    else
      rover = (rObj < rover->_data) ? rover->_left : rover->_right;
  }
  return found;
}


// "occurrences" counts the occurrences of the specified item in this tree.
//
template <class T> unsigned CmTBinaryTree<T>::occurrences(const T& rObj) const
{
  unsigned                 occur = 0;
  CmTBinaryTreeIterator<T> iter(*this);
  while (!iter.done())
    if (iter.next() == rObj) occur++;
  return occur;
}


// "newIterator" creates and returns a new tree iterator.
//
template <class T> CmTIterator<T>* CmTBinaryTree<T>::newIterator() const
{
  return new CmTBinaryTreeIterator<T>(*this);
}


// "removeNode" removes the specified node from the tree.
//
template <class T>
bool CmTBinaryTree<T>::removeNode(CmTBinaryTreeNode<T>* parent,
                                  CmTBinaryTreeNode<T>* node)
{
  CmTBinaryTreeNode<T> *temp1 = node;

  if (node->_right == NULL)
    node = node->_left;

  else if (node->_right->_left == NULL)
  {
    node        = node->_right;
    node->_left = temp1->_left;
  }

  else
  {
    CmTBinaryTreeNode<T> *temp2 = node->_right;
    while (temp2->_left->_left) temp2 = temp2->_left;
    node         = temp2->_left;
    temp2->_left = node->_right;
    node->_left  = temp1->_left;
    node->_right = temp1->_right;
  }

  if (!parent)
    _root = node;
  else
  {
    if (temp1->_data < parent->_data) parent->_left  = node;
    else                              parent->_right = node;
  }

  delete temp1;
  _size--;
  return true;
}


// "balanceHalf" is called recursively for tree balancing.
// 
template <class T>
void CmTBinaryTree<T>::balanceHalf(const CmTQueue<T>& que, int low, int high)
{
  if (low <= high)
  {
    int mid = (low + high) / 2;
    add(que[mid]);
    balanceHalf(que, mid + 1, high);
    balanceHalf(que, low, mid-1);
  }
}


// "killNode" is recursively called to destroy branches of a node.
//
template <class T> void CmTBinaryTree<T>::killNode(CmTBinaryTreeNode<T>* sub)
{
  if (sub->_left)
  {
    killNode(sub->_left);
    delete sub->_left;
    sub->_left = NULL;
  }

  if (sub->_right)
  {
    killNode(sub->_right);
    delete sub->_right;
    sub->_right = NULL;
  }
}


// "CmTBinaryTreeIterator" is the iterator constructor.
//
template <class T>
CmTBinaryTreeIterator<T>::CmTBinaryTreeIterator(const CmTBinaryTree<T>& tr)
                        : CmTIterator<T>(tr), _node(tr._root)
{
  if (_node)
  {
    while (_node->_left)
    {
      _stack.push(_node);
      _node = _node->_left;
    }
  }
}


// "tree" returns a reference to the tree being iterated.
//
template <class T>
const CmTBinaryTree<T>& CmTBinaryTreeIterator<T>::tree() const
{
  return (const CmTBinaryTree<T>&) _container;
}


// "done" returns true if the iterator can proceed no further.
//
template <class T> bool CmTBinaryTreeIterator<T>::done() const
{
  return (!_node) ? true : false;
}


// "next" returns the current item and advances the iterator.
//
template <class T> const T& CmTBinaryTreeIterator<T>::next()
{
  if (!_node) return tree()._error;               // Iteration is over.

  const T& retValue = _node->_data;               // Save item reference.

  if (_node->_right)                              // If there is a right
  {                                               // branch, advance to it
    _stack.push(_node);                           // and descend all the way
    _node = _node->_right;                        // down it's left side.
    while (_node->_left)
    {
      _stack.push(_node);
      _node = _node->_left;
    }
  }

  else                                            // Otherwise, back up the
  {                                               // tree until we are back
    CmTBinaryTreeNode<T>* temp;                   // at the top or until we
    do                                            // are no longer a right
    {                                             // branch.
      temp  = _node;
      _node = (_stack.size() == 0) ? NULL : _stack.pop();
    } while (_node && _node->_right == temp);
  }
  return retValue;
}


// "previous" returns the current item and advances the iterator.
//
template <class T> const T& CmTBinaryTreeIterator<T>::previous()
{
  if (!_node) return tree()._error;               // Iteration is over.

  const T& retValue = _node->_data;               // Save item reference.

  if (_node->_left)                               // If there is a left
  {                                               // branch, advance to it
    _stack.push(_node);                           // and descend all the way
    _node = _node->_left;                         // down it's right side.
    while (_node->_right)
    {
      _stack.push(_node);
      _node = _node->_right;
    }
  }

  else                                            // Otherwise, back up the
  {                                               // tree until we are back
    CmTBinaryTreeNode<T>* temp;                   // at the top or until we
    do                                            // are no longer a left
    {                                             // branch.
      temp  = _node;
      _node = (_stack.size() == 0) ? NULL : _stack.pop();
    } while (_node && _node->_left == temp);
  }
  return retValue;
}


// "current" returns the curren item.
//
template <class T> const T& CmTBinaryTreeIterator<T>::current() const
{
  return (_node) ? _node->_data : tree()._error;
}


// "first" moves the iterator to the first item.
//
template <class T> void CmTBinaryTreeIterator<T>::first()
{
  _stack.removeAll();
  _node = tree()._root;
  if (_node)
  {
    while (_node->_left)
    {
      _stack.push(_node);
      _node = _node->_left;
    }
  }
}


// "last" moves the iterator to the last item.
//
template <class T> void CmTBinaryTreeIterator<T>::last()
{
  _stack.removeAll();
  _node = tree()._root;
  if (_node)
  {
    while (_node->_right)
    {
      _stack.push(_node);
      _node = _node->_right;
    }
  }
}
