// cmtcont.cc
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Abstract container template implementation.
// -----------------------------------------------------------------


// "CmTContainer" is the container copy constructor.
//
template <class T> CmTContainer<T>::CmTContainer(const CmTContainer<T>&)
                                  : _error(* new T)
{}


// "~CmTContainer" is the container destructor.
//
template <class T> CmTContainer<T>::~CmTContainer()
{
  delete &_error;
}


// "[]" indexing operator returns the item at the specified index.
//
template <class T> const T& CmTContainer<T>::operator[](int idx) const
{
  return at(idx);
}


// "size" returns the current container size.
//
template <class T> unsigned CmTContainer<T>::size() const
{
  return _size;
}


// "total" returns the total number of items.
//
template <class T> unsigned CmTContainer<T>::total() const
{
  return _size;
}


// "isEmpty" returns true if the container is empty.
//
template <class T> bool CmTContainer<T>::isEmpty() const
{
  return (_size == 0);
}


// "at" returns the item at the specified index.
//
template <class T> const T& CmTContainer<T>::at(int idx) const
{
  if (idx < 0 || idx >= size()) return _error;
  CmTIterator<T> *iter = newIterator();
  int             ii   = 0;
  const T*        out;

  while (!iter->done() && ii <= idx)
  {
    if (ii++ == idx) out = & iter->current();
    iter->next();
  }
  delete iter;
  return *out;
}


// "lookup" returns a reference to the first occurrence of an item
// that is equal to the specified item.
//
template <class T> const T& CmTContainer<T>::lookup(const T& rItem) const
{
  CmTIterator<T> *iter = newIterator();
  const T*        out  = NULL;
  while (!iter->done() && !out)
  {
    const T& temp = iter->next();
    if (temp == rItem) out = &temp;
  }
  delete iter;
  return *out;
}


// "contains" returns true if the container contains an item that
// is equal to the specified item.
//
template <class T> bool CmTContainer<T>::contains(const T& rItem) const
{
  CmTIterator<T> *iter  = newIterator();
  bool            found = false;
  while (!iter->done() && !found)
    if (rItem == iter->next()) found = true;
  delete iter;
  return found;
}


// "occurrences" returns the number of items in this container
// that are equal to the specified item.
//
template <class T> unsigned CmTContainer<T>::occurrences(const T& rItem) const
{
  CmTIterator<T> *iter  = newIterator();
  unsigned        occur = 0;
  while (!iter->done())
    if (rItem == iter->next())
      occur++;
  delete iter;
  return occur;
}


#include <cm/include/cmtlist.h>

// "lookupAll" allocates and returns a container with all items
// in this container equal to the specified item.
//
template <class T>
CmTContainer<T>* CmTContainer<T>::lookupAll(const T& rItem) const
{
  CmTLinkedList<T> *pList = new CmTLinkedList<T>;
  CmTIterator<T>   *iter  = newIterator();
  while (!iter->done())
  {
    const T& temp = iter->next();
    if (temp == rItem) pList->add(temp);
  }
  delete iter;
  return pList;
}


// "addAllFrom" adds all the items from the specified container to
// this container.
//
template <class T>
bool CmTContainer<T>::addAllFrom(const CmTContainer<T>* container)
{
  CmTIterator<T> *iter    = container->newIterator();
  bool            success = true;
  while (!iter->done() && success)
    success = add(iter->next());
  delete iter;
  return success;
}


// "forEach" iterates through all the items in the container calling
// the specified function for each item until there are not more items
// of the function returns false.
//
template <class T> void
CmTContainer<T>::forEach(bool (*pFunc)(const T&, void*), void* pData)
{
  bool            cont = true;
  CmTIterator<T>* iter = newIterator();
  while (!iter->done() && cont)
    cont = (*pFunc)(iter->next(), pData);
  delete iter;
}


// "firstThat" returns the first item in the container that is passed
// into the specified function and returns true.
//
template <class T> T
CmTContainer<T>::firstThat(bool (*pFunc)(const T&, void*), void* pData) const
{
  T               outItem;
  bool            found  = false;
  CmTIterator<T>* iter  = newIterator();
  while (!iter->done() && !found)
  {
    const T& rItem = iter->next();
    if ((*pFunc)(rItem, pData))
    {
      outItem = rItem;
      found   = true;
    }
  }
  delete iter;
  return outItem;
}


// "lastThat" returns the last item in the container that is passed
// into the specified function and returns true.
//
template <class T> T
CmTContainer<T>::lastThat(bool (*pFunc)(const T&, void*), void* pData) const
{
  T               outItem;
  bool            found = false;
  CmTIterator<T>* iter  = newIterator();
  iter->last();
  while (!iter->done() && !found)
  {
    const T& rItem = iter->previous();
    if ((*pFunc)(rItem, pData))
    {
      outItem = rItem;
      found   = true;
    }
  }
  delete iter;
  return outItem;
}


// "query" cycles through all the items in the container calling the
// specified function for each.  Every item in which the function
// returns true gets added to a subset container which is then returned
// from the function.
//
template <class T> CmTContainer<T>*
CmTContainer<T>::query(bool (*pFunc)(const T&, void*), void* pData) const
{
  CmTLinkedList<T>* pList = new CmTLinkedList<T>;
  CmTIterator<T>*   iter  = newIterator();
  while (!iter->done())
  {
    const T& rItem = iter->next();
    if ((*pFunc)(rItem, pData)) pList->add(rItem);
  }
  delete iter;
  return pList;
}


// "copy" copies the contents of the specified container into
// this container.
//
template <class T> void CmTContainer<T>::copy(const CmTContainer<T>& C)
{
  _size = C._size;
  CmTIterator<T> *iter = C.newIterator();
  while (!iter->done()) add(iter->next());
  delete iter;
}
