// cmtcont.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Abstract container template definition.
// -----------------------------------------------------------------

#ifndef _CMTCONT_H
#define _CMTCONT_H

#include <cm/include/cmtiter.h>

template <class T> class CmTContainer {              // Container definition.
public:
  typedef bool (*CmTQueryFunc)(const T&, void*);     // Query function type.

  CmTContainer() : _error(* new T), _size(0) {}      // Default constructor.
  CmTContainer(const CmTContainer<T>&);              // Copy constructor.

  virtual ~CmTContainer();                           // Virtual destructor.

  const T& operator[](int) const;                    // Get object at index.

  virtual unsigned size   () const;                  // Return container size.
  virtual unsigned total  () const;                  // Return number of items.
  virtual bool     isEmpty() const;                  // Is container empty?

  virtual bool add      (const T&)              = 0; // Add item to container.
  virtual bool remove   (const T&)              = 0; // Remove item.
  virtual bool remove   (const CmTIterator<T>&) = 0; // Remove iterator item.
  virtual bool removeAt (int)                   = 0; // Remove at index.
  virtual void removeAll()                      = 0; // Remove all items.

  virtual const T& at         (int)      const;      // Get object at index.
  virtual const T& lookup     (const T&) const;      // Find equal object.
  virtual bool     contains   (const T&) const;      // Container has object?
  virtual unsigned occurrences(const T&) const;      // Count occurrences.

  virtual CmTIterator<T>* newIterator() const = 0;   // Get container iterator.

  CmTContainer<T>* lookupAll (const T&) const;         // Find all equal items.
  bool             addAllFrom(const CmTContainer<T>*); // Combine containers.

  // Call the specified function for every item in the container
  // until the end is reached or the function returns false.
  void forEach(CmTQueryFunc, void* = NULL);

  // Return the first item to satisfy the conditional test performed
  // in the specified function.
  T firstThat(CmTQueryFunc, void* = NULL) const;

  // Return the last item to satisfy the conditional test performed
  // in the specified function.
  T lastThat(CmTQueryFunc, void* = NULL) const;

  // Call the specified function for every item in the container
  // adding all items for which the function returns true to an
  // output container.
  CmTContainer<T>* query(CmTQueryFunc, void* = NULL) const;

protected:
  void copy(const CmTContainer<T>&);                // Internal copy method.

  T&       _error;                                  // Error object.
  unsigned _size;                                   // Container size.
};

#if defined(__TURBOC__) || defined(__xlC__) || defined(_MSC_VER) || defined(__GNUG__)
#include <cm/include/cmtcont.cc>
#endif

// "CmTForEach" and "CmTEndFor" define a foreach loop where the item,
// type and collection are passed in and the item will be set equal to
// the next item in the collection for each loop iteration.
//
#define CmTForEach(obj, type, cltn) {                           \
        CmTIterator<type> *___cm_iterator = cltn.newIterator(); \
        while (!___cm_iterator->done()) {                       \
          obj = ___cm_iterator->next();
#define CmTEndFor } delete ___cm_iterator; }

// "CmTRepeat" and "CmTUntil" define a pascal style repeat/until loop
// where the item, type, and collection are passed in and the item will
// be set equal to next item in the collection for each loop iteration.
//
#define CmTRepeat(obj, type, cltn) {                            \
        CmTIterator<type> *___cm_iterator = cltn.newIterator(); \
        do {                                                    \
          obj = ___cm_iterator->next();
#define CmTUntil(condition) } while (!(condition)); delete ___cm_iterator; }

// "CmTDo" and "CmTWhile" define a typical do/while loop where the item,
// type, and collection are passed in and the item will be set equal to
// the next item in the collection for each loop iteration.
//
#define CmTDo(obj, type, cltn) {                                \
        CmTIterator<type> *___cm_iterator = cltn.newIterator(); \
        do {                                                    \
          obj = ___cm_iterator->next();
#define CmTWhile(condition) } while (condition); delete ___cm_iterator; }

#endif
