// cmtdlist.cc
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Doubly linked list template implementation.
// -----------------------------------------------------------------


// "CmTDLinkedList" is the list copy constructor.
//
template <class T>
CmTDLinkedList<T>::CmTDLinkedList(const CmTDLinkedList<T>& list)
                : CmTContainer<T>(list), _first(NULL), _last(NULL)
{
  copy(list);
}


// "CmTDLinkedList" constructs a list from the specified container.
//
template <class T>
CmTDLinkedList<T>::CmTDLinkedList(const CmTContainer<T>& cont)
                 : _first(NULL), _last(NULL)
{
  copy(cont);
}


// "~CmTDLinkedList" is the list destructor.
//
template <class T> CmTDLinkedList<T>::~CmTDLinkedList()
{
  removeAll();
}


// "=" operator copies the contents of the input list into this list.
//
template <class T>
CmTDLinkedList<T>& CmTDLinkedList<T>::operator=(const CmTDLinkedList<T>& list)
{
  if (&list != this)
  {
    removeAll();
    copy     (list);
  }
  return *this;
}


// "add" adds the specified item to the end of this list.
//
template <class T> bool CmTDLinkedList<T>::add(const T& rObj)
{
  return append(rObj);
}


// "append" adds the specified item to the end of this list.
//
template <class T> bool CmTDLinkedList<T>::append(const T& obj)
{
  CmTDLink<T> *newlink = new CmTDLink<T>(obj);
  if (!newlink) return false;

  if (!_first)
    _first = _last = newlink;
  else
  {
    _last->_next   = newlink;
    newlink->_prev = _last;
    _last          = newlink;
  }
  _size++;
  return true;
}


// "prepend" adds the specified item to the beginning of this list.
//
template <class T> bool CmTDLinkedList<T>::prepend(const T& obj)
{
  CmTDLink<T> *newlink = new CmTDLink<T>(obj);
  if (!newlink) return false;

  if (!_first)
    _first = _last = newlink;
  else
  {
    newlink->_next = _first;
    _first->_prev  = newlink;
    _first         = newlink;
  }
  _size++;
  return true;
}


// "insertAt" inserts the specified item at the specified index.
//
template <class T> bool CmTDLinkedList<T>::insertAt(int idx, const T& rObj)
{
  if (idx < 0 || idx >= _size) return false;

  int          ii    = 0;
  CmTDLink<T> *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  if (!rover) return false;

  if (rover == _first) return prepend(rObj);

  CmTDLink<T> *newlink = new CmTDLink<T>(rObj);
  newlink->_prev      = rover->_prev;
  rover->_prev->_next = newlink;
  newlink->_next      = rover;
  rover->_prev        = newlink;
  _size++;
  return true;
}


// "replaceAt" replaces the item at the specified index with the
// specified item.
//
template <class T> bool CmTDLinkedList<T>::replaceAt(int idx, const T& rObj)
{
  if (idx < 0 || idx >= _size) return false;

  int         ii    = 0;
  CmTDLink<T> *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  if (!rover) return false;
  rover->_data = rObj;
  return true;
}


// "remove" removes the item equal to the specified item from the list.
//
template <class T> bool CmTDLinkedList<T>::remove(const T& rObj)
{
  CmTDLink<T> *rover = _first;
  int          ii = 0, idx = -1;
  while (idx < 0 && rover)
  {
    if (rover->_data == rObj) idx = ii;
    rover = rover->_next; ii++;
  }
  return (idx >= 0) ? removeAt(idx) : false;
}


// "remove" removes the item equal to the item pointed to by the
// specified iterator from the list.
//
template <class T> bool CmTDLinkedList<T>::remove(const CmTIterator<T>& iter)
{
  return remove(iter.current());
}


// "removeAt" removes the item at the specified list index.
//
template <class T> bool CmTDLinkedList<T>::removeAt(int idx)
{
  if (idx < 0 || idx >= _size) return false;

  if (idx == 0)                                   // Item to remove is
  {                                               // first in the list.
    removeFirst();
    return true;
  }

  if (idx == _size - 1)                           // Item to remove is
  {                                               // last in the list.
    removeLast();
    return true;
  }

  int          ii    = 0;                         // Position to item.
  CmTDLink<T> *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }

  if (rover == NULL) return false;                // Item was not found.
  rover->_prev->_next = rover->_next;             // Remove item.
  rover->_next->_prev = rover->_prev;

  delete rover;                                   // Delete the link.
  _size--;
  return true;
}


// "removeAll" clears the contents of this list.
//
template <class T> void CmTDLinkedList<T>::removeAll()
{
  CmTDLink<T> *temp, *rover = _first;
  while (rover)
  {
    temp = rover->_next;
    delete rover;
    rover = temp;
  }
  _first = _last = NULL;
  _size  = 0;
}


// "removeFirst" removes and returns the first item in the list.
//
template <class T> T CmTDLinkedList<T>::removeFirst()
{
  if (!_first) return _error;
  T            ret  = _first->_data;
  CmTDLink<T> *temp = _first;
  _first = _first->_next;
  if (!_first) _last = NULL;
  else _first->_prev = NULL;
  delete temp;
  _size--;
  return ret;
}


// "removeLast" removes and returns the last item in the list.
//
template <class T> T CmTDLinkedList<T>::removeLast()
{
  if (!_last) return _error;
  T            ret  = _last->_data;
  CmTDLink<T> *temp = _last;
  _last = _last->_prev;
  if (!_last) _first = NULL;
  else _last->_next  = NULL;
  delete temp;
  _size--;
  return ret;
}


// "insertAfter" searches for the first item equal to the specified
// item in the list and inserts the second item immediately after.
//
template <class T>
bool CmTDLinkedList<T>::insertAfter(const T& after, const T& rObj)
{
  if (!_first) return false;

  CmTDLink<T> *rover = _first;
  bool         found = false;
  while (rover && !found)
  {
    if (rover->_data == after) found = true;
    else                       rover = rover->_next;
  }
  if (!found) return false;

  CmTDLink<T> *newlink = new CmTDLink<T>(rObj);
  if (!newlink) return false;

  if (rover->_next) rover->_next->_prev = newlink;
  else _last = newlink;
  newlink->_next = rover->_next;
  newlink->_prev = rover;
  rover->_next   = newlink;
  _size++;
  return true;
}


// "insertAfter" searches for the first item equal to the item
// pointed to by the specified iterator and inserts the specified
// item immediately after.
//
template <class T>
bool CmTDLinkedList<T>::insertAfter(const CmTIterator<T>& iter, const T& rObj)
{
  return insertAfter(iter.current(), rObj);
}


// "insertBefore" searches for the first item equal to the specified
// item in the list and inserts the second item immediately before.
//
template <class T>
bool CmTDLinkedList<T>::insertBefore(const T& before, const T& rObj)
{
  if (!_first) return false;

  CmTDLink<T> *rover = _first;
  bool         found = false;
  while (rover && !found)
  {
    if (rover->_data == before) found = true;
    else                        rover = rover->_next;
  }
  if (!found) return false;

  CmTDLink<T> *newlink = new CmTDLink<T>(rObj);
  if (!newlink) return false;

  if (rover->_prev) rover->_prev->_next = newlink;
  else _first = newlink;
  newlink->_prev = rover->_prev;
  newlink->_next = rover;
  rover->_prev   = newlink;
  _size++;
  return true;
}


// "insertBefore" searches for the first item equal to the item
// pointed to by the specified iterator and inserts the specified
// item immediately before.
//
template <class T>
bool CmTDLinkedList<T>::insertBefore(const CmTIterator<T>& iter, const T& rObj)
{
  return insertBefore(iter.current(), rObj);
}


// "first" returns a pointer to the first item in the list.
//
template <class T> const T& CmTDLinkedList<T>::first() const
{
  return (_first) ? _first->_data : _error;
}


// "last" returns a pointer to the last item in the list.
//
template <class T> const T& CmTDLinkedList<T>::last() const
{
  return (_last ) ? _last->_data  : _error;
}


// "index" returns the index of the specified item in the list.
//
template <class T> int CmTDLinkedList<T>::index(const T& rObj) const
{
  CmTDLink<T> *rover  = _first;
  int         outidx = -1;
  int         idx    = 0;
  while (outidx < 0 && rover)
  {
    if (rover->_data == rObj) outidx = idx;
    rover = rover->_next;
    idx++;
  }
  return outidx;
}


// "index" returns the index of the item pointed to by the specified
// iterator.
//
template <class T>
int CmTDLinkedList<T>::index(const CmTIterator<T>& iter) const
{
  return index(iter.current());
}


// "at" returns the item at the specified list index.
//
template <class T> const T& CmTDLinkedList<T>::at(int idx) const
{
  if (idx < 0 || idx >= _size) return _error;

  int         ii    = 0;
  CmTDLink<T> *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  return (rover) ? rover->_data : _error;
}


// "lookup" searches for the item equal to the specified item in the
// list and returns the first occurrence.
//
template <class T> const T& CmTDLinkedList<T>::lookup(const T& rObj) const
{
  CmTDLink<T> *rover = _first;
  bool        found = false;
  while (rover && !found)
  {
    if (rover->_data == rObj) found = true;
    else                      rover = rover->_next;
  }
  return (found) ? rover->_data : _error;
}


// "contains" checks to see if the item equal to the specified item
// is contained in this list.
//
template <class T> bool CmTDLinkedList<T>::contains(const T& rObj) const
{
  CmTDLink<T> *rover = _first;
  bool        found = false;
  while (rover && !found)
  {
    if (rover->_data == rObj) found = true;
    else                      rover = rover->_next;
  }
  return found;
}


// "occurrences" counts the number of occurrences of the specified item.
//
template <class T> unsigned CmTDLinkedList<T>::occurrences(const T& rObj) const
{
  CmTDLink<T> *rover = _first;
  unsigned    occur = 0;
  while (rover)
  {
    if (rover->_data == rObj) occur++;
    rover = rover->_next;
  }
  return occur;
}


// "newIterator" creates and returns a new list iterator.
//
template <class T> CmTIterator<T>* CmTDLinkedList<T>::newIterator() const
{
  return new CmTDLinkedListIterator<T>(*this);
}


// "CmTDLinkedListIterator" is the iterator constructor.
//
template <class T>
CmTDLinkedListIterator<T>::CmTDLinkedListIterator(const CmTDLinkedList<T>& L)
                        : CmTIterator<T>(L), _link(L._first)
{}


// "list" returns a reference to the list being iterated.
//
template <class T>
const CmTDLinkedList<T>& CmTDLinkedListIterator<T>::list() const
{
  return (const CmTDLinkedList<T>&) _container;
}


// "done" returns true if the iterator can advance no further.
//
template <class T> bool CmTDLinkedListIterator<T>::done() const
{
  return (!_link);
}


// "next" returns the current list item and increments the iterator to
// the next item in the list.
//
template <class T> const T& CmTDLinkedListIterator<T>::next()
{
  if (!_link) return list()._error;
  CmTDLink<T> *temp = _link;
  _link = _link->_next;
  return temp->_data;
}


// "previous" backs the iterator up one item and returns that item.
//
template <class T> const T& CmTDLinkedListIterator<T>::previous()
{
  if (!_link) return list()._error;
  CmTDLink<T> *temp = _link;
  _link = _link->_prev;
  return temp->_data;
}


// "current" returns the current item in the list.
//
template <class T> const T& CmTDLinkedListIterator<T>::current() const
{
  return (_link) ? _link->_data : list()._error;
}


// "first" moves the iterator to the first item.
//
template <class T> void CmTDLinkedListIterator<T>::first()
{
  _link = list()._first;
}


// "last" moves the iterator to the last item.
//
template <class T> void CmTDLinkedListIterator<T>::last()
{
  _link = list()._last;
}
