// cmthash.cc
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Hash table template implementation.
// -----------------------------------------------------------------


// "CmTHashTable" is the hash table constructor.
//
template <class T> CmTHashTable<T>::CmTHashTable(unsigned sz)
                                  : _hFunc(NULL), _total(0)
{
  _buckets = (sz > 0) ? new CmTHashBucket<T>[sz] : NULL;
  _size    = (_buckets) ? sz : 0;
}


// "CmTHashTable" constructs a hash table that will use the specified
// function for hashing.
//
template <class T>
CmTHashTable<T>::CmTHashTable(unsigned (*func)(const T&, unsigned), unsigned sz)
               : _hFunc(func), _total(0)
{
  _buckets = (sz > 0) ? new CmTHashBucket<T>[sz] : NULL;
  _size    = (_buckets) ? sz : 0;
}


// "CmTHashTable" is the hash table copy constructor.
//
template <class T>
CmTHashTable<T>::CmTHashTable(const CmTHashTable<T>& table)
               : CmTContainer<T>(table), _hFunc(table._hFunc), _total(0)
{
  _buckets = (table._size > 0) ? new CmTHashBucket<T>[table._size] : NULL;
  _size    = (_buckets) ? table._size : 0;
  copy(table);
}


// "CmTHashTable" constructs a table from the specified container.
//
template <class T>
CmTHashTable<T>::CmTHashTable(unsigned (*func)(const T&, unsigned),
                              const CmTContainer<T>& cont)
               : _hFunc(func), _total(0)
{
  _buckets = (cont.total() > 0) ? new CmTHashBucket<T>[cont.total()] : NULL;
  _size    = (_buckets) ? cont.total() : 0;
  copy(cont);
}


// "CmTHashTable" is the hash table destructor.
//
template <class T> CmTHashTable<T>::~CmTHashTable()
{
  removeAll();
  delete[] _buckets;
}


// "=" assignment operator copies the specified hash table into this table.
//
template <class T>
CmTHashTable<T>& CmTHashTable<T>::operator=(const CmTHashTable<T>& table)
{
  if (&table != this)
  {
    removeAll();
    delete[] _buckets;
    _hFunc   = table._hFunc;
    _total   = 0;
    _buckets = new CmTHashBucket<T>[table._size];
    _size    = (_buckets != NULL) ? table._size : 0;
    copy(table);
  }
  return *this;
}


// "setHashFunction" sets the hash function to be used on the input items.
//
template <class T>
void CmTHashTable<T>::setHashFunction(unsigned (*func)(const T&, unsigned))
{
  _hFunc = func;
}


// "total" returns the number of items in the table.
//
template <class T> unsigned CmTHashTable<T>::total() const
{
  return _total;
}


// "resize" resizes the hash table.
//
template <class T> bool CmTHashTable<T>::resize(unsigned newsize)
{
  if (newsize <= 0) newsize = 1;

  unsigned          newtotal   = 0;
  CmTHashBucket<T> *newbuckets = new CmTHashBucket<T>[newsize];
  if (!newbuckets) return false;

  CmTHashTableIterator<T> iter(*this);
  while (!iter.done())
  {
    T   item = iter.next();
    int idx  = (*_hFunc)(item, newsize);
    newbuckets[idx].append(item);
    newtotal++;
  }

  delete[] _buckets;
  _size    = newsize;
  _buckets = newbuckets;
  _total   = newtotal;
  return true;
}

// "isEmpty" checks whether or not the table is empty.
//
template <class T> bool CmTHashTable<T>::isEmpty() const
{
  return (_total == 0) ? true : false;
}


// "add" adds a new item to the table.
//
template <class T> bool CmTHashTable<T>::add(const T& rObj)
{
  if (!_hFunc) return false;
  unsigned idx = (*_hFunc)(rObj, _size);
  bool success = _buckets[idx].append(rObj);
  if (success) _total++;
  return success;
}


// "remove" removes the first occurrence of an item equal to the specified
// item from the table.
//
template <class T> bool CmTHashTable<T>::remove(const T& rObj)
{
  if (!_hFunc) return false;
  unsigned idx = (*_hFunc)(rObj, _size);
  bool success = _buckets[idx].remove(rObj);
  if (success) _total--;
  return success;
}


// "remove" removes the item pointed to by the iterator.
//
template <class T> bool CmTHashTable<T>::remove(const CmTIterator<T>& iter)
{
  return remove(iter.current());
}


// "removeAt" removes the item at the specified index.
//
template <class T> bool CmTHashTable<T>::removeAt(int idx)
{
  if (_total == 0) return false;

  int             bucket = 0;
  int             bidx   = 0;
  CmTHashNode<T> *node   = NULL;
  while (_buckets[bucket]._size == 0 && bucket < _size) bucket++;
  if (bucket < _size) node = _buckets[bucket]._first;

  int ix = 0;
  while (node && ix < idx)
  {
    if (node->_next)
    {
      node = node->_next;
      bidx++;
    }
    else
    {
      bucket++;
      bidx = 0;
      while (bucket < _size && _buckets[bucket]._size == 0) bucket++;
      node = (bucket < _size) ? _buckets[bucket]._first : NULL;
    }
    ix++;
  }
  return (ix == idx) ? _buckets[bucket].removeAt(bidx) : false;
}


// "removeAll" removes all items from the table.
//
template <class T> void CmTHashTable<T>::removeAll()
{
  for (int ii = 0; ii < _size; ii++)
    _buckets[ii].removeAll();
  _total = 0;
}


// "lookup" returns the first occurrence of an item equal to the specified
// item.
//
template <class T> const T& CmTHashTable<T>::lookup(const T& rObj) const
{
  if (!_hFunc) return _error;
  unsigned idx = (*_hFunc)(rObj, _size);
  T* out = _buckets[idx].lookup(rObj);
  return (out) ? *out : _error;
}


// "contains" checks if an item equal to the specified item exists
// in the table.
//
template <class T> bool CmTHashTable<T>::contains(const T& rObj) const
{
  if (!_hFunc) return false;
  unsigned idx = (*_hFunc)(rObj, _size);
  return _buckets[idx].contains(rObj);
}


// "occurrences" counts the number of items in the table equal to the
// specified item.
//
template <class T> unsigned CmTHashTable<T>::occurrences(const T& rObj) const
{
  CmTHashTableIterator<T> iter(*this);
  unsigned                num = 0;
  while (!iter.done())
    if (iter.next() == rObj) num++;
  return num;
}


// "newIterator" creates and returns a new table iterator.
//
template <class T> CmTIterator<T>* CmTHashTable<T>::newIterator() const
{
  return new CmTHashTableIterator<T>(*this);
}


// "append" appends a new item to the hash bucket.
//
template <class T> bool CmTHashBucket<T>::append(const T& rObj)
{
  CmTHashNode<T> *newnode = new CmTHashNode<T>(rObj);
  if (!newnode) return false;

  if (!_first)
    _first = _last = newnode;
  else
  {
    _last->_next = newnode;
    _last        = newnode;
  }
  _size++;
  return true;
}


// "contains" checks if an item equal to the specified item exists
// in this hash bucket.
//
template <class T> bool CmTHashBucket<T>::contains(const T& rObj) const
{
  CmTHashNode<T> *rover = _first;
  bool            found = false;
  while (rover && !found)
  {
    if (rover->_data == rObj) found = true;
    else                      rover = rover->_next;
  }
  return found;
}


// "remove" removes the first occurrence of an item equal to the
// specified item from this hash bucket.
//
template <class T> bool CmTHashBucket<T>::remove(const T& rObj)
{
  CmTHashNode<T> *rover = _first;
  int             ii = 0, idx = -1;
  while (idx < 0 && rover)
  {
    if (rover->_data == rObj) idx = ii;
    rover = rover->_next; ii++;
  }
  return (idx >= 0) ? removeAt(idx) : false;
}


// "removeAt" removes the item at the specified index.
//
template <class T> bool CmTHashBucket<T>::removeAt(int idx)
{
  if (idx < 0 || idx >= _size) return false;

  if (idx == 0)                             // Object is first in the list.
  {
    CmTHashNode<T> *temp = _first;
    if (_first == _last) _first = _last = NULL;
    else                 _first = _first->_next;
    _size--;
    delete temp;
    return true;
  }

  int             ii     = 0;
  CmTHashNode<T> *rover1 = _first;          // Search for object in list.
  CmTHashNode<T> *rover2 = _last;

  while (rover1 != NULL && ii < idx)
  {
    rover2 = rover1;
    rover1 = rover1->_next;
    ii++;
  }

  if (rover1 == NULL) return false;         // Object was not found.

  if (rover1 == _last)                      // Object is last in list.
  {
    _last         = rover2;
    rover2->_next = NULL;
  }

  else                                      // Remove object from list.
    rover2->_next = rover2->_next->_next;

  _size--;
  delete rover1;                            // Delete the hash node.
  return true;
}


// "lookup" returns the first occurrence of an item equal to the
// specified item.
//
template <class T> T* CmTHashBucket<T>::lookup(const T& rObj) const
{
  CmTHashNode<T> *rover = _first;
  CmTHashNode<T> *temp  = NULL;
  while (rover && !temp)
  {
    if (rover->_data == rObj)
      temp = rover;
    else
      rover = rover->_next;
  }
  return (temp) ? &(temp->_data) : NULL;
}


// "removeAll" removes all items from this hash bucket.
//
template <class T> void CmTHashBucket<T>::removeAll()
{
  CmTHashNode<T> *temp, *rover = _first;
  while (rover)
  {
    temp = rover->_next;
    delete rover;
    rover = temp;
  }
  _size  = 0;
  _first = _last = NULL;
}


// "CmTHashTableIterator" is the iterator constructor.  The node
// pointer is positioned at the first bucket containing items.
//
template <class T>
CmTHashTableIterator<T>::CmTHashTableIterator(const CmTHashTable<T>& Tb)
                       : CmTIterator<T>(Tb), _bucket(0), _node(NULL)
{
  while (table()._buckets[_bucket]._size == 0 && _bucket < table()._size)
    _bucket++;
  if (_bucket < table()._size) _node = table()._buckets[_bucket]._first;
}


// "table" returns a reference to the table being iterated.
//
template <class T>
const CmTHashTable<T>& CmTHashTableIterator<T>::table() const
{
  return (const CmTHashTable<T>&) _container;
}


// "done" checks if the iterator can advance any further.
//
template <class T> bool CmTHashTableIterator<T>::done() const
{
  return (!_node);
}


// "next" returns the current item and advances the iterator to the
// next item.
//
template <class T> const T& CmTHashTableIterator<T>::next()
{
  if (!_node) return table()._error;

  const T& rObj = _node->_data;         // Save current item reference.

  if (_node->_next)                     // Advance to next node.
    _node = _node->_next;
  else                                  // Advance to next bucket with items.
  {
    _bucket++;
    while (_bucket < table()._size && table()._buckets[_bucket]._size == 0)
      _bucket++;
    if (_bucket < table()._size) _node = table()._buckets[_bucket]._first;
    else                         _node = NULL;
  }
  return rObj;
}


// "previous" decrements the iterator by one object and then returns
// that object.
//
template <class T> const T& CmTHashTableIterator<T>::previous()
{
  if (!_node) return table()._error;
  const T& rval = _node->_data;

  if (_node == table()._buckets[_bucket]._first)
  {
    --_bucket;
    while (_bucket >= 0 && table()._buckets[_bucket]._size == 0)
      _bucket--;
    if (_bucket >= 0) _node = table()._buckets[_bucket]._last;
    else              _node = NULL;
  }
  else
  {
    CmTHashNode<T> *rover = table()._buckets[_bucket]._first;
    while (rover && rover->_next != _node)
      rover = rover->_next;
    _node = rover;
  }
  return rval;
}


// "current" returns the current object.
//
template <class T> const T& CmTHashTableIterator<T>::current() const
{
  return (_node) ? _node->_data : table()._error;
}


// "first" moves the iterator to the first item.
//
template <class T> void CmTHashTableIterator<T>::first()
{
  _bucket = 0;
  _node   = NULL;
  while (table()._buckets[_bucket]._size == 0 && _bucket < table()._size)
    _bucket++;
  if (_bucket < table()._size) _node = table()._buckets[_bucket]._first;
}


// "last" moves the iterator to the last item.
//
template <class T> void CmTHashTableIterator<T>::last()
{
  _bucket = table()._size - 1;
  _node   = NULL;
  while (_bucket >= 0 && table()._buckets[_bucket]._size == 0)
    _bucket--;
  if(_bucket >= 0) _node = table()._buckets[_bucket]._last;
}
