// cmthash.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Hash table template definition.
// -----------------------------------------------------------------

#ifndef _CMTHASH_H
#define _CMTHASH_H

#include <cm/include/cmtcont.h>

template <class T> class CmTHashBucket;         // Bucket class stub.
template <class T> class CmTHashNode;           // Node class stub.
template <class T> class CmTHashTableIterator;  // Iterator class stub.

template <class T> class CmTHashTable : public CmTContainer<T> {
public:
  typedef unsigned (*CmTHashFunc)(const T&, unsigned); // Hash function type.

  CmTHashTable(unsigned = 10);                  // Default table constructor.
  CmTHashTable(CmTHashFunc, unsigned = 10);     // Construct with hash func.
  CmTHashTable(const CmTHashTable<T>&);         // Table copy constructor.
  CmTHashTable(CmTHashFunc, const CmTContainer<T>&); // Copy from container.
 ~CmTHashTable();                               // Table destructor.

  CmTHashTable<T>& operator=(const CmTHashTable<T>&);  // Assignment.

  void setHashFunction(CmTHashFunc);            // Set item hashing function.

  unsigned total  () const;                     // Return number of items.
  bool     resize (unsigned);                   // Resize the table.
  bool     isEmpty() const;                     // Is table empty.

  bool add      (const T&);                     // Add item to table.
  bool remove   (const T&);                     // Remove equal item.
  bool remove   (const CmTIterator<T>&);        // Remove iterator item.
  bool removeAt (int);                          // Remove at index.
  void removeAll();                             // Remove all items.

  const T& lookup     (const T&) const;         // Return equal item from table.
  bool     contains   (const T&) const;         // See if equal item in table.
  unsigned occurrences(const T&) const;         // How many occurrences of item.

  CmTIterator<T>* newIterator() const;          // Create and return iterator.

private:
  CmTHashFunc       _hFunc;                     // Hash function pointer.
  unsigned          _total;                     // Number of items in table.
  CmTHashBucket<T> *_buckets;                   // Hash bucket list.

  friend CmTHashTableIterator<T>;               // Iterator can access.
};

template <class T> class CmTHashBucket {        // Bucket definition.
private:
  CmTHashBucket() : _size(0), _first(NULL), _last(NULL) {}  // Constructor.
 ~CmTHashBucket() { removeAll(); }              // Bucket destructor.

  unsigned size() const { return _size; }       // Return bucket size.

  bool append   (const T&);                     // Append item to bucket.
  bool contains (const T&) const;               // Is item in bucket?
  bool remove   (const T&);                     // Remove item from bucket.
  bool removeAt (int);                          // Remove at index.
  T*   lookup   (const T&) const;               // Return equal item.
  void removeAll();                             // Remove all items.

  unsigned        _size;                        // Bucket size.
  CmTHashNode<T> *_first;                       // First node in bucket.
  CmTHashNode<T> *_last;                        // Last node in bucket.

  friend CmTHashTable<T>;                       // Table can access.
  friend CmTHashTableIterator<T>;               // Iterator can access.
};

template <class T> class CmTHashNode {          // Node definition.
private:
  CmTHashNode(const T& O) : _next(NULL), _data(O) {}  // Constructor.

  CmTHashNode<T> *_next;                        // Next node in bucket.
  T               _data;                        // Item.

  friend CmTHashTable<T>;                       // Table can access.
  friend CmTHashBucket<T>;                      // Bucket can access.
  friend CmTHashTableIterator<T>;               // Iterator can access.
};

template <class T> class CmTHashTableIterator : public CmTIterator<T> {
public:
  CmTHashTableIterator(const CmTHashTable<T>&); // Iterator constructor.

  bool     done    () const;                    // Check if done iterating.
  const T& next    ();                          // Return and advance.
  const T& previous();                          // Return and backup.
  const T& current () const;                    // Return current item.
  void     first   ();                          // Move to first item.
  void     last    ();                          // Move to last item.

  const CmTHashTable<T>& table() const;         // Return table reference.

private:
  int             _bucket;                      // Current hash bucket.
  CmTHashNode<T> *_node;                        // Current bucket node.
};

#if defined(__TURBOC__) || defined(__xlC__) || defined(_MSC_VER) || defined(__GNUG__)
#include <cm/include/cmthash.cc>
#endif

#endif
