// cmtime.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Time class definition.
// -----------------------------------------------------------------

#ifndef _CMTIME_H
#define _CMTIME_H

#include <cm/include/cmdate.h>

class CmTime : public CmObject {              // Time class definition.
public:
  CmTime();                                   // Default time constructor.
  CmTime(unsigned long s) : _sec(s) {}        // Construct from seconds.
  CmTime(unsigned, unsigned, unsigned = 0);   // Construct from h, m, s.
  CmTime(const CmDate&, unsigned = 0, unsigned = 0, unsigned = 0);

  bool     isDST      () const;               // Check if daylight savings.
  CmTime   localTime  () const;               // Get local time.
  unsigned hour       () const;               // Get hour number.
  unsigned hourGMT    () const;               // Get hour number (Greenwich).
  unsigned minute     () const;               // Get minute number.
  unsigned minuteGMT  () const;               // Get minute num  (Greenwich).
  unsigned second     () const;               // Get second number.
  void     makeCurrent();                     // Make this current time.

  unsigned long seconds() const;              // Returns seconds.

  operator CmDate() const;                    // Make date from time.

  bool operator< (const CmTime&) const;       // Check if time <  this.
  bool operator<=(const CmTime&) const;       // Check if time <= this.
  bool operator> (const CmTime&) const;       // Check if time >  this.
  bool operator>=(const CmTime&) const;       // Check if time >= this.
  bool operator==(const CmTime&) const;       // Check if time == this.
  bool operator!=(const CmTime&) const;       // Check if time != this.

  bool isEqual(const CmObject*) const;        // Compare times.
  int  compare(const CmObject*) const;        // Compare times.

  unsigned hash(unsigned) const;              // Hash time.

  void printOn(ostream&) const;               // Print time to stream.

  bool write(CmIOFile&) const;                // Write time to file.
  bool read (CmIOFile&);                      // Read time from file.

  CMOBJECT_DEFINE(CmTime)                     // Define object funcs.

  static CmTime beginDST (unsigned);          // Get start of DST for year.
  static CmTime endDST   (unsigned);          // Get end of DST for year.
  static CmTime localTime(const CmDate&, unsigned = 0,  // Get local time.
                           unsigned = 0, unsigned = 0);

private:
  unsigned long _sec;                         // Seconds.
  static   int  _displayStyle;                // Display style.
};

class CmTimeFormat {                          // Time format definition.
private:
  friend ostream& time12Hour(ostream&);       // 12 hour clock manipulator.
  friend ostream& time24Hour(ostream&);       // 24 hour clock manipulator.
  friend ostream& timeDate  (ostream&);       // Date with time manipulator.
  friend ostream& timeNoDate(ostream&);       // No date manipulator.
  static int      _hourIndex;                 // 12/24 hour index from ios.
  static int      _dateIndex;                 // Date/nodate index from ios.
  friend CmTime;                              // Time class can access.
};

// "seconds" returns the seconds value.
inline unsigned long CmTime::seconds() const
{ return _sec; }

// "<" checks if the input is less than this time.
inline bool CmTime::operator<(const CmTime& T) const
{ return (_sec < T._sec); }

// "<=" checks if the input is less than or equal to this time.
inline bool CmTime::operator<=(const CmTime& T) const
{ return (_sec <= T._sec); }

// ">" checks if the input is greater than this time.
inline bool CmTime::operator>(const CmTime& T) const
{ return (_sec > T._sec); }

// ">=" checks if the input is greater than or equal to this time.
inline bool CmTime::operator>=(const CmTime& T) const
{ return (_sec >= T._sec); }

// "==" checks if the input is equal to this time.
inline bool CmTime::operator==(const CmTime& T) const
{ return (_sec == T._sec); }

// "!=" checks if the input is not equal to this time.
inline bool CmTime::operator!=(const CmTime& T) const
{ return (_sec != T._sec); }

#endif
