// cmtlist.cc
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Singly linked list template implementation.
// -----------------------------------------------------------------


// "CmTLinkedList" is the list copy constructor.
//
template <class T>
CmTLinkedList<T>::CmTLinkedList(const CmTLinkedList<T>& list)
                : CmTContainer<T>(list), _first(NULL), _last(NULL)
{
  copy(list);
}


// "CmTLinkedList" constructs a list from the specified container.
//
template <class T>
CmTLinkedList<T>::CmTLinkedList(const CmTContainer<T>& cont)
                : _first(NULL), _last(NULL)
{
  copy(cont);
}


// "~CmTLinkedList" is the list destructor.
//
template <class T> CmTLinkedList<T>::~CmTLinkedList()
{
  removeAll();
}


// "=" operator copies the contents of the input list into this list.
//
template <class T>
CmTLinkedList<T>& CmTLinkedList<T>::operator=(const CmTLinkedList<T>& list)
{
  if (&list != this)
  {
    removeAll();
    copy     (list);
  }
  return *this;
}


// "add" adds the specified item to the end of this list.
//
template <class T> bool CmTLinkedList<T>::add(const T& rObj)
{
  return append(rObj);
}


// "append" adds the specified item to the end of this list.
//
template <class T> bool CmTLinkedList<T>::append(const T& obj)
{
  CmTLink<T> *newlink = new CmTLink<T>(obj);
  if (!newlink) return false;

  if (!_first)
    _first = _last = newlink;
  else
  {
    _last->_next = newlink;
    _last        = newlink;
  }
  _size++;
  return true;
}


// "prepend" adds the specified item to the beginning of this list.
//
template <class T> bool CmTLinkedList<T>::prepend(const T& obj)
{
  CmTLink<T> *newlink = new CmTLink<T>(obj);
  if (!newlink) return false;

  if (!_first)
    _first = _last = newlink;
  else
  {
    newlink->_next = _first;
    _first         = newlink;
  }
  _size++;
  return true;
}


// "insertAt" inserts the specified item at the specified index.
//
template <class T> bool CmTLinkedList<T>::insertAt(int idx, const T& rObj)
{
  if (idx < 0 || idx >= _size) return false;

  int         ii    = 0;
  CmTLink<T> *rover1 = _first;
  CmTLink<T> *rover2 = NULL;
  while (rover1 && ii < idx)
  {
    rover2 = rover1;
    rover1 = rover1->_next;
    ii++;
  }
  if (!rover1) return false;

  CmTLink<T> *newlink = new CmTLink<T>(rObj);
  if (rover1 == rover2)
  {
    newlink->_next = _first;
    _first         = newlink;
  }
  else
  {
    newlink->_next = rover2->_next;
    rover2->_next  = newlink;
  }
  _size++;
  return true;
}


// "replaceAt" replaces the item at the specified index with the
// specified item.
//
template <class T> bool CmTLinkedList<T>::replaceAt(int idx, const T& rObj)
{
  if (idx < 0 || idx >= _size) return false;

  int         ii    = 0;
  CmTLink<T> *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  if (!rover) return false;
  rover->_data = rObj;
  return true;
}


// "remove" removes the item equal to the specified item from the list.
//
template <class T> bool CmTLinkedList<T>::remove(const T& rObj)
{
  CmTLink<T> *rover = _first;
  int         ii = 0, idx = -1;
  while (idx < 0 && rover)
  {
    if (rover->_data == rObj) idx = ii;
    rover = rover->_next; ii++;
  }
  return (idx >= 0) ? removeAt(idx) : false;
}


// "remove" removes the item equal to the item pointed to by the
// specified iterator from the list.
//
template <class T> bool CmTLinkedList<T>::remove(const CmTIterator<T>& iter)
{
  return remove(iter.current());
}


// "removeAt" removes the item at the specified list index.
//
template <class T> bool CmTLinkedList<T>::removeAt(int idx)
{
  if (idx < 0 || idx >= _size) return false;

  if (idx == 0)                                   // Item to remove is
  {                                               // first in the list.
    removeFirst();
    return true;
  }

  if (idx == _size - 1)                           // Item to remove is
  {                                               // last in the list.
    removeLast();
    return true;
  }

  int         ii    = 0;                          // Position to item.
  CmTLink<T> *rover = _first;
  CmTLink<T> *prev  = NULL;
  while (rover && ii < idx)
  {
    prev  = rover;
    rover = rover->_next;
    ii++;
  }

  if (rover == NULL) return false;                // Item was not found.
  prev->_next = prev->_next->_next;               // Remove link from list.

  delete rover;                                   // Delete the link.
  _size--;
  return true;
}


// "removeAll" clears the contents of this list.
//
template <class T> void CmTLinkedList<T>::removeAll()
{
  CmTLink<T> *temp, *rover = _first;
  while (rover)
  {
    temp = rover->_next;
    delete rover;
    rover = temp;
  }
  _first = _last = NULL;
  _size  = 0;
}


// "removeFirst" removes and returns the first item in the list.
//
template <class T> T CmTLinkedList<T>::removeFirst()
{
  if (!_first) return _error;
  T           ret  = _first->_data;
  CmTLink<T> *temp = _first;
  _first = _first->_next;
  if (!_first) _last = NULL;
  delete temp;
  _size--;
  return ret;
}


// "removeLast" removes and returns the last item in the list.
//
template <class T> T CmTLinkedList<T>::removeLast()
{
  if (!_last) return _error;
  if (_last == _first)
  {
    T ret = _first->_data;
    delete _first;
    _first = _last = NULL;
    _size  = 0;
    return ret;
  }

  CmTLink<T> *rover = _first;
  while (rover->_next->_next)
    rover = rover->_next;
  T ret = rover->_next->_data;
  delete rover->_next;
  _last = rover;
  _last->_next = NULL;
  _size--;
  return ret;
}


// "insertAfter" searches for the first item equal to the specified
// item in the list and inserts the second item immediately after.
//
template <class T>
bool CmTLinkedList<T>::insertAfter(const T& after, const T& rObj)
{
  if (!_first) return false;

  CmTLink<T> *rover = _first;
  bool        found = false;
  while (rover && !found)
  {
    if (rover->_data == after) found = true;
    else                       rover = rover->_next;
  }

  if (!found) return false;

  CmTLink<T> *newlink = new CmTLink<T>(rObj);
  if (!newlink) return false;
  newlink->_next = rover->_next;
  rover->_next   = newlink;
  if (rover == _last) _last = newlink;
  _size++;
  return true;
}


// "insertAfter" searches for the first item equal to the item
// pointed to by the specified iterator and inserts the specified
// item immediately after.
//
template <class T>
bool CmTLinkedList<T>::insertAfter(const CmTIterator<T>& iter, const T& rObj)
{
  return insertAfter(iter.current(), rObj);
}


// "insertBefore" searches for the first item equal to the specified
// item in the list and inserts the second item immediately before.
//
template <class T>
bool CmTLinkedList<T>::insertBefore(const T& before, const T& rObj)
{
  if (!_first) return false;

  CmTLink<T> *rover1 = _first;
  CmTLink<T> *rover2 = _first;

  while (rover1 != NULL && rover1->_data != before)
  {
    rover2 = rover1;
    rover1 = rover1->_next;
  }

  if (rover1 == NULL) return false;

  CmTLink<T> *newlink = new CmTLink<T>(rObj);
  if (rover1 == rover2)
  {
    newlink->_next = _first;
    _first         = newlink;
  }
  else
  {
    newlink->_next = rover2->_next;
    rover2->_next  = newlink;
  }
  _size++;
  return true;
}


// "insertBefore" searches for the first item equal to the item
// pointed to by the specified iterator and inserts the specified
// item immediately before.
//
template <class T>
bool CmTLinkedList<T>::insertBefore(const CmTIterator<T>& iter, const T& rObj)
{
  return insertBefore(iter.current(), rObj);
}


// "first" returns a pointer to the first item in the list.
//
template <class T> const T& CmTLinkedList<T>::first() const
{
  return (_first) ? _first->_data : _error;
}


// "last" returns a pointer to the last item in the list.
//
template <class T> const T& CmTLinkedList<T>::last() const
{
  return (_last ) ? _last->_data  : _error;
}


// "index" returns the index of the specified item in the list.
//
template <class T> int CmTLinkedList<T>::index(const T& rObj) const
{
  CmTLink<T> *rover  = _first;
  int         outidx = -1;
  int         idx    = 0;
  while (outidx < 0 && rover)
  {
    if (rover->_data == rObj) outidx = idx;
    rover = rover->_next;
    idx++;
  }
  return outidx;
}


// "index" returns the index of the item pointed to by the specified
// iterator.
//
template <class T>
int CmTLinkedList<T>::index(const CmTIterator<T>& iter) const
{
  return index(iter.current());
}


// "at" returns the item at the specified list index.
//
template <class T> const T& CmTLinkedList<T>::at(int idx) const
{
  if (idx < 0 || idx >= _size) return _error;

  int         ii    = 0;
  CmTLink<T> *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  return (rover) ? rover->_data : _error;
}


// "lookup" searches for the item equal to the specified item in the
// list and returns the first occurrence.
//
template <class T> const T& CmTLinkedList<T>::lookup(const T& rObj) const
{
  CmTLink<T> *rover = _first;
  bool        found = false;
  while (rover && !found)
  {
    if (rover->_data == rObj) found = true;
    else                      rover = rover->_next;
  }
  return (found) ? rover->_data : _error;
}


// "contains" checks to see if the item equal to the specified item
// is contained in this list.
//
template <class T> bool CmTLinkedList<T>::contains(const T& rObj) const
{
  CmTLink<T> *rover = _first;
  bool        found = false;
  while (rover && !found)
  {
    if (rover->_data == rObj) found = true;
    else                      rover = rover->_next;
  }
  return found;
}


// "occurrences" counts the number of occurrences of the specified item.
//
template <class T> unsigned CmTLinkedList<T>::occurrences(const T& rObj) const
{
  CmTLink<T> *rover = _first;
  unsigned    occur = 0;
  while (rover)
  {
    if (rover->_data == rObj) occur++;
    rover = rover->_next;
  }
  return occur;
}


// "newIterator" creates and returns a new list iterator.
//
template <class T> CmTIterator<T>* CmTLinkedList<T>::newIterator() const
{
  return new CmTLinkedListIterator<T>(*this);
}


// "CmTLinkedListIterator" is the iterator constructor.
//
template <class T>
CmTLinkedListIterator<T>::CmTLinkedListIterator(const CmTLinkedList<T>& L)
                        : CmTIterator<T>(L), _link(L._first)
{}


// "list" returns a reference to the list being iterated.
//
template <class T>
const CmTLinkedList<T>& CmTLinkedListIterator<T>::list() const
{
  return (const CmTLinkedList<T>&) _container;
}


// "done" returns true if the iterator can advance no further.
//
template <class T> bool CmTLinkedListIterator<T>::done() const
{
  return (!_link);
}


// "next" returns the current list item and increments the iterator to
// the next item in the list.
//
template <class T> const T& CmTLinkedListIterator<T>::next()
{
  if (!_link) return list()._error;
  CmTLink<T> *temp = _link;
  _link = _link->_next;
  return temp->_data;
}


// "previous" backs the iterator up one item and returns that item.
//
template <class T> const T& CmTLinkedListIterator<T>::previous()
{
  if (!_link) return list()._error;
  CmTLink<T> *temp = _link;
  if (_link == list()._first)
    _link = NULL;
  else
  {
    CmTLink<T> *rover = list()._first;
    while (rover && rover->_next != _link)
      rover = rover->_next;
    _link = rover;
  }
  return temp->_data;
}


// "current" returns the current item in the list.
//
template <class T> const T& CmTLinkedListIterator<T>::current() const
{
  return (_link) ? _link->_data : list()._error;
}


// "first" moves the iterator to the first item.
//
template <class T> void CmTLinkedListIterator<T>::first()
{
  _link = list()._first;
}


// "last" moves the iterator to the last item.
//
template <class T> void CmTLinkedListIterator<T>::last()
{
  _link = list()._last;
}
