// cmtordar.cc
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Ordered array template implementation.
// -----------------------------------------------------------------


// "CmTOrderedArray" is the default array constructor.
//
template <class T>
CmTOrderedArray<T>::CmTOrderedArray(unsigned sz, unsigned dt)
                  : _delta(dt), _total(0)
{
  _entries = (sz > 0) ? new T[sz] : NULL;
  _size    = (_entries) ? sz : 0;
}


// "CmTOrderedArray" is the array copy constructor.
//
template <class T>
CmTOrderedArray<T>::CmTOrderedArray(const CmTOrderedArray<T>& arr)
                  : CmTContainer<T>(arr), _delta(arr._delta), _total(0)
{
  _entries = (arr._size > 0) ? new T[arr._size] : NULL;
  _size    = (_entries) ? arr._size : 0;
  copy(arr);
}


// "CmTOrderedArray" constructs an array from the specified container.
//
template <class T>
CmTOrderedArray<T>::CmTOrderedArray(const CmTContainer<T>& cont)
                  : _delta(0), _total(0)
{
  _entries = (cont.total() > 0) ? new T[cont.total()] : NULL;
  _size    = (_entries) ? cont.total() : 0;
  copy(cont);
}


// "=" assignment operator copies the specified array into this array.
//
template <class T> CmTOrderedArray<T>&
CmTOrderedArray<T>::operator=(const CmTOrderedArray<T>& arr)
{
  if (&arr != this)
  {
    delete[] _entries;
    _delta   = arr._delta;
    _total   = 0;
    _entries = (arr._size > 0) ? new T[arr._size] : NULL;
    _size    = (_entries) ? arr._size : 0;
    copy(arr);
  }
  return *this;
}


// "delta" sets a new delta value for automatic growing.
//
template <class T> void CmTOrderedArray<T>::delta(unsigned dt)
{
  _delta = dt;
}


// "delta" returns the current delta value.
//
template <class T> unsigned CmTOrderedArray<T>::delta() const
{
  return _delta;
}


// "total" returns the number of items in the array.
//
template <class T> unsigned CmTOrderedArray<T>::total() const
{
  return _total;
}


// "resize" resizes the storage allocated for this array.
//
template <class T> bool CmTOrderedArray<T>::resize(unsigned newsize)
{
  T* newentries = NULL;
  if (newsize == 0)
    removeAll();

  else
  {
    newentries = new T[newsize];
    if (!newentries) return false;

    if (newsize < _total) _total = newsize;

    for (int ii = 0; ii < _total; ii++)
      newentries[ii] = _entries[ii];
  }

  delete[] _entries;
  _entries = newentries;
  _size    = newsize;
  return true;
}


// "isEmpty" returns whether or not this array has any items in it.
//
template <class T> bool CmTOrderedArray<T>::isEmpty() const
{
  return (_total == 0) ? true : false;
}


// "add" appends the specified item to this array.
//
template <class T> bool CmTOrderedArray<T>::add(const T& rObj)
{
  if (_total >= _size)
    if (_delta == 0 || !resize(_size + _delta))
      return false;

  int idx = shouldGo(rObj);
  if (idx == -1) return false;

  if (idx != _total)
  {
    for (int ii = _total; ii > idx; ii--)
      _entries[ii] = _entries[ii-1];
  }

  _entries[idx] = rObj;
  _total++;
  return true;
}


// "remove" removes the first occurrence of an item equal to the specified
// item in the array.
//
template <class T> bool CmTOrderedArray<T>::remove(const T& rObj)
{
  if (_total == 0) return false;
  int  ii    = 0;
  bool found = false;
  while (ii < _total && !found)
    if (_entries[ii++] == rObj) found = true;
  return (found) ? removeAt(ii-1) : false;
}


// "remove" removes the item pointed to by the specified iterator.
//
template <class T> bool CmTOrderedArray<T>::remove(const CmTIterator<T>& iter)
{
  return remove(iter.current());
}


// "removeAt" removes the item at the specified index.
//
template <class T> bool CmTOrderedArray<T>::removeAt(int idx)
{
  if (idx < 0 || idx >= _total) return false;

  for (int ii = idx; ii < _total-1; ii++)
    _entries[ii] = _entries[ii+1];
  _total--;
  return true;
}


// "removeAll" removes all items from the array.
//
template <class T> void CmTOrderedArray<T>::removeAll()
{
  _total = 0;
}


// "index" returns the index of the first occurrence of an item equal
// to the specified item.
//
template <class T> int CmTOrderedArray<T>::index(const T& rObj) const
{
  int ii = 0, idx = -1;
  while (ii < _total && idx == -1)
    if (_entries[ii++] == rObj) idx = ii - 1;
  return idx;
}


// "index" returns the index of the item pointed to by the specified
// iterator.
//
template <class T>
int CmTOrderedArray<T>::index(const CmTIterator<T>& iter) const
{
  return index(iter.current());
}


// "shouldGo" reports the index where the specified object would be
// inserted if it were added to the array.
//
template <class T> int CmTOrderedArray<T>::shouldGo(const T& rObj) const
{
  if (_total == _size) return -1;

  if (_total == 0) return 0;
  if (rObj >= _entries[_total-1]) return _total;

  int idx = -1, ii = 0;
  while (ii < _total && idx == -1)
    if (rObj < _entries[ii++]) idx = ii - 1;
  return idx;
}


// "at" returns the item at the specified index.
//
template <class T> const T& CmTOrderedArray<T>::at(int idx) const
{
  return (idx >= 0 && idx < _total) ? _entries[idx] : _error;
}


// "lookup" returns the first occurrence of the item equal to the
// specified item in the array.
//
template <class T> const T& CmTOrderedArray<T>::lookup(const T& rObj) const
{
  int idx = index(rObj);
  return (idx > -1) ? _entries[idx] : _error;
}


// "contains" checks the array for an item equal to the input.
//
template <class T> bool CmTOrderedArray<T>::contains(const T& rObj) const
{
  return (index(rObj) > -1) ? true : false;
}


// "occurrences" counts the number of occurrences of items equal to the
// specified item.
//
template <class T> unsigned CmTOrderedArray<T>::occurrences(const T& rObj) const
{
  int      ii  = 0;
  unsigned num = 0;
  while (ii < _total)
    if (_entries[ii++] == rObj) num++;
  return num;
}


// "newIterator" creates and returns a new array iterator.
//
template <class T> CmTIterator<T>* CmTOrderedArray<T>::newIterator() const
{
  return new CmTOrderedArrayIterator<T>(*this);
}


// "array" returns a reference to the array being iterated.
//
template <class T>
const CmTOrderedArray<T>& CmTOrderedArrayIterator<T>::array() const
{
  return (const CmTOrderedArray<T>&) _container;
}


// "done" returns whether or not this iterator can advance any further.
//
template <class T> bool CmTOrderedArrayIterator<T>::done() const
{
  return (_index >= 0 && _index < array()._total) ? false : true;
}


// "next" returns the current item in the array and advances the
// iterator to the next item.
//
template <class T> const T& CmTOrderedArrayIterator<T>::next()
{
  return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index++] : array()._error;
}


// "previous" returns the current item in the array and decrements the
// iterator to the previous item.
//
template <class T> const T& CmTOrderedArrayIterator<T>::previous()
{
  return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index--] : array()._error;
}


// "current" returns the current array item.
//
template <class T> const T& CmTOrderedArrayIterator<T>::current() const
{
  return (_index >= 0 && _index < array()._total) ?
          array()._entries[_index] : array()._error;
}


// "first" moves the iterator to the first item.
//
template <class T> void CmTOrderedArrayIterator<T>::first()
{
  _index = 0;
}


// "last" moves the iterator to the last item.
//
template <class T> void CmTOrderedArrayIterator<T>::last()
{
  _index = array()._total - 1;
}
