// cmtvec.cc
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Dynamic array template implementation.
// -----------------------------------------------------------------


// "CmTVector" is the default vector constructor.
//
template <class T> CmTVector<T>::CmTVector(unsigned sz, const T* items)
                               : _size(0), _array(NULL)
{
  assign(items, sz);
}


// "CmTVector" constructs a new vector and fills it with the specified
// fill item.
//
template <class T> CmTVector<T>::CmTVector(unsigned sz, const T& fill)
                               : _size(0), _array(NULL)
{
  assign(fill, sz);
}


// "CmTVector" is the vector copy constructor.
//
template <class T> CmTVector<T>::CmTVector(const CmTVector<T>& vec)
                               : _size(0), _array(NULL)
{
  assign(vec._array, vec._size);
}


// "~CmTVector" is the vector destructor.
//
template <class T> CmTVector<T>::~CmTVector()
{
  clear();
}


// "=" assignment operator copies the contents from an input array into
// this array.
//
template <class T>
CmTVector<T>& CmTVector<T>::operator=(const CmTVector<T>& vec)
{
  if (&vec != this) assign(vec._array, vec._size);
  return *this;
}


// "+=" appends a copy of the specified vector onto this vector.
//
template <class T>
CmTVector<T>& CmTVector<T>::operator+=(const CmTVector<T>& vec)
{
  return append(vec._array, vec._size);
}


// "+" creates a new vector containing a copy of this vector with a
// copy of the specified vector appended on the end.
//
template <class T>
CmTVector<T> CmTVector<T>::operator+(const CmTVector<T>& vec) const
{
  CmTVector<T> out(*this);
  out.append(vec._array, vec._size);
  return out;
}


// "[]" operator returns the entry at the specified index in the array.
//
template <class T> T& CmTVector<T>::operator[](int idx)
{
  if (idx < 0) idx = 0;
  if (idx >= _size)
    if (!resize(idx + 1)) idx = _size;
  return _array[idx];
}


// "[]" operator returns the entry at the specified index in the array.
//
template <class T> const T& CmTVector<T>::operator[](int idx) const
{
  if (idx < 0) idx = 0;
  if (idx >= _size)
    if (!((CmTVector<T>*) this)->resize(idx + 1)) idx = _size;
  return _array[idx];
}


// "()" operator returns a vector containing n items from this vector
// starting at the specified index.
//
template <class T>
CmTVector<T> CmTVector<T>::operator()(int idx, unsigned n) const
{
  return subVector(idx, n);
}


// "const T*" operator returns a pointer to the array.
//
template <class T> CmTVector<T>::operator const T*() const
{
  return _array;
}


// "size" returns the vector size.
//
template <class T> unsigned CmTVector<T>::size() const
{
  return _size;
}


// "array" returns the array pointer in the vector.
//
template <class T> const T* CmTVector<T>::array() const
{
  return _array;
}


// "assign" replaces the contents of the vector with the specified
// array of items.
//
template <class T>
CmTVector<T>& CmTVector<T>::assign(const T* pItems, unsigned sz)
{
  if (!pItems || !sz || !resize(sz)) return *this;
  for (int ii = 0; ii < sz; ii++)
    _array[ii] = pItems[ii];
  return *this;
}


// "assign" resizes the vector using the specified size and fills the
// vector with copies of the specified object.
//
template <class T>
CmTVector<T>& CmTVector<T>::assign(const T& rItem, unsigned sz)
{
  if (!sz || !resize(sz)) return *this;
  for (int ii = 0; ii < sz; ii++)
    _array[ii] = rItem;
  return *this;
}


// "append" appends the specified array of items to the end of this
// vector.
//
template <class T>
CmTVector<T>& CmTVector<T>::append(const T* pItems, unsigned sz)
{
  unsigned current = size();
  if (!pItems || !sz || !resize(current + sz)) return *this;
  for (int ii = current; ii < size(); ii++)
    _array[ii] = pItems[ii-current];
  return *this;
}


// "append" increases the vector size using the specified size and
// fills the new portion of the vector with copies of the specified item.
//
template <class T>
CmTVector<T>& CmTVector<T>::append(const T& rItem, unsigned sz)
{
  unsigned current = size();
  if (!sz || !resize(current + sz)) return *this;
  for (int ii = current; ii < size(); ii++)
    _array[ii] = rItem;
  return *this;
}


// "insert" inserts the items from the specified vector into this vector at
// the specified index moving the existing items to make room.
//
template <class T>
CmTVector<T>& CmTVector<T>::insert(int idx, const CmTVector<T>& vec)
{
  return insert(idx, vec.array(), vec.size());
}


// "insert" inserts the items in the specified array into this vector at
// the specified index moving the existing items to make room.
//
template <class T>
CmTVector<T>& CmTVector<T>::insert(int idx, const T* pItems, unsigned sz)
{
  unsigned size1 = size();

  if (idx < 0 || idx > size1 || !pItems || !resize(size1+sz))
    return *this;

  for (int ii = size()-1; ii >= (idx+sz); ii--)
    _array[ii] = _array[ii-sz];
  for (ii = 0; ii < sz; ii++)
    _array[idx+ii] = pItems[ii];
  return *this;
}


// "insert" inserts copies of the specified item into this vector at the
// specified index moving the existing items to make room.
//
template <class T>
CmTVector<T>& CmTVector<T>::insert(int idx, const T& rItem, unsigned sz)
{
  int      ii;
  unsigned size1 = size();

  if (idx < 0 || idx > size1 || !resize(size1+sz))
    return *this;

  for (ii = size()-1; ii >= (idx+sz); ii--)
    _array[ii] = _array[ii-sz];
  for (ii = 0; ii < sz; ii++)
    _array[idx+ii] = rItem;
  return *this;
}


// "remove" removes the specified number of items from the vector at
// the specified index.
//
template <class T>
CmTVector<T>& CmTVector<T>::remove(int idx, unsigned nItems)
{
  unsigned size1 = size();

  if (idx < 0 || idx > size1 || nItems <= 0 || (idx+nItems) > size1)
    return *this;

  for (int ii = idx; ii < (size1-nItems); ii++)
    _array[ii] = _array[ii+nItems];
  resize(size1-nItems);
  return *this;
}


// "subVector" returns a vector containing n items from this vector
// starting at the specified index.
//
template <class T>
CmTVector<T> CmTVector<T>::subVector(int idx, unsigned n) const
{
  if (idx < 0 || (idx+n) > _size) return CmTVector<T>();
  return CmTVector<T>(n, &(_array[idx]));
}


// "clear" clears out this array resetting the size to zero.
//
template <class T> void CmTVector<T>::clear()
{
  if (_array)
  {
    delete[] _array;
    _size  = 0;
    _array = NULL;
  }
}


// "resize" resizes the array to match a specified size.
//
template <class T> bool CmTVector<T>::resize(unsigned newsize)
{
  if (newsize == 0)
    clear();
  else
  {
    T *newentries = new T[newsize];
    if (!newentries) return false;
    if (_size > 0)
    {
      unsigned limit = (_size > newsize) ? newsize : _size;
      for (int ii = 0; ii < limit; ii++)
        newentries[ii] = _array[ii];
      delete[] _array;
    }
    _size  = newsize;
    _array = newentries;
  }
  return true;
}
