// cmtypinf.h
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Type id and type info class definitions.
// -----------------------------------------------------------------

#ifndef _CMTYPINF_H
#define _CMTYPINF_H

#include <cm/include/cmdefs.h>
#include <string.h>

class CmTypeInfo;                               // Type info class stub.

class CmTypeId {                                // Type id definition.
public:
  CmTypeId(const CmTypeInfo* p) : _id(p) {}     // Type id constructor.

  const char* name() const;                     // Return type name.

  bool operator==(const CmTypeId&) const;       // See if type info is equal.

  const CmTypeInfo* typeInfo() const;           // Get type info pointer.

private:
  const CmTypeInfo* _id;                        // Type info pointer.
};

class CmTypeInfo {                              // Type info definition.
public:
  CmTypeInfo(const char*, const CmTypeInfo*[]); // Type info constructor.

  inline const char* name() const;              // Return type name.

  bool isEqual(const CmTypeInfo*) const;        // See if type info is equal.
  bool canCast(const CmTypeInfo*) const;        // Check if we can cast down.
  bool hasBase(const CmTypeInfo*) const;        // Check if type has base class.

  static const CmTypeInfo _infoObj;             // Type info's type info.

  virtual CmTypeId getTypeId() const;           // Get object type id.
  static  CmTypeId typeId   ();                 // Get class type id.

private:
  CmTypeInfo(const CmTypeInfo&) {}              // No copies of type info.
  void operator=(const CmTypeInfo&) {}          // No copies of type info.

  const char        *_name;                     // Type name.
  const CmTypeInfo **_bases;                    // List of base classes.
};

// "name" returns the type name from the type id's type info.
inline const char* CmTypeId::name() const
{ return _id->name(); }

// "typeInfo" returns a pointer to the type info object.
inline const CmTypeInfo* CmTypeId::typeInfo() const
{ return _id; }

// "==" returns true if the specified type id is from the same type.
inline bool CmTypeId::operator==(const CmTypeId& id) const
{ return (_id == id._id) ? true : false; }

// "name" returns the type name from the type info.
inline const char* CmTypeInfo::name() const
{ return _name; }

// "isEqual" returns true if the specified type info is from the same type.
inline bool CmTypeInfo::isEqual(const CmTypeInfo* pInfo) const
{ return (this == pInfo || !strcmp(_name, pInfo->_name)) ? true : false; }

// "canCast" returns true if we can cast from this to the specified type.
inline bool CmTypeInfo::canCast(const CmTypeInfo* pInfo) const
{ return (isEqual(pInfo) || pInfo->hasBase(this)) ? true : false; }

// "getTypeId" returns the type id for this object.
inline CmTypeId CmTypeInfo::getTypeId() const
{ return &_infoObj; }

// "typeId" returns the type id for this class.
inline CmTypeId CmTypeInfo::typeId()
{ return &_infoObj; }

// Macro definitions can be used to obtain type information and for pointer
// casting.
//
// CmStaticTypeId - given a type name returns it's type information.
// CmPtrTypeId    - given a pointer returns it's type information.
// CmRefTypeId    - given a reference returns it's type information.
// CmPtrCast      - given a type name and a pointer, casts the pointer to
//                  that type and returns the new pointer or NULL if the
//                  cast is not appropriate.
//
#define CmStaticTypeId(tp) tp::typeId()
#define CmPtrTypeId(p) ((p)->getTypeId())
#define CmRefTypeId(r) ((r).getTypeId())
#define CmPtrCast(tp,p) \
 (tp::typeId().typeInfo()->canCast((p)->getTypeId().typeInfo()) ? (tp*)(p) : 0)

// Macro defines the static type info, virtual function, and static function
// needed in any classe that is to use run time type information.
//
#define CMTYPEID_DEFINE                                          \
        static const CmTypeInfo _infoObj;                        \
        virtual CmTypeId getTypeId() const { return &_infoObj; } \
        static  CmTypeId typeId   ()       { return &_infoObj; }

// Macro initializes the static type info for any class that is the
// first in a heirarchy of classes using run time type information.
//
#define CMTYPEID_INIT_BASE(name)                                  \
        static const CmTypeInfo* name ## TypeInfo [] = { 0 };     \
        const CmTypeInfo name::_infoObj(#name, name ## TypeInfo);

// Macro initializes the static type info for any class that singly
// derives from another class containing run time type information.
//
#define CMTYPEID_INIT(name,parent)                                \
        static const CmTypeInfo* name ## TypeInfo [] =            \
               {&parent::_infoObj, 0 };                           \
        const CmTypeInfo name::_infoObj(#name, name ## TypeInfo);

// Macro begins the initialization of the static type info for any
// class that derives from more than one class containing run time
// type information.
//
#define BEGIN_CMTYPEID_INIT(name)                        \
        static const CmTypeInfo* name ## TypeInfo [] = {

// Macro is used during the initialization of the static type info
// to add another parent to the list of parents being derived from
// using run time type information.
//
#define CMTYPEID_PARENT(parent) \
        &parent::_infoObj

// Macro ends the initialization of the static type info for any
// class that derives from more than one class containing run time
// type information.
//
#define END_CMTYPEID_INIT(name)                                   \
        0 };                                                      \
        const CmTypeInfo name::_infoObj(#name, name ## TypeInfo);

#endif
