// cmarray.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Array implementation.
// -----------------------------------------------------------------

#include <stdlib.h>
#include <cm/include/cmarray.h>
#include <cm/include/cmio.h>

// Macro calls initialize the static type info data.
CMOBJECT_INIT(CmArray, CmContainer)
CMOBJECT_INIT(CmArrayIterator, CmIterator)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmArray)


// "CmArray" is the default array constructor.
//
CmArray::CmArray(unsigned sz, unsigned dt) : _delta(dt), _total(0)
{
  _entries = (sz > 0) ? new CmObject*[sz] : NULL;
  _size    = (_entries) ? sz : 0;
}


// "CmArray" is the array copy constructor.
//
CmArray::CmArray(const CmArray& arr) : _delta(arr._delta), _total(0)
{
  _entries = (arr._size > 0) ? new CmObject*[arr._size] : NULL;
  _size    = (_entries) ? arr._size : 0;
  copy(arr);
}


// "CmArray" constructs an array from the specified container.
//
CmArray::CmArray(const CmContainer& cont) : _delta(0), _total(0)
{
  _entries = (cont.total() > 0) ? new CmObject*[cont.total()] : NULL;
  _size    = (_entries) ? cont.total() : 0;
  copy(cont);
}


// "~CmArray" is the array destructor.
//
CmArray::~CmArray()
{
  removeAll();
  delete[] _entries;
}


// "=" assignement operator copies the contents of the specified
// array into this array.
//
CmArray& CmArray::operator=(const CmArray& arr)
{
  if (&arr != this)
  {
    removeAll();
    delete[] _entries;
    _delta   = arr._delta;
    _total   = 0;
    _entries = (arr._size > 0) ? new CmObject*[arr._size] : NULL;
    _size    = (_entries) ? arr._size : 0;
    copy(arr);
  }
  return *this;
}


// "resize" resizes the storage allocated for this array.
//
bool CmArray::resize(unsigned newsize)
{
  CmObject **newentries = NULL;
  if (newsize == 0)
    removeAll();

  else
  {
    newentries = new CmObject*[newsize];
    if (!newentries) return false;
    if (newsize < _total)
    {
      for (int ii = newsize; ii < _total; ii++)
        if (ownsObjects()) delete _entries[ii];
      _total = newsize;
    }
    memcpy(newentries, _entries, _total * sizeof(CmObject*));
  }

  delete[] _entries;
  _entries = newentries;
  _size    = newsize;
  return true;
}


// "isEmpty" checks if the array is empty or not.
//
bool CmArray::isEmpty() const
{
  return (_total == 0) ? true : false;
}


// "add" appends the specified object to the array.
//
bool CmArray::add(CmObject* pObj)
{
  if (!pObj) return false;
  if (_total >= _size)
    if (_delta == 0 || !resize(_size + _delta))
      return false;
  _entries[_total++] = pObj;
  return true;
}


// "insertAt" inserts the specified object at the given index moving
// the other objects to make room.
//
bool CmArray::insertAt(int idx, CmObject* pObj)
{
  if (!pObj || idx < 0 || idx >= _total) return false;
  if (_total >= _size)
    if (_delta == 0 || !resize(_size + _delta))
      return false;

  size_t sz = (_total - idx) * sizeof(CmObject*);
  memmove(&(_entries[idx+1]), &(_entries[idx]), sz);
  _entries[idx] = pObj;
  _total++;
  return true;
}


// "replaceAt" replaces the object at the specified index with the
// specified object.
//
bool CmArray::replaceAt(int idx, CmObject* pObj)
{
  if (!pObj || idx < 0 || idx >= _total) return false;
  if (ownsObjects()) delete _entries[idx];
  _entries[idx] = pObj;
  return true;
}


// "remove" removes the first occurrence of an object that isEqual to
// the specified object from the array.
//
bool CmArray::remove(const CmObject* pObj)
{
  if (_total == 0 || !pObj) return false;
  int  ii    = 0;
  bool found = false;
  while (ii < _total && !found)
    if (_entries[ii++]->isEqual(pObj)) found = true;
  return (found) ? removeAt(ii-1) : false;
}


// "remove" removes the object pointed to by the specified iterator
// from the array.
//
bool CmArray::remove(const CmIterator& iter)
{
  if (_total == 0 || !iter.current()) return false;
  int  ii    = 0;
  bool found = false;
  while (ii < _total && !found)
    if (_entries[ii++] == iter.current()) found = true;
  return (found) ? removeAt(ii-1) : false;
}


// "removeAt" removes the object at the specified index.
//
bool CmArray::removeAt(int idx)
{
  if (idx < 0 || idx >= _total) return false;

  if (ownsObjects()) delete _entries[idx];
  size_t sz = (_total - idx - 1) * sizeof(CmObject*);
  memmove(&(_entries[idx]), &(_entries[idx+1]), sz);
  _entries[_total-1] = NULL;
  _total--;
  return true;
}


// "removeAll" removes all objects from the array.
//
void CmArray::removeAll()
{
  for (int ii = 0; ii < _total; ii++)
  {
    if (ownsObjects()) delete _entries[ii];
    _entries[ii] = NULL;
  }
  _total = 0;
}


// "index" returns the index of the first occurrence of an object
// that isEqual to the specified object.
//
int CmArray::index(const CmObject* pObj) const
{
  if (_total == 0 || !pObj) return -1;
  int ii = 0, idx = -1;
  while (ii < _total && idx == -1)
    if (_entries[ii++]->isEqual(pObj)) idx = ii - 1;
  return idx;
}


// "index" returns the array index of the object pointed to by the
// specified iterator.
//
int CmArray::index(const CmIterator& iter) const
{
  if (_total == 0 || !iter.current()) return -1;
  int ii = 0, idx = -1;
  while (ii < _total && idx == -1)
    if (_entries[ii++] == iter.current()) idx = ii - 1;
  return idx;
}


// "quickSort" uses the standard library to perform a quick sort
// on this array.
//
void CmArray::quickSort()
{
  qsort(_entries, _total, sizeof(CmObject*), &CmArray::cmpObjs);
}


// "lookup" returns the first occurrence of an object that isEqual
// to the specified object.
//
CmObject* CmArray::lookup(const CmObject* pObj) const
{
  int idx = index(pObj);
  return (idx > -1) ? _entries[idx] : NULL;
}


// "occurrences" returns the number of objects in the array
// isEqual to the specified object.
//
unsigned CmArray::occurrences(const CmObject* pObj) const
{
  if (_total == 0 || !pObj) return 0;

  int      ii  = 0;
  unsigned num = 0;
  while (ii < _total)
    if (_entries[ii++]->isEqual(pObj)) num++;
  return num;
}


// "newIterator" creates and returns an array iterator.
//
CmIterator* CmArray::newIterator() const
{
  return new CmArrayIterator(*this);
}


// "write" writes the array objects to the specified file.
//
bool CmArray::write(CmIOFile& file) const
{
  if (!writeBase (file))   return false;
  if (!file.write(_delta)) return false;
  if (!file.write(_total)) return false;

  bool success = true;
  int  ii      = 0;
  while (ii < _total && success)
    success = CmIO::writeObject(file, _entries[ii++]);
  return success;
}


// "read" reads the array objects from the specified file.
//
bool CmArray::read(CmIOFile& file)
{
  unsigned size = readBase(file);

  unsigned total;
  if (!file.read(_delta)) return false;
  if (!file.read(total))  return false;
  if (!resize   (size))   return false;

  removeAll();

  bool success = true;
  int  ii      = 0;
  while (ii < total && success)
  {
    CmObject *pObj = CmIO::readObject(file);
    if (pObj) success = add(pObj); ii++;
  }
  return success;
}


// "cmpObjs" is called by the standard library qsort function for
// sorting the array into ascending order.
//
int CmArray::cmpObjs(const void* obj1, const void* obj2)
{
  CmObject *pObj1 = *((CmObject**)obj1);
  CmObject *pObj2 = *((CmObject**)obj2);
  return pObj1->compare(pObj2);
}
