// cmbdict.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// BTree dictionary implementation.
// -----------------------------------------------------------------

#include <cm/include/cmbdict.h>

// Macro call initializes the static type info data.
CMOBJECT_INIT(CmBTreeDictionary, CmBTree)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmBTreeDictionary)


// "CmBTreeDictionary" is the default dictionary constructor.
//
CmBTreeDictionary::CmBTreeDictionary(unsigned sz)
                 : CmBTree(sz)
{}


// "CmBTreeDictionary" is the dictionary copy constructor.
//
CmBTreeDictionary::CmBTreeDictionary(const CmBTreeDictionary& dict)
                 : CmBTree(dict)
{}


// "~CmBTreeDictionary" is the dictionary destructor.
//
CmBTreeDictionary::~CmBTreeDictionary()
{
  removeAll();
}


// "=" assignment operator copies the specified dictionary into
// this dictionary.
//
CmBTreeDictionary& CmBTreeDictionary::operator=(const CmBTreeDictionary& dict)
{
  return (CmBTreeDictionary&) CmBTree::operator=(dict);
}


// "add" adds an association to the dictionary.
//
bool CmBTreeDictionary::add(CmObject* pObj)
{
  if (CmPtrCast(CmAssociation,pObj) && !CmBTree::contains(pObj))
    return CmBTree::add(pObj);
  return false;
}


// "add" adds a key/object pair to the dictionary.
//
bool CmBTreeDictionary::add(CmObject* pKey, CmObject* pObj)
{
  CmAssociation *pAssoc = new CmAssociation(pKey, pObj);
  if (!pAssoc) return false;

  if (CmBTree::contains(pAssoc))
  {
    delete pAssoc;
    return false;
  }
  else
    return CmBTree::add(pAssoc);
}


// "replace" looks for the first occurrence of an association with
// a key that isEqual to the specified key and replaces it's object
// value with the specified object.
//
bool CmBTreeDictionary::replace(const CmObject* pKey, CmObject* pObj)
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  if (!pAssoc) return false;

  if (ownsObjects()) delete pAssoc->object();
  pAssoc->object(pObj);
  return true;
}


// "remove" looks for the first occurrence of an association with
// a key that isEqual to the specified key and removes it.
//
bool CmBTreeDictionary::remove(const CmObject* pKey)
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  if (pAssoc)
  {
    CmObject* pKey    = pAssoc->key   ();
    CmObject* pObject = pAssoc->object();
    bool success = CmBTree::remove(pAssoc);
    if (success && ownsObjects())
    {
      delete pKey;
      delete pObject;
    }
    return success;
  }
  return false;
}


// "remove" removes the association pointed to by the iterator.
//
bool CmBTreeDictionary::remove(const CmIterator& iter)
{
  return CmBTree::remove(iter);
}


// "removeAll" removes all of the associations from the dictionary.
//
void CmBTreeDictionary::removeAll()
{
  if (ownsObjects())
  {
    CmBTreeIterator iter(*this);
    while (!iter.done())
    {
      CmAssociation *pAssoc = (CmAssociation*) iter.next();
      delete pAssoc->key   ();
      delete pAssoc->object();
    }
  }
  CmBTree::removeAll();
}


// "lookup" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association's
// key value.
//
CmObject* CmBTreeDictionary::lookup(const CmObject* pKey) const
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  return (pAssoc) ? pAssoc->key() : NULL;
}


// "lookupObject" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association's
// object value.
//
CmObject* CmBTreeDictionary::lookupObject(const CmObject* pKey) const
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  return (pAssoc) ? pAssoc->object() : NULL;
}


// "lookupObject" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association.
//
CmAssociation* CmBTreeDictionary::lookupAssoc(const CmObject* pKey) const
{
  CmAssociation assoc((CmObject*) pKey, NULL);
  return (CmAssociation*) CmBTree::lookup(&assoc);
}


// "contains" checks to see is an association exists with a key which
// isEqual to the specified key.
//
bool CmBTreeDictionary::contains(const CmObject* pKey) const
{
  return (lookupAssoc(pKey) == NULL) ? false : true;
}


// "occurrences" returns the number of associations with a key which
// isEqual to the specified key.  (This should always be 1).
//
unsigned CmBTreeDictionary::occurrences(const CmObject* pKey) const
{
  CmAssociation assoc((CmObject*) pKey, NULL);
  return CmBTree::occurrences(&assoc);
}
