// cmbintr.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Binary tree implementation.
// -----------------------------------------------------------------

#include <cm/include/cmbintr.h>

// Macro calls initialize the static type info data.
CMOBJECT_INIT(CmBinaryTree, CmContainer)
CMOBJECT_INIT(CmBinaryTreeIterator, CmIterator)
CMOBJECT_INIT(CmBinaryTreeNode, CmObject)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmBinaryTree)


// "CmBinaryTree" is the tree copy constructor.
//
CmBinaryTree::CmBinaryTree(const CmBinaryTree& tree) : _root(NULL)
{
  copy   (tree);
  balance();
}


// "CmBinaryTree" constructs a binary tree from the specified
// container.
//
CmBinaryTree::CmBinaryTree(const CmContainer& cont) : _root(NULL)
{
  copy   (cont);
  balance();
}


// "~CmBinaryTree" is the tree destructor.
//
CmBinaryTree::~CmBinaryTree()
{
  removeAll();
}


// "=" assignment operator copies the contents of the specified tree
// into this tree.
//
CmBinaryTree& CmBinaryTree::operator=(const CmBinaryTree& tree)
{
  if (&tree != this)
  {
    removeAll();
    copy     (tree);
    balance  ();
  }
  return *this;
}


// "add" adds the specified object to this tree.
//
bool CmBinaryTree::add(CmObject* pObj)
{
  if (!pObj) return false;

  CmBinaryTreeNode *newnode = new CmBinaryTreeNode(pObj);
  if (!newnode) return false;

  if (!_root)                                // Tree is empty.  Add at root.
    _root = newnode;

  else                                       // Tree is not empty.
  {
    int               comp   = pObj->compare(_root->_data);
    CmBinaryTreeNode *rover  = (comp < 0) ? _root->_left : _root->_right;
    CmBinaryTreeNode *parent = _root;

    while (rover)                            // Search for vacant slot.
    {
      parent = rover;
      comp   = pObj->compare(rover->_data);
      rover  = (comp < 0) ? rover->_left : rover->_right;
    }

    if (comp < 0) parent->_left  = newnode;  // Add new node appropriately.
    else          parent->_right = newnode;
  }
  _size++;
  return true;
}


// "remove" removes the first occurrence of an object that isEqual
// to the specified object from the tree.
//
bool CmBinaryTree::remove(const CmObject* pObj)
{
  if (!_root || !pObj) return false;           // Exit if empty tree.

  CmBinaryTreeNode *rover   = _root;           // Init for search.
  CmBinaryTreeNode *parent  = NULL;
  bool              done    = false;
  bool              success = false;

  while (rover && !done)                       // Search for object.
  {
    int comp = pObj->compare(rover->_data);
    if (comp != 0)                             // Advance to next node.
    {
      parent = rover;
      rover  = (comp < 0) ? rover->_left : rover->_right;
    }

    else                                       // Object was found.
    {
      success = removeNode(parent, rover);
      done    = true;
    }
  }
  return success;
}


// "remove" removes the object pointed to by the specified iterator
// object.
//
bool CmBinaryTree::remove(const CmIterator& iter)
{
  if (!_root || !iter.current()) return false; // Exit if empty tree.

  CmBinaryTreeNode *rover   = _root;           // Init for search.
  CmBinaryTreeNode *parent  = NULL;
  bool              done    = false;
  bool              success = false;

  while (rover && !done)                       // Search for object.
  {
    int comp = iter.current()->compare(rover->_data);
    if (comp != 0)                             // Advance to next node.
    {
      parent = rover;
      rover  = (comp < 0) ? rover->_left : rover->_right;
    }

    else if (comp == 0 && rover->_data != iter.current())
    {
      parent = rover;
      rover  = rover->_right;
    }

    else                                       // Object was found.
    {
      success = removeNode(parent, rover);
      done    = true;
    }
  }
  return success;
}


// "removeAt" removes the object at the specified index.
//
bool CmBinaryTree::removeAt(int idx)
{
  CmObject *pObj = at(idx);
  if (!_root || !pObj) return false;           // Exit if empty tree.

  CmBinaryTreeNode *rover   = _root;           // Init for search.
  CmBinaryTreeNode *parent  = NULL;
  bool              done    = false;
  bool              success = false;

  while (rover && !done)                       // Search for object.
  {
    if (rover->_data == pObj)                  // This is it.
    {
      success = removeNode(parent, rover);
      done    = true;
    }
    else
    {
      int comp = pObj->compare(rover->_data);  // Do object compare.
      if (comp != 0)                           // Advance to next node.
      {
        parent = rover;
        rover  = (comp < 0) ? rover->_left : rover->_right;
      }
      else if (comp == 0 && rover->_data != pObj)
      {
        parent = rover;
        rover  = rover->_right;
      }
    }
  }
  return success;
}


// "removeAll" removes all the objects from the tree.
//
void CmBinaryTree::removeAll()
{
  if (_root)
  {
    killNode(_root, ownsObjects());
    if (ownsObjects()) delete _root->_data;
    delete _root;
    _root = NULL;
    _size = 0;
  }
}


// "balance" balances the current contents of the binary tree.
//
void CmBinaryTree::balance()
{
  CmQueue que;
  que.ownsObjects(false);

  CmBinaryTreeIterator iter(*this);
  while (!iter.done()) que.push(iter.next());

  bool owns = ownsObjects();
  if (owns) ownsObjects(false);
  removeAll();
  if (owns) ownsObjects(true);

  balanceHalf(que, 0, que.size() - 1);
}


// "lookup" returns the first occurrence of an object which isEqual
// to the specified object.
//
CmObject* CmBinaryTree::lookup(const CmObject *pObj) const
{
  if (!_root || !pObj) return NULL;

  CmBinaryTreeNode *rover = _root;
  CmObject         *ret   = NULL;

  while (rover && !ret)
  {
    int comp = pObj->compare(rover->_data);
    if (comp == 0) ret    = rover->_data;
    else           rover  = (comp < 0) ? rover->_left : rover->_right;
  }
  return ret;
}


// "contains" checks to see if an object that isEqual to the specified
// object exists in the tree.
//
bool CmBinaryTree::contains(const CmObject *pObj) const
{
  return (lookup(pObj) == NULL) ? false : true;
}


// "occurrences" returns the number of objects in the tree
// isEqual to the specified object.
//
unsigned CmBinaryTree::occurrences(const CmObject *pObj) const
{
  unsigned             total = 0;
  CmBinaryTreeIterator iter(*this);
  while (!iter.done())
    if (pObj->isEqual(iter.next()))
      total++;
  return total;
}


// "newIterator" creates and returns a new tree iterator.
//
CmIterator* CmBinaryTree::newIterator() const
{
  return new CmBinaryTreeIterator(*this);
}


// "write" writes the tree contents to the specified file.
//
bool CmBinaryTree::write(CmIOFile& file) const
{
  return CmContainer::write(file);
}


// "read" reads objects from the specified file.
//
bool CmBinaryTree::read(CmIOFile& file)
{
  bool success = CmContainer::read(file);
  if (success) balance();
  return success;
}


// "removeNode" removes the specified node from the tree.
//
bool CmBinaryTree::removeNode(CmBinaryTreeNode *parent, CmBinaryTreeNode *node)
{
  CmBinaryTreeNode *temp1 = node;

  if (node->_right == NULL)
    node = node->_left;

  else if (node->_right->_left == NULL)
  {
    node        = node->_right;
    node->_left = temp1->_left;
  }

  else
  {
    CmBinaryTreeNode *temp2 = node->_right;
    while (temp2->_left->_left) temp2 = temp2->_left;
    node         = temp2->_left;
    temp2->_left = node->_right;
    node->_left  = temp1->_left;
    node->_right = temp1->_right;
  }

  if (!parent)
    _root = node;

  else
  {
    int comp = temp1->_data->compare(parent->_data);
    if (comp < 0) parent->_left  = node;
    else          parent->_right = node;
  }

  if (ownsObjects()) delete temp1->_data;
  delete temp1;
  _size--;
  return true;
}


// "balanceHalf" is called recursively to balance part of the tree.
//
void CmBinaryTree::balanceHalf(const CmQueue& que, int low, int high)
{
  if (low <= high)
  {
    int mid = (low + high) / 2;
    add(que[mid]);
    balanceHalf(que, mid + 1, high);
    balanceHalf(que, low, mid - 1);
  }
}


// "killNode" is a recursively called static function which takes a
// subtree and destroys the left and right branches.
//
void CmBinaryTree::killNode(CmBinaryTreeNode* subTree, bool owns)
{
  if (subTree->_left)
  {
    killNode(subTree->_left, owns);
    if (owns) delete subTree->_left->_data;
    delete subTree->_left;
    subTree->_left = NULL;
  }

  if (subTree->_right)
  {
    killNode(subTree->_right, owns);
    if (owns) delete subTree->_right->_data;
    delete subTree->_right;
    subTree->_right = NULL;
  }
}


// "CmBinaryTreeIterator" is the constructor for the tree iterator
// class.
//
CmBinaryTreeIterator::CmBinaryTreeIterator(const CmBinaryTree &Tr)
                    : CmIterator(Tr), _node(Tr._root)
{
  _stack.ownsObjects(false);
  if (_node)
  {
    while (_node->_left)
    {
      _stack.push(_node);
      _node = _node->_left;
    }
  }
}


// "next" returns a pointer to the object pointed to by the iterator.  The
// iterator is then advanced to the next node in the tree.
//
CmObject* CmBinaryTreeIterator::next()
{
  if (!_node) return NULL;                        // Iteration is over.

  CmObject* retValue = _node->_data;              // Save off object pointer.

  if (_node->_right)                              // If there is a right
  {                                               // branch, advance to it
    _stack.push(_node);                           // and descend all the way
    _node = _node->_right;                        // down it's left side.
    while (_node->_left)
    {
      _stack.push(_node);
      _node = _node->_left;
    }
  }

  else                                            // Otherwise, back up the
  {                                               // tree until we are back
    CmBinaryTreeNode* temp;                       // at the top or until we
    do                                            // are no longer a right
    {                                             // branch.
      temp  = _node;
      _node = (_stack.size() == 0) ? NULL : (CmBinaryTreeNode*) _stack.pop();
    } while (_node && _node->_right == temp);
  }
  return retValue;
}


// "previous" returns a pointer to the current object and decrements
// the iterator.
//
CmObject* CmBinaryTreeIterator::previous()
{
  if (!_node) return NULL;                        // Iteration is over.

  CmObject* retValue = _node->_data;              // Save off object pointer.

  if (_node->_left)                               // If there is a left
  {                                               // branch, advance to it
    _stack.push(_node);                           // and descend all the way
    _node = _node->_left;                         // down it's right side.
    while (_node->_right)
    {
      _stack.push(_node);
      _node = _node->_right;
    }
  }

  else                                            // Otherwise, back up the
  {                                               // tree until we are back
    CmBinaryTreeNode* temp;                       // at the top or until we
    do                                            // are no longer a left
    {                                             // branch.
      temp  = _node;
      _node = (_stack.size() == 0) ? NULL : (CmBinaryTreeNode*) _stack.pop();
    } while (_node && _node->_left == temp);
  }
  return retValue;
}


// "first" moves the iterator to the first object in the tree.
//
void CmBinaryTreeIterator::first()
{
  _stack.removeAll();
  _node = tree()._root;
  if (_node)
  {
    while (_node->_left)
    {
      _stack.push(_node);
      _node = _node->_left;
    }
  }
}


// "last" moves the iterator to the right most tree node thus
// pointing to the last object in the tree.
//
void CmBinaryTreeIterator::last()
{
  _stack.removeAll();
  _node = tree()._root;
  if (_node)
  {
    while (_node->_right)
    {
      _stack.push(_node);
      _node = _node->_right;
    }
  }
}
