// cmcont.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Abstract container implementation.
// -----------------------------------------------------------------

#include <cm/include/cmcont.h>
#include <cm/include/cmlist.h>
#include <cm/include/cmio.h>

// Macro call initializes the static type info data.
CMOBJECT_INIT(CmContainer, CmObject)


// "at" returns a pointer to the object at the specified index in
// the container.
//
CmObject* CmContainer::at(int idx) const
{
  if (idx < 0 || idx >= size()) return NULL;
  CmIterator *iter = newIterator();
  CmObject   *out  = NULL;
  int         ii   = 0;
  while (!iter->done() && ii <= idx)
  {
    if (ii++ == idx) out = iter->current();
    iter->next();
  }
  delete iter;
  return out;
}


// "lookup" returns a pointer to the first occurrence of an object
// that isEqual to the specified object.
//
CmObject* CmContainer::lookup(const CmObject* pObj) const
{
  if (!pObj) return NULL;
  CmIterator *iter = newIterator();
  CmObject   *out  = NULL;
  while (!iter->done() && !out)
  {
    CmObject *temp = iter->next();
    if (pObj->isEqual(temp)) out = temp;
  }
  delete iter;
  return out;
}


// "contains" returns true if the container contains an object
// that isEqual to the specified object.
//
bool CmContainer::contains(const CmObject* pObj) const
{
  if (!pObj) return false;
  CmIterator *iter  = newIterator();
  bool        found = false;
  while (!iter->done() && !found)
    if (pObj->isEqual(iter->next()))
      found = true;
  delete iter;
  return found;
}


// "occurrences" returns the number of objects in this container
// that are isEqual to the specified object.
//
unsigned CmContainer::occurrences(const CmObject* pObj) const
{
  if (!pObj) return 0;
  CmIterator *iter  = newIterator();
  unsigned    occur = 0;
  while (!iter->done())
    if (pObj->isEqual(iter->next()))
      occur++;
  delete iter;
  return occur;
}


// "lookupAll" allocates and returns a container with pointers to
// all objects in this container equal to the specified object.
//
CmContainer* CmContainer::lookupAll(const CmObject* pObj) const
{
  CmLinkedList *pList = new CmLinkedList;
  CmIterator   *iter  = newIterator();

  pList->ownsObjects(false);
  while (!iter->done())
  {
    CmObject* temp = iter->next();
    if (temp->isEqual(pObj)) pList->add(temp);
  }

  delete iter;
  return pList;
}


// "addAllFrom" adds all the objects from the specified container
// to this container.
//
bool CmContainer::addAllFrom(const CmContainer* container)
{
  CmIterator *iter = container->newIterator();
  while (!iter->done())
  {
    if (ownsObjects()) add(iter->next()->newCopy());
    else               add(iter->next());
  }
  delete iter;
  return true;
}


// "printOn" prints the type of container and the container contents
// on the specified output stream.
//
void CmContainer::printOn(ostream& os) const
{
  os << "Type " << CmPtrTypeId(this).name() << " contains -" << endl;
  CmIterator *iter = newIterator();
  while (!iter->done())
    os << *(iter->next()) << endl;
  delete iter;
}


// "write" writes the container contents to the specified file.
//
bool CmContainer::write(CmIOFile& file) const
{
  if (!writeBase(file)) return false;                // Write base container.

  bool        success = true;
  CmIterator *iter    = newIterator();
  while (!iter->done() && success)                   // Cycle through objects.
    success = CmIO::writeObject(file, iter->next()); // Write the object.
  delete iter;
  return success;
}


// "read" reads objects from the specified file into this container.
//
bool CmContainer::read(CmIOFile& file)
{
  unsigned size = readBase(file);                 // Read base container.

  removeAll();                                    // Empty the container.

  int  ii      = 0;
  bool success = true;
  while (ii < size && success)                    // For each object.
  {
    CmObject *pObj = CmIO::readObject(file);      // Read object.
    if (pObj) success = add(pObj); ii++;          // Add to container.
  }
  return success;
}


// "forEach" iterates through all the objects in the container calling
// the specified function for each object until there are no more objects
// or the function returns false.
//
void CmContainer::forEach(CmQueryFunc pFunc, void* pData)
{
  bool        cont = true;
  CmIterator* iter = newIterator();
  while (!iter->done() && cont)
    cont = (*pFunc)(iter->next(), pData);
  delete iter;
}


// "forEach" iterates through all the objects in the container calling
// the specified object member function for each object until there are
// no more objects or the function returns false.
//
void CmContainer::forEach(CmQueryMemFunc pFunc, void* pData)
{
  bool        cont = true;
  CmIterator* iter = newIterator();
  while (!iter->done() && cont)
  {
    CmObject *pObj = iter->next();
    cont = (pObj->*pFunc)(pData);
  }
  delete iter;
}


// "firstThat" returns the first object in the container that is
// passed into the specified function and returns true.
//
CmObject* CmContainer::firstThat(CmQueryFunc pFunc, void* pData) const
{
  CmObject   *outObj = NULL;
  CmIterator *iter   = newIterator();
  while (!iter->done() && !outObj)
  {
    CmObject *pObj = iter->next();
    if ((*pFunc)(pObj, pData)) outObj = pObj;
  }
  delete iter;
  return outObj;
}


// "firstThat" returns the first object in the container who's specified
// member function returns true.
//
CmObject* CmContainer::firstThat(CmQueryMemFunc pFunc, void* pData) const
{
  CmObject   *outObj = NULL;
  CmIterator *iter   = newIterator();
  while (!iter->done() && !outObj)
  {
    CmObject *pObj = iter->next();
    if ((pObj->*pFunc)(pData)) outObj = pObj;
  }
  delete iter;
  return outObj;
}


// "lastThat" returns the last object in the container that is
// passed into the specified function and returns true.
//
CmObject* CmContainer::lastThat(CmQueryFunc pFunc, void* pData) const
{
  CmObject   *outObj = NULL;
  CmIterator *iter   = newIterator();
  iter->last();
  while (!iter->done() && !outObj)
  {
    CmObject *pObj = iter->previous();
    if ((*pFunc)(pObj, pData)) outObj = pObj;
  }
  delete iter;
  return outObj;
}


// "lastThat" returns the last object in the container who's specified
// member function returns true.
//
CmObject* CmContainer::lastThat(CmQueryMemFunc pFunc, void* pData) const
{
  CmObject   *outObj = NULL;
  CmIterator *iter   = newIterator();
  iter->last();
  while (!iter->done() && !outObj)
  {
    CmObject *pObj = iter->previous();
    if ((pObj->*pFunc)(pData)) outObj = pObj;
  }
  delete iter;
  return outObj;
}


// "query" cycles through all the objects in the container calling the
// specified function for each.  Every object in which the function returns
// returns true gets added to a subset container which is then returned
// from the function.
//
CmContainer* CmContainer::query(CmQueryFunc pFunc, void* pData) const
{
  CmLinkedList *pList = new CmLinkedList;
  CmIterator   *iter  = newIterator();

  pList->ownsObjects(false);
  while (!iter->done())
  {
    CmObject* pObj = iter->next();
    if ((*pFunc)(pObj, pData)) pList->add(pObj);
  }

  delete iter;
  return pList;
}


// "query" cycles through all the objects in the container calling the
// specified object member function for each.  Every object in which the
// function returns returns true gets added to a subset container which
// is then returned from the function.
//
CmContainer* CmContainer::query(CmQueryMemFunc pFunc, void* pData) const
{
  CmLinkedList *pList = new CmLinkedList;
  CmIterator   *iter  = newIterator();

  pList->ownsObjects(false);
  while (!iter->done())
  {
    CmObject* pObj = iter->next();
    if ((pObj->*pFunc)(pData)) pList->add(pObj);
  }

  delete iter;
  return pList;
}


// "copy" copies the contents from the specified container into this
// container.
//
void CmContainer::copy(const CmContainer& C)
{
  _ownsObjects = C._ownsObjects;
  CmIterator *iter = C.newIterator();
  while (!iter->done())
  {
    if (ownsObjects()) add(iter->next()->newCopy());
    else               add(iter->next());
  }
  delete iter;
}


// "writeBase" writes the base container members to the specified file.
//
bool CmContainer::writeBase(CmIOFile& file) const
{
  bool success = file.write(_ownsObjects);
  if (success) success = file.write(_size);
  return success;
}


// "readBase" reads the base container members from the specified file
// and returns the container size (the _size member is not set here).
//
unsigned CmContainer::readBase(CmIOFile& file)
{
  unsigned size;
  bool success = file.read(_ownsObjects);
  if (success) success = file.read(size);
  return (success) ? size : (unsigned) 0;
}
