// cmdict.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Dictionary implementation.
// -----------------------------------------------------------------

#include <cm/include/cmdict.h>

// Macro call initializes the static type info data.
CMOBJECT_INIT(CmDictionary, CmDLinkedList)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmDictionary)


// "CmDictionary" is the default dictionary constructor.
//
CmDictionary::CmDictionary()
{}


// "CmDictionary" is the dictionary copy constructor.
//
CmDictionary::CmDictionary(const CmDictionary& dict)
            : CmDLinkedList(dict)
{}


// "~CmDictionary" is the dictionary destructor.
//
CmDictionary::~CmDictionary()
{
  removeAll();
}


// "=" assignment operator copies the specified dictionary into
// this dictionary.
//
CmDictionary& CmDictionary::operator=(const CmDictionary& dict)
{
  return (CmDictionary&) CmDLinkedList::operator=(dict);
}


// "add" adds an association to the dictionary.
//
bool CmDictionary::add(CmObject* pObj)
{
  if (CmPtrCast(CmAssociation,pObj) && !CmDLinkedList::contains(pObj))
    return CmDLinkedList::add(pObj);
  return false;
}


// "add" adds a key/object pair to the dictionary.
//
bool CmDictionary::add(CmObject* pKey, CmObject* pObj)
{
  CmAssociation *pAssoc = new CmAssociation(pKey, pObj);
  if (!pAssoc) return false;

  if (CmDLinkedList::contains(pAssoc))
  {
    delete pAssoc;
    return false;
  }
  else
    return CmDLinkedList::add(pAssoc);
}


// "replace" looks for the first occurrence of an association with
// a key that isEqual to the specified key and replaces it's object
// value with the specified object.
//
bool CmDictionary::replace(const CmObject* pKey, CmObject* pObj)
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  if (!pAssoc) return false;

  if (ownsObjects()) delete pAssoc->object();
  pAssoc->object(pObj);
  return true;
}


// "remove" looks for the first occurrence of an association with
// a key that isEqual to the specified key and removes it.
//
bool CmDictionary::remove(const CmObject* pKey)
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  if (pAssoc)
  {
    CmObject* pKey    = pAssoc->key   ();
    CmObject* pObject = pAssoc->object();
    bool success = CmDLinkedList::remove(pAssoc);
    if (success && ownsObjects())
    {
      delete pKey;
      delete pObject;
    }
    return success;
  }
  return false;
}


// "remove" removes the association pointed to by the iterator.
//
bool CmDictionary::remove(const CmIterator& iter)
{
  return CmDLinkedList::remove(iter);
}


// "removeAll" removes all of the associations from the dictionary.
//
void CmDictionary::removeAll()
{
  if (ownsObjects())
  {
    CmDLinkedListIterator iter(*this);
    while (!iter.done())
    {
      CmAssociation *pAssoc = (CmAssociation*) iter.next();
      delete pAssoc->key   ();
      delete pAssoc->object();
    }
  }
  CmDLinkedList::removeAll();
}


// "lookup" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association's
// key value.
//
CmObject* CmDictionary::lookup(const CmObject* pKey) const
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  return (pAssoc) ? pAssoc->key() : NULL;
}


// "lookupObject" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association's
// object value.
//
CmObject* CmDictionary::lookupObject(const CmObject* pKey) const
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  return (pAssoc) ? pAssoc->object() : NULL;
}


// "lookupObject" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association.
//
CmAssociation* CmDictionary::lookupAssoc(const CmObject* pKey) const
{
  CmAssociation assoc((CmObject*) pKey, NULL);
  return (CmAssociation*) CmDLinkedList::lookup(&assoc);
}


// "contains" checks to see is an association exists with a key which
// isEqual to the specified key.
//
bool CmDictionary::contains(const CmObject* pKey) const
{
  return (lookupAssoc(pKey) == NULL) ? false : true;
}


// "occurrences" returns the number of associations with a key which
// isEqual to the specified key.  (This should always be 1).
//
unsigned CmDictionary::occurrences(const CmObject* pKey) const
{
  CmAssociation assoc((CmObject*) pKey, NULL);
  return CmDLinkedList::occurrences(&assoc);
}
