// cmdlist.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Doubly linked list implementation.
// -----------------------------------------------------------------

#include <cm/include/cmdlist.h>

// Macro calls initialize the static type info data.
CMOBJECT_INIT(CmDLinkedList, CmContainer)
CMOBJECT_INIT(CmDLinkedListIterator, CmIterator)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmDLinkedList)


// "CmDLinkedList" is the list copy constructor.
//
CmDLinkedList::CmDLinkedList(const CmDLinkedList& list)
             : _first(NULL), _last(NULL)
{
  copy(list);
}


// "CmDLinkedList" constructs a list from the specified container.
//
CmDLinkedList::CmDLinkedList(const CmContainer& cont)
             : _first(NULL), _last(NULL)
{
  copy(cont);
}


// "~CmDLinkedList" is the list destructor.
//
CmDLinkedList::~CmDLinkedList()
{
  removeAll();
}


// "=" assignment operator copies the contents of the specified list
// into this list.
//
CmDLinkedList& CmDLinkedList::operator=(const CmDLinkedList& list)
{
  if (&list != this)
  {
    removeAll();
    copy     (list);
  }
  return *this;
}


// "add" adds the specified object to the end of this list.
//
bool CmDLinkedList::add(CmObject* pObj)
{
  return append(pObj);
}


// "append" adds the specified object to the end of this list.
//
bool CmDLinkedList::append(CmObject* pObj)
{
  if (!pObj) return false;
  CmDLink *newlink = new CmDLink(pObj);
  if (!newlink) return false;

  if (!_first)
    _first = _last = newlink;
  else
  {
    _last->_next   = newlink;
    newlink->_prev = _last;
    _last          = newlink;
  }
  _size++;
  return true;
}


// "prepend" adds the specified object to the start of this list.
//
bool CmDLinkedList::prepend(CmObject* pObj)
{
  if (!pObj) return false;
  CmDLink *newlink = new CmDLink(pObj);
  if (!newlink) return false;

  if (!_first)
    _first = _last = newlink;
  else
  {
    newlink->_next = _first;
    _first->_prev  = newlink;
    _first         = newlink;
  }
  _size++;
  return true;
}


// "insertAt" inserts the specified object at the specified index.
//
bool CmDLinkedList::insertAt(int idx, CmObject* pObj)
{
  if (idx < 0 || idx >= _size) return false;
  CmDLinkedListIterator iter(*this);
  int                   ii = 0;
  while (!iter.done() && ii < idx)
  {
    iter.next();
    ii++;
  }
  return insertBefore(iter, pObj);
}


// "replaceAt" replaces the object at the specified index with the
// specified object.
//
bool CmDLinkedList::replaceAt(int idx, CmObject* pObj)
{
  if (idx < 0 || idx >= _size) return false;

  int      ii    = 0;
  CmDLink *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  if (!rover) return false;
  if (ownsObjects()) delete rover->_data;
  rover->_data = pObj;
  return true;
}


// "remove" removes the first occurrence of an object that isEqual
// to the specified object from the list.
//
bool CmDLinkedList::remove(const CmObject* pObj)
{
  if (!pObj || !_first) return false;
  CmObject *remObj = CmDLinkedList::lookup(pObj);
  return (remObj) ? removeAbs(remObj) : false;
}


// "remove" removes the object pointed to by the specified iterator
// object.
//
bool CmDLinkedList::remove(const CmIterator& iter)
{
  CmObject *remObj = iter.current();
  return (remObj) ? removeAbs(remObj) : false;
}


// "removeAt" removes the object at the specified list index.
//
bool CmDLinkedList::removeAt(int idx)
{
  CmObject *remObj = at(idx);
  return (remObj) ? removeAbs(remObj) : false;
}


// "removeAll" removes all the objects from the list.
//
void CmDLinkedList::removeAll()
{
  CmDLink *temp, *rover = _first;
  while (rover)
  {
    temp = rover->_next;
    if (ownsObjects()) delete rover->_data;
    delete rover;
    rover = temp;
  }
  _size  = 0;
  _first = _last = NULL;
}


// "removeFirst" removes the first object from the list.  The object is
// not deleted regardless of the ownership.
//
CmObject* CmDLinkedList::removeFirst()
{
  if (!_first) return NULL;

  CmObject *pObj = _first->_data;
  CmDLink  *temp = _first;
  _first = _first->_next;
  if (!_first) _last = NULL;
  else _first->_prev = NULL;
  delete temp;
  _size--;
  return pObj;
}


// "removeLast" removes the last object from the list.  The object is
// not deleted regardless of the ownership.
//
CmObject* CmDLinkedList::removeLast()
{
  if (!_last) return NULL;
  if (_last == _first) return removeFirst();

  CmObject *pObj = _last->_data;
  CmDLink  *temp = _last;
  _last = _last->_prev;
  if (!_last) _first = NULL;
  else _last->_next  = NULL;
  delete temp;
  _size--;
  return pObj;
}


// "insertAfter" searches for the first occurrence of an object that
// isEqual to the first specified object and inserts the second
// specified object immediately after.
//
bool CmDLinkedList::insertAfter(const CmObject* after, CmObject* pObj)
{
  if (!after || !pObj || !_first) return false;

  CmDLink *rover = _first;
  bool     found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(after)) found = true;
    else                              rover = rover->_next;
  }
  if (!found) return false;

  CmDLink *newlink = new CmDLink(pObj);
  if (!newlink) return false;

  if (rover->_next) rover->_next->_prev = newlink;
  else _last = newlink;
  newlink->_next = rover->_next;
  newlink->_prev = rover;
  rover->_next   = newlink;
  _size++;
  return true;
}


// "insertAfter" searches for the object pointed to by the specified
// iterator and inserts the specified object immediately after.
//
bool CmDLinkedList::insertAfter(const CmIterator& iter, CmObject* pObj)
{
  if (!iter.current() || !pObj || !_first) return false;

  CmObject *after = iter.current();
  CmDLink  *rover = _first;
  bool      found = false;
  while (rover && !found)
  {
    if (rover->_data == after) found = true;
    else                       rover = rover->_next;
  }
  if (!found) return false;

  CmDLink *newlink = new CmDLink(pObj);
  if (!newlink) return false;

  if (rover->_next) rover->_next->_prev = newlink;
  else _last = newlink;
  newlink->_next = rover->_next;
  newlink->_prev = rover;
  rover->_next   = newlink;
  _size++;
  return true;
}


// "insertBefore" searches for the first occurrence of an object that
// isEqual to the first specified object and inserts the second
// specified object immediately before.
//
bool CmDLinkedList::insertBefore(const CmObject* before, CmObject* pObj)
{
  if (!before || !pObj || !_first) return false;

  CmDLink *rover = _first;
  bool     found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(before)) found = true;
    else                               rover = rover->_next;
  }
  if (!found) return false;

  CmDLink *newlink = new CmDLink(pObj);
  if (!newlink) return false;

  if (rover->_prev) rover->_prev->_next = newlink;
  else _first = newlink;
  newlink->_prev = rover->_prev;
  newlink->_next = rover;
  rover->_prev   = newlink;
  _size++;
  return true;
}


// "insertBefore" searches for the object pointed to by the specified
// iterator and inserts the specified object immediately before.
//
bool CmDLinkedList::insertBefore(const CmIterator& iter, CmObject* pObj)
{
  if (!iter.current() || !pObj || !_first) return false;

  CmObject *before = iter.current();
  CmDLink *rover = _first;
  bool     found = false;
  while (rover && !found)
  {
    if (rover->_data == before) found = true;
    else                        rover = rover->_next;
  }
  if (!found) return false;

  CmDLink *newlink = new CmDLink(pObj);
  if (!newlink) return false;

  if (rover->_prev) rover->_prev->_next = newlink;
  else _first = newlink;
  newlink->_prev = rover->_prev;
  newlink->_next = rover;
  rover->_prev   = newlink;
  _size++;
  return true;
}


// "index" returns the index in this list of the first occurrence of
// an object that isEqual to the specified object.
//
int CmDLinkedList::index(const CmObject* pObj) const
{
  CmDLink *rover  = _first;
  int      outidx = -1, idx = 0;
  while (outidx < 0 && rover)
  {
    if (rover->_data->isEqual(pObj)) outidx = idx;
    rover = rover->_next;
    idx++;
  }
  return outidx;
}


// "index" returns the index in the list of the object pointed to by
// the specified iterator.
//
int CmDLinkedList::index(const CmIterator& iter) const
{
  CmDLink *rover  = _first;
  int      outidx = -1, idx = 0;
  while (outidx < 0 && rover)
  {
    if (rover->_data == iter.current()) outidx = idx;
    rover = rover->_next;
    idx++;
  }
  return outidx;
}


// "at" returns the object at the given list index.
//
CmObject* CmDLinkedList::at(int idx) const
{
  if (idx < 0 || idx >= _size) return NULL;

  int      ii    = 0;
  CmDLink *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  return (rover) ? rover->_data : NULL;
}


// "lookup" returns the first occurrence of an object which isEqual
// to the specified object.
//
CmObject* CmDLinkedList::lookup(const CmObject* pObj) const
{
  if (!pObj || !_first) return NULL;

  CmDLink *rover = _first;
  bool     found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(pObj)) found = true;
    else                             rover = rover->_next;
  }
  return (found) ? rover->_data : NULL;
}


// "contains" checks to see if an object that isEqual to the specified
// object exists in the list.
//
bool CmDLinkedList::contains(const CmObject* pObj) const
{
  if (!pObj || !_first) return false;

  CmDLink *rover = _first;
  bool    found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(pObj)) found = true;
    else                             rover = rover->_next;
  }
  return found;
}


// "occurrences" returns the number of objects in the list
// isEqual to the specified object.
//
unsigned CmDLinkedList::occurrences(const CmObject* pObj) const
{
  if (!pObj || !_first) return 0;

  CmDLink *rover = _first;
  unsigned total = 0;
  while (rover)
  {
    if (rover->_data->isEqual(pObj)) total++;
    rover = rover->_next;
  }
  return total;
}


// "newIterator" creates and returns a list iterator.
//
CmIterator* CmDLinkedList::newIterator() const
{
  return new CmDLinkedListIterator(*this);
}


// "removeAbs" removes the specific object from the list.
//
bool CmDLinkedList::removeAbs(CmObject* pObj)
{
  if (!pObj || !_first) return false;

  CmDLink *rover = _first;
  bool     found = false;

  while (rover && !found)
  {
    if (rover->_data == pObj) found = true;
    else                      rover = rover->_next;
  }

  if (!found) return false;
  if (rover == _first) _first = rover->_next;
  if (rover == _last ) _last  = rover->_prev;
  if (rover->_prev) rover->_prev->_next = rover->_next;
  if (rover->_next) rover->_next->_prev = rover->_prev;
  if (ownsObjects()) delete rover->_data;
  delete rover;
  _size--;
  return true;
}


// "next" returns the current object in the list and increments the
// iterator to the next link.
//
CmObject* CmDLinkedListIterator::next()
{
  if (!_link) return NULL;
  CmObject *rval = _link->_data;
  _link = _link->_next;
  return rval;
}


// "previous" backs the iterator up one object and returns that object.
//
CmObject* CmDLinkedListIterator::previous()
{
  if (!_link) return NULL;
  CmObject *rval = _link->_data;
  _link = _link->_prev;
  return rval;
}
