// cmhash.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Hash table implementation.
// -----------------------------------------------------------------

#include <cm/include/cmhash.h>
#include <cm/include/cmio.h>

// Macro calls initialize the static type info data.
CMOBJECT_INIT(CmHashTable, CmContainer)
CMOBJECT_INIT(CmHashTableIterator, CmIterator)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmHashTable)


// "CmHashTable" is the default table constructor.
//
CmHashTable::CmHashTable(unsigned sz) : _total(0)
{
  _buckets = (sz > 0) ? new CmHashBucket[sz] : NULL;
  _size    = (_buckets) ? sz : 0;
}


// "CmHashTable" is the table copy constructor.
//
CmHashTable::CmHashTable(const CmHashTable& table) : _total(0)
{
  _buckets = (table._size > 0) ? new CmHashBucket[table._size] : NULL;
  _size    = (_buckets) ? table._size : 0;
  copy(table);
}


// "CmHashTable" constructs a table from the specified container.
//
CmHashTable::CmHashTable(const CmContainer& cont) : _total(0)
{
  _buckets = (cont.total() > 0) ? new CmHashBucket[cont.total()] : NULL;
  _size    = (_buckets) ? cont.total() : 0;
  copy(cont);
}


// "CmHashTable" is the table destructor.
//
CmHashTable::~CmHashTable()
{
  removeAll();
  delete[] _buckets;
}


// "=" assignment operator copies the contents of the specified
// table into this table.
//
CmHashTable& CmHashTable::operator=(const CmHashTable& table)
{
  if (&table != this)
  {
    removeAll();
    delete[] _buckets;
    _total   = 0;
    _buckets = new CmHashBucket[table._size];
    _size    = (_buckets != NULL) ? table._size : 0;
    copy(table);
  }
  return *this;
}


// "resize" resizes the hash table.
//
bool CmHashTable::resize(unsigned newsize)
{
  if (newsize <= 0) newsize = 1;

  unsigned      newtotal   = 0;
  CmHashBucket *newbuckets = new CmHashBucket[newsize];
  if (!newbuckets) return false;

  CmHashTableIterator iter(*this);
  while (!iter.done())
  {
    CmObject *pObj = iter.next();
    newbuckets[pObj->hash(newsize)].append(pObj);
    newtotal++;
  }

  delete[] _buckets;
  _size    = newsize;
  _buckets = newbuckets;
  _total   = newtotal;
  return true;
}


// "add" adds the specified object to the hash table.
//
bool CmHashTable::add(CmObject* pObj)
{
  if (!pObj) return false;
  bool success = _buckets[pObj->hash(_size)].append(pObj);
  if (success) _total++;
  return success;
}


// "remove" removes the first occurrence of an object that isEqual
// to the specified object from the table.
//
bool CmHashTable::remove(const CmObject* pObj)
{
  if (!pObj) return false;
  bool success = _buckets[pObj->hash(_size)].remove(pObj, ownsObjects());
  if (success) _total--;
  return success;
}


// "remove" removes the object pointed to by the iterator from the
// table.
//
bool CmHashTable::remove(const CmIterator& iter)
{
  int  ii      = 0;
  bool success = false;
  while (ii < _size && !success)
    success = _buckets[ii++].removeAbs(iter.current(), ownsObjects());
  return success;
}


// "removeAt" removes the object at the specified index.
//
bool CmHashTable::removeAt(int idx)
{
  CmObject *pObj = at(idx);
  if (!pObj) return false;

  int  ii      = 0;
  bool success = false;
  while (ii < _size && !success)
    success = _buckets[ii++].removeAbs(pObj, ownsObjects());
  return success;
}


// "lookup" returns the first occurrence of an object which isEqual
// to the specified object.
//
CmObject* CmHashTable::lookup(const CmObject* pObj) const
{
  if (!pObj) return NULL;
  return _buckets[pObj->hash(_size)].lookup(pObj);
}


// "contains" checks to see if an object that isEqual to the specified
// object exists in the table.
//
bool CmHashTable::contains(const CmObject* pObj) const
{
  if (!pObj) return false;
  return (_buckets[pObj->hash(_size)].lookup(pObj) == NULL) ? false : true;
}


// "occurrences" returns the number of objects in the table
// isEqual to the specified object.
//
unsigned CmHashTable::occurrences(const CmObject* pObj) const
{
  if (!pObj) return false;
  CmHashTableIterator iter(*this);
  unsigned            num = 0;
  while (!iter.done())
    if (iter.next()->isEqual(pObj)) num++;
  return num;
}


// "removeAll" removes all objects from the table.
//
void CmHashTable::removeAll()
{
  for (int ii = 0; ii < _size; ii++)
    _buckets[ii].removeAll(ownsObjects());
  _total = 0;
}


// "newIterator" creates and returns a new hash table iterator.
//
CmIterator* CmHashTable::newIterator() const
{
  return new CmHashTableIterator(*this);
}


// "write" writes the array objects to the specified file.
//
bool CmHashTable::write(CmIOFile& file) const
{
  if (!writeBase(file))    return false;            // Write base container.
  if (!file.write(_total)) return false;            // Write number of objects.

  bool success  = true;
  CmHashTableIterator iter(*this);
  while (!iter.done() && success)                   // For each object.
    success = CmIO::writeObject(file, iter.next()); // Write object.
  return success;
}


// "read" reads the array objects from the specified file.
//
bool CmHashTable::read(CmIOFile& file)
{
  unsigned size = readBase(file);                 // Read base container.

  int total;
  if (!file.read(total)) return false;            // Read number of objects.
  if (!resize(size))     return false;            // Resize the table.

  removeAll();

  bool success = true;
  int  ii      = 0;
  while (ii < total && success)                   // For each object,
  {
    CmObject *pObj = CmIO::readObject(file);      // Read object.
    if (pObj) success = add(pObj); ii++;          // Add to table.
  }
  return success;
}


// "append" adds a new object to the end of this hash bucket.
//
bool CmHashBucket::append(CmObject* pObj)
{
  CmHashNode *newnode = new CmHashNode(pObj);
  if (!newnode) return false;

  if (!_first)
    _first = _last = newnode;
  else
  {
    _last->_next = newnode;
    _last        = newnode;
  }
  _size++;
  return true;
}


// "remove" removes the first occurrence of an object that isEqual
// to the specified object from the bucket.
//
bool CmHashBucket::remove(const CmObject* pObj, bool delFlag)
{
  CmObject *temp = lookup(pObj);
  return (temp) ? removeAbs(temp, delFlag) : false;
}


// "removeAbs" removes the exact object from the bucket.
//
bool CmHashBucket::removeAbs(const CmObject* pObj, bool delFlag)
{
  if (!pObj || !_first) return false;

  if (_first->_data == pObj)                // Object is first in the list.
  {
    CmHashNode *temp = _first;
    if (_first == _last) _first = _last = NULL;
    else                 _first = _first->_next;
    _size--;
    if (delFlag) delete temp->_data;
    delete temp;
    return true;
  }

  CmHashNode *rover1 = _first;              // Search for object in list.
  CmHashNode *rover2 = _last;

  while (rover1 != NULL && rover1->_data != pObj)
  {
    rover2 = rover1;
    rover1 = rover1->_next;
  }

  if (rover1 == NULL) return false;         // Object was not found.

  if (rover1 == _last)                      // Object is last in list.
  {
    _last         = rover2;
    rover2->_next = NULL;
  }

  else                                      // Remove object from list.
    rover2->_next = rover2->_next->_next;

  if (delFlag) delete rover1->_data;        // Delete object (if requested)
  delete rover1;                            // and delete the hash node.
  return true;
}


// "removeAll" removes all objects from this list and deletes the objects
// if requested to do so.
//
void CmHashBucket::removeAll(bool delFlag)
{
  CmHashNode *temp, *rover = _first;
  while (rover)
  {
    temp = rover->_next;
    if (delFlag) delete rover->_data;
    delete rover;
    rover = temp;
  }
  _size  = 0;
  _first = _last = NULL;
}


// "lookup" returns the first occurrence of an object which isEqual
// to the specified object.
//
CmObject* CmHashBucket::lookup(const CmObject* pObj) const
{
  CmHashNode *rover = _first;
  bool        found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(pObj)) found = true;
    else                             rover = rover->_next;
  }

  return (found) ? rover->_data : NULL;
}


// "contains" checks to see if an object that isEqual to the specified
// object exists in the bucket.
//
bool CmHashBucket::contains(const CmObject* pObj) const
{
  CmHashNode *rover = _first;
  bool        found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(pObj)) found = true;
    else                             rover = rover->_next;
  }
  return found;
}


// "CmHashTableIterator" is the iterator constructor.
//
CmHashTableIterator::CmHashTableIterator(const CmHashTable& T)
                   : CmIterator(T), _bucket(0), _node(NULL)
{
  while (table()._buckets[_bucket]._size == 0 && _bucket < table()._size)
    _bucket++;
  if (_bucket < table()._size) _node = table()._buckets[_bucket]._first;
}


// "next" returns the current object pointed to by the iterator and
// then advances the iterator to the next object.
//
CmObject* CmHashTableIterator::next()
{
  if (!_node) return NULL;

  CmObject *pObj = _node->_data;

  if (_node->_next)
    _node = _node->_next;
  else
  {
    _bucket++;
    while (_bucket < table()._size && table()._buckets[_bucket]._size == 0)
      _bucket++;
    _node = (_bucket < table()._size) ? table()._buckets[_bucket]._first : NULL;
  }
  return pObj;
}


// "previous" decrements the iterator by one object and then returns
// that object.
//
CmObject* CmHashTableIterator::previous()
{
  if (!_node) return NULL;
  CmObject *rval = _node->_data;

  if (_node == table()._buckets[_bucket]._first)
  {
    --_bucket;
    while (_bucket >= 0 && table()._buckets[_bucket]._size == 0) _bucket--;
    _node = (_bucket >= 0) ? table()._buckets[_bucket]._last : NULL;
  }
  else
  {
    CmHashNode *rover = table()._buckets[_bucket]._first;
    while (rover && rover->_next != _node)
      rover = rover->_next;
    _node = rover;
  }
  return rval;
}


// "first" moves the iterator to the first object in the table.
//
void CmHashTableIterator::first()
{
  _bucket = 0;
  _node   = NULL;
  while (table()._buckets[_bucket]._size == 0 && _bucket < table()._size)
    _bucket++;
  if (_bucket < table()._size) _node = table()._buckets[_bucket]._first;
}


// "last" moves the iterator to the last object in the table.
//
void CmHashTableIterator::last()
{
  _bucket = table()._size - 1;
  _node   = NULL;
  while (_bucket >= 0 && table()._buckets[_bucket]._size == 0)
    _bucket--;
  if(_bucket >= 0) _node = table()._buckets[_bucket]._last;
}
