// cmhdict.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Hash dictionary implementation.
// -----------------------------------------------------------------

#include <cm/include/cmhdict.h>

// Macro call initializes the static type info data.
CMOBJECT_INIT(CmHashDictionary, CmHashTable)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmHashDictionary)


// "CmHashDictionary" is the default dictionary constructor.
//
CmHashDictionary::CmHashDictionary(unsigned sz)
                 : CmHashTable(sz)
{}


// "CmHashDictionary" is the dictionary copy constructor.
//
CmHashDictionary::CmHashDictionary(const CmHashDictionary& dict)
             : CmHashTable(dict)
{}


// "~CmHashDictionary" is the dictionary destructor.
//
CmHashDictionary::~CmHashDictionary()
{
  removeAll();
}


// "=" assignment operator copies the specified dictionary into
// this dictionary.
//
CmHashDictionary& CmHashDictionary::operator=(const CmHashDictionary& dict)
{
  return (CmHashDictionary&) CmHashTable::operator=(dict);
}


// "add" adds an association to the dictionary.
//
bool CmHashDictionary::add(CmObject* pObj)
{
  if (CmPtrCast(CmAssociation,pObj) && !CmHashTable::contains(pObj))
    return CmHashTable::add(pObj);
  return false;
}


// "add" adds a key/object pair to the dictionary.
//
bool CmHashDictionary::add(CmObject* pKey, CmObject* pObj)
{
  CmAssociation *pAssoc = new CmAssociation(pKey, pObj);
  if (!pAssoc) return false;

  if (CmHashTable::contains(pAssoc))
  {
    delete pAssoc;
    return false;
  }
  else
    return CmHashTable::add(pAssoc);
}


// "replace" looks for the first occurrence of an association with
// a key that isEqual to the specified key and replaces it's object
// value with the specified object.
//
bool CmHashDictionary::replace(const CmObject* pKey, CmObject* pObj)
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  if (!pAssoc) return false;

  if (ownsObjects()) delete pAssoc->object();
  pAssoc->object(pObj);
  return true;
}


// "remove" looks for the first occurrence of an association with
// a key that isEqual to the specified key and removes it.
//
bool CmHashDictionary::remove(const CmObject* pKey)
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  if (pAssoc)
  {
    CmObject* pKey    = pAssoc->key   ();
    CmObject* pObject = pAssoc->object();
    bool success = CmHashTable::remove(pAssoc);
    if (success && ownsObjects())
    {
      delete pKey;
      delete pObject;
    }
    return success;
  }
  return false;
}


// "remove" removes the association pointed to by the iterator.
//
bool CmHashDictionary::remove(const CmIterator& iter)
{
  return CmHashTable::remove(iter);
}


// "removeAll" removes all of the associations from the dictionary.
//
void CmHashDictionary::removeAll()
{
  if (ownsObjects())
  {
    CmHashTableIterator iter(*this);
    while (!iter.done())
    {
      CmAssociation *pAssoc = (CmAssociation*) iter.next();
      delete pAssoc->key   ();
      delete pAssoc->object();
    }
  }
  CmHashTable::removeAll();
}


// "lookup" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association's
// key value.
//
CmObject* CmHashDictionary::lookup(const CmObject* pKey) const
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  return (pAssoc) ? pAssoc->key() : NULL;
}


// "lookupObject" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association's
// object value.
//
CmObject* CmHashDictionary::lookupObject(const CmObject* pKey) const
{
  CmAssociation *pAssoc = lookupAssoc(pKey);
  return (pAssoc) ? pAssoc->object() : NULL;
}


// "lookupObject" looks for the first occurrence of an association with
// a key that isEqual to the specified key and returns that association.
//
CmAssociation* CmHashDictionary::lookupAssoc(const CmObject* pKey) const
{
  CmAssociation assoc((CmObject*) pKey, NULL);
  return (CmAssociation*) CmHashTable::lookup(&assoc);
}


// "contains" checks to see is an association exists with a key which
// isEqual to the specified key.
//
bool CmHashDictionary::contains(const CmObject* pKey) const
{
  return (lookupAssoc(pKey) == NULL) ? false : true;
}


// "occurrences" returns the number of associations with a key which
// isEqual to the specified key.  (This should always be 1).
//
unsigned CmHashDictionary::occurrences(const CmObject* pKey) const
{
  CmAssociation assoc((CmObject*) pKey, NULL);
  return CmHashTable::occurrences(&assoc);
}
