// cmio.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Object I/O management class implementation.
// -----------------------------------------------------------------

#include <cm/include/cmio.h>

class CmRegistry : public CmObject {           // Class registry definition.
public:
  CmRegistry(const char*, CmCreateFunc);       // Registry constructor.
  CmRegistry(const char*);                     // Registry constructor.
 ~CmRegistry() { delete[] _name; }             // Registry destructor.

  bool isEqual(const CmObject*) const;         // Compare names.
  int  compare(const CmObject*) const;         // Compare names.

  CMOBJECT_DEFINE(CmRegistry)                  // Define base.

  char*        _name;                          // Class name.
  CmCreateFunc _func;                          // Create function.
};

class CmObjRef : public CmObject {             // Object reference definition.
public:
  CmObjRef(int k)                              // Construct with key.
          : _key(k), _object(NULL) {}

  CmObjRef(CmObject* o)                        // Construct with object.
          : _key(0), _object(o) {}

  CmObjRef(int k, CmObject* o)                 // Construct with both.
          : _key(k), _object(o) {}

  bool isEqual(const CmObject*) const;         // Reference compare.
  int  compare(const CmObject*) const;         // Reference compare.

  CMOBJECT_DEFINE(CmObjRef)                    // Define base functions.

  enum { KEYS=0, OBJECTS=1 };                  // Compare type values.

  int        _key;                             // Key value.
  CmObject  *_object;                          // Object pointer.
  static int _compareBy;                       // Compare policy.
};

// Macros initialize static type info.
CMOBJECT_INIT(CmRegistry, CmObject)
CMOBJECT_INIT(CmObjRef, CmObject)

// Initialize the object reference compare policy.
int CmObjRef::_compareBy = CmObjRef::OBJECTS;

// Initialize the IO class registry.
CmBTree* CmIO::_classes = NULL;

// Initialize the IO reference table.
CmBTree ___cmGlobal_Reference_Table;
CmBTree* CmIO::_table = &___cmGlobal_Reference_Table;

// Initialize the IO reference flag.
bool CmIO::_references = true;


// "CmRegistry" is the registry constructor.
//
CmRegistry::CmRegistry(const char* name, CmCreateFunc func)
{
  _name = (name) ? new char[strlen(name)+1] : NULL;
  if (_name) strcpy(_name, name);
  _func = func;
}


// "CmRegistry" constructs a registry using just the name.
//
CmRegistry::CmRegistry(const char* name)
{
  _name = (name) ? new char[strlen(name)+1] : NULL;
  if (_name) strcpy(_name, name);
}


// "isEqual" compares the names of this registry and the specified
// registry.
//
bool CmRegistry::isEqual(const CmObject* pObj) const
{
  CmRegistry *pReg = CmPtrCast(CmRegistry, pObj);
  return (pReg) ? ((!strcmp(_name, pReg->_name)) ? true : false) : false;
}


// "compare" compares the names of this registry and the specified
// registry.
//
int CmRegistry::compare(const CmObject* pObj) const
{
  CmRegistry *pReg = CmPtrCast(CmRegistry, pObj);
  return (pReg) ? strcmp(_name, pReg->_name) : -1;
}


// "isEqual" compares the specified object with this reference using
// the current compare policy.
//
bool CmObjRef::isEqual(const CmObject* pObj) const
{
  CmObjRef *pRef = CmPtrCast(CmObjRef, pObj);
  if (!pRef) return false;
  if (CmObjRef::_compareBy == CmObjRef::KEYS)
    return (_key == pRef->_key) ? true : false;
  else
    return (_object == pRef->_object) ? true : false;
}


// "compare" compares the specified object with this reference using
// the current compare policy.
//
int CmObjRef::compare(const CmObject* pObj) const
{
  CmObjRef *pRef = CmPtrCast(CmObjRef, pObj);
  if (!pRef) return -1;
  if (CmObjRef::_compareBy == CmObjRef::KEYS)
    return ((_key == pRef->_key) ? 0 : ((_key > pRef->_key) ? 1 : -1));
  else
    return ((_object==pRef->_object) ? 0 : ((_object>pRef->_object) ? 1 : -1));
}


// "CmRegisterClass" calls the static function to register the class
// name and corresponding object create function with the global registry.
//
void CmRegisterClass(const char* name, CmCreateFunc func)
{
  CmIO::registerClass(name, func);
}


// "CmUnRegister" calls the static function to unregister the named
// class from the global registry.
//
void CmUnRegister(const char* name)
{
  CmIO::unRegister(name);
}


// "registerClass" registers the class name and corresponding object
// create function with the global registry.
//
void CmIO::registerClass(const char* name, CmCreateFunc func)
{
  if (!(CmIO::_classes)) CmIO::_classes = new CmBTree;
  CmIO::_classes->add(new CmRegistry(name, func));
}


// "unRegister" unregisters the named class with the global registry.
//
void CmIO::unRegister(const char* name)
{
  CmRegistry temp(name);
  CmIO::_classes->remove(&temp);
}


// "reset" is called to reset the global reference table after
// I/O is done.
//
void CmIO::reset()
{
  CmIO::_table->removeAll();
}


// "writeObject" writes the specified object to the specified file
// and updates the global reference table.
//
bool CmIO::writeObject(CmIOFile& file, CmObject* obj)
{
  if (!file.write(CmPtrTypeId(obj).name()))    // Write class name.
    return false;

  int key, flag = CmIO::OBJECT;
  if (CmIO::maintainReferences())
  {
    key  = CmIO::getFromTable(obj);            // See if object is in table.
    flag = (key >= 0) ? CmIO::REFERENCE : CmIO::OBJECT;
  }
  bool success = file.write(flag);             // Write reference flag.

  if (success && flag == CmIO::REFERENCE)      // If object was written,
    success = file.write(key);                 // write table key.
  else if (success)                            // Object was not written.
  {
    if (CmIO::maintainReferences())            // Add to reference table.
      success = CmIO::addToTable(obj, CmObjRef::OBJECTS);
    if (success) success = obj->write(file);   // Write object.
  }
  return success;
}


// "readObject" creates and reads the next object from the specified
// file.
//
CmObject* CmIO::readObject(CmIOFile& file)
{
  char *letters = file.read();                 // Read class name.
  if (!letters) return NULL;

  int flag;
  if (!file.read(flag))                        // Read reference flag.
  {
    delete[] letters;
    return NULL;
  }

  CmObject *pObj = NULL;
  if (flag == CmIO::REFERENCE)                 // If object was reference,
  {
    int key;
    if (file.read(key))                        // Read table index.
      pObj = CmIO::getFromTable(key);          // Get pointer from table.
  }
  else
  {
    pObj = CmIO::createObject(letters);        // Create new object.
    if (pObj)
    {
      bool success = pObj->read(file);         // Read object data.
      if (!success)
      {
        delete pObj;
        pObj = NULL;
      }
      else
        CmIO::addToTable(pObj, CmObjRef::KEYS); // Add to reference table.
    }
  }
  delete[] letters;
  return pObj;                                 // Return new object.
}


// "createObject" gets the object create function from the global
// registry based on the specified class name, creates a new object,
// and returns it.
//
CmObject* CmIO::createObject(const char* name)
{
  CmRegistry temp(name);
  CmRegistry *pReg = (CmRegistry*) CmIO::_classes->lookup(&temp);
  if (!pReg) return NULL;

  CmCreateFunc func = pReg->_func;
  return (*func)();
}


// "addToTable" adds the specified object to the reference table.
//
bool CmIO::addToTable(CmObject* pObj, int policy)
{
  int key = CmIO::_table->size();
  CmObjRef *pRef = new CmObjRef(key, pObj);
  CmObjRef::_compareBy = policy;
  return CmIO::_table->add(pRef);
}


// "getFromTable" returns the object in the reference table associated
// with the specified key.
//
CmObject* CmIO::getFromTable(int key)
{
  CmObjRef::_compareBy = CmObjRef::KEYS;
  CmObjRef temp(key);
  CmObjRef* pRef = (CmObjRef*) CmIO::_table->lookup(&temp);
  return (pRef) ? pRef->_object : NULL;
}


// "getFromTable" returns the key in the reference table associated
// with the specified object.
//
int CmIO::getFromTable(CmObject* pObj)
{
  CmObjRef::_compareBy = CmObjRef::OBJECTS;
  CmObjRef temp(pObj);
  CmObjRef* pRef = (CmObjRef*) CmIO::_table->lookup(&temp);
  return (pRef) ? pRef->_key : -1;
}
