// cmiofile.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Binary file interface implementation.
// -----------------------------------------------------------------

#include <string.h>
#include <cm/include/cmiofile.h>


// Define a file buffer size of 16k.
//
#define CM_BUF_SIZE 16384


// "CmIOFile" is the default file constructor.
//
CmIOFile::CmIOFile() : _name(NULL), _file(NULL)
{}


// "CmIOFile" is the file constructor.  The file is opened/created at
// construct time and the error parameter signals success or failure.
//
CmIOFile::CmIOFile(const char* filename)
{
  size_t lnth = (filename) ? strlen(filename) : 0;

  _file = NULL;
  _name = (lnth) ? new char[lnth+1] : NULL; 

  if (_name)
  {
    strcpy(_name, filename);
    _file = fopen(_name, "rb+");
    if (!_file) _file = fopen(_name, "wb+");
    if (_file) setvbuf(_file, NULL, _IOFBF, CM_BUF_SIZE);
  }
}


// "CmIOFile" is the file copy constructor.  An invalid file
// is created.
//
CmIOFile::CmIOFile(const CmIOFile& file)
{
  size_t lnth = (file._name) ? strlen(file._name) : 0;
  _name = (lnth) ? new char[lnth+1] : NULL;
  if (_name) strcpy(_name, file._name);
  _file = NULL;
}


// "~CmIOFile" is the file destructor.  The file is closed at
// destruct time.
//
CmIOFile::~CmIOFile()
{
  delete[] _name;
  if (_file) fclose(_file);
}


// "=" assignment operator copies the file name from the specified
// file.  However, an invalid file is created.
//
CmIOFile& CmIOFile::operator=(const CmIOFile& file)
{
  if (&file != this)
  {
    if (_file) fclose(_file);
    delete[] _name;
    size_t lnth = (file._name) ? strlen(file._name) : 0;
    _name = (lnth) ? new char[lnth+1] : NULL;
    if (_name) strcpy(_name, file._name);
    _file = NULL;
  }
  return *this;
}


// "open" opens the named file.
//
bool CmIOFile::open(const char* filename)
{
  if (_file) return false;

  if (_name) delete[] _name;
  _name = (strlen(filename) > 0) ? new char[strlen(filename)+1] : NULL;

  if (_name)
  {
    strcpy(_name, filename);
    _file = fopen(_name, "rb+");
    if (!_file) _file = fopen(_name, "wb+");
    if (_file) setvbuf(_file, NULL, _IOFBF, CM_BUF_SIZE);
  }
  return (_file) ? true : false;
}


// "close" closes the file.
//
bool CmIOFile::close()
{
  int out = 1;
  if (_file)
  {
    out = fclose(_file);
    _file = NULL;
  }
  return (out == 0) ? true : false;
}


// "remove" closes and removes this file from the disk.
//
bool CmIOFile::remove()
{
  if (!_file) return false;
  close();
  return (::remove(_name) == 0) ? true : false;
}


// "rename" renames this file.
//
bool CmIOFile::rename(const char* newname)
{
  if (::rename(_name, newname) != 0) return false;
  delete[] _name;
  _name = (strlen(newname) > 0) ? new char[strlen(newname)+1] : NULL;
  if (_name) strcpy(_name, newname);
  return (_name) ? true : false;
}


// "read" is the direct input function for this file.
//
size_t CmIOFile::read(void* ptr, size_t size, size_t nobj)
{
  return (_file) ? fread(ptr, size, nobj, _file) : 0;
}


// "read" is a direct input function reading only one object.
//
bool CmIOFile::read(void* ptr, size_t size)
{
  return (!_file) ? false : ((fread(ptr, size, 1, _file) == 1) ? true : false);
}


// "write" is the direct output function for this file.
//
size_t CmIOFile::write(void* ptr, size_t size, size_t nobj)
{
  return (_file) ? fwrite(ptr, size, nobj, _file) : 0;
}


// "write" is a direct output function writing only one object.
//
bool CmIOFile::write(void* ptr, size_t size)
{
  return (!_file) ? false : ((fwrite(ptr, size, 1, _file) == 1) ? true : false);
}


// "read" reads one character from the file.
//
bool CmIOFile::read(char& c)
{
  return read((void*) &c, sizeof(char));
}


// "read" reads one short from the file.
//
bool CmIOFile::read(short& s)
{
  return read((void*) &s, sizeof(short));
}


// "read" reads one integer from the file.
//
bool CmIOFile::read(int& i)
{
  return read((void*) &i, sizeof(int));
}


// "read" reads one unsigned character from the file.
//
bool CmIOFile::read(long& l)
{
  return read((void*) &l, sizeof(long));
}


// "read" reads one unsigned character from the file.
//
bool CmIOFile::read(unsigned char& uc)
{
  return read((void*) &uc, sizeof(unsigned char));
}


// "read" reads one unsigned character short the file.
//
bool CmIOFile::read(unsigned short& us)
{
  return read((void*) &us, sizeof(unsigned short));
}


// "read" reads one unsigned int from the file.
//
bool CmIOFile::read(unsigned int& ui)
{
  return read((void*) &ui, sizeof(unsigned int));
}


// "read" reads one unsigned long from the file.
//
bool CmIOFile::read(unsigned long& ul)
{
  return read((void*) &ul, sizeof(unsigned long));
}


// "read" reads one float from the file.
//
bool CmIOFile::read(float& f)
{
  return read((void*) &f, sizeof(float));
}


// "read" reads one double from the file.
//
bool CmIOFile::read(double& d)
{
  return read((void*) &d, sizeof(double));
}


// "read" reads an array of characters from the file.
//
bool CmIOFile::read(char* c, unsigned u)
{
  return (read(c, sizeof(char), u) == u);
}


// "read" reads an array of shorts from the file.
//
bool CmIOFile::read(short* s, unsigned u)
{
  return (read(s, sizeof(short), u) == u);
}


// "read" reads an array of integers from the file.
//
bool CmIOFile::read(int* i, unsigned u)
{
  return (read(i, sizeof(int), u) == u);
}


// "read" reads an array of longs from the file.
//
bool CmIOFile::read(long* l, unsigned u)
{
  return (read(l, sizeof(long), u) == u);
}


// "read" reads an array of floats from the file.
//
bool CmIOFile::read(float* f, unsigned u)
{
  return (read(f, sizeof(float), u) == u);
}


// "read" reads an array of doubles from the file.
//
bool CmIOFile::read(double* d, unsigned u)
{
  return (read(d, sizeof(double), u) == u);
}


// "write" writes one character to the file.
//
bool CmIOFile::write(char c)
{
  return write((void*) &c, sizeof(char));
}


// "write" writes one short to the file.
//
bool CmIOFile::write(short s)
{
  return write((void*) &s, sizeof(short));
}


// "write" writes one integer to the file.
//
bool CmIOFile::write(int i)
{
  return write((void*) &i, sizeof(int));
}


// "write" writes one long to the file.
//
bool CmIOFile::write(long l)
{
  return write((void*) &l, sizeof(long));
}


// "write" writes one unsigned character to the file.
//
bool CmIOFile::write(unsigned char uc)
{
  return write((void*) &uc, sizeof(unsigned char));
}


// "write" writes one unsigned short to the file.
//
bool CmIOFile::write(unsigned short us)
{
  return write((void*) &us, sizeof(unsigned short));
}


// "write" writes one unsigned integer to the file.
//
bool CmIOFile::write(unsigned int ui)
{
  return write((void*) &ui, sizeof(unsigned int));
}


// "write" writes one unsigned long to the file.
//
bool CmIOFile::write(unsigned long ul)
{
  return write((void*) &ul, sizeof(unsigned long));
}


// "write" writes one float to the file.
//
bool CmIOFile::write(float f)
{
  return write((void*) &f, sizeof(float));
}


// "write" writes one double to the file.
//
bool CmIOFile::write(double d)
{
  return write((void*) &d, sizeof(double));
}


// "write" writes an array of characters to the file.
//
bool CmIOFile::write(char* c, unsigned u)
{
  return (write(c, sizeof(char), u) == u);
}


// "write" writes an array of shorts to the file.
//
bool CmIOFile::write(short* s, unsigned u)
{
  return (write(s, sizeof(short), u) == u);
}


// "write" writes an array of integers to the file.
//
bool CmIOFile::write(int* i, unsigned u)
{
  return (write(i, sizeof(int), u) == u);
}


// "write" writes an array of longs to the file.
//
bool CmIOFile::write(long* l, unsigned u)
{
  return (write(l, sizeof(long), u) == u);
}


// "write" writes an array of floats to the file.
//
bool CmIOFile::write(float* f, unsigned u)
{
  return (write(f, sizeof(float), u) == u);
}


// "write" writes an array of doubles to the file.
//
bool CmIOFile::write(double* d, unsigned u)
{
  return (write(d, sizeof(double), u) == u);
}


#if !defined(NO_STANDARD_BOOL)

// "read" reads one boolean from the file.
//
bool CmIOFile::read(bool& b)
{
  return read((void*) &b, sizeof(bool));
}


// "read" reads an array of booleans from the file.
//
bool CmIOFile::read(bool* b, unsigned u)
{
  return (read(b, sizeof(bool), u) == u);
}


// "write" writes one boolean to the file.
//
bool CmIOFile::write(bool b)
{
  return write((void*) &b, sizeof(bool));
}


// "write" writes an array of booleans to the file.
//
bool CmIOFile::write(bool* b, unsigned u)
{
  return (write(b, sizeof(bool), u) == u);
}

#endif


// "write" writes the specified character string length and the
// string text to the file.
//
bool CmIOFile::write(const char* s)
{
  int lnth = (s) ? strlen(s) + 1 : 0;
  if (!write(lnth)) return false;
  return (lnth>0) ? write((char*) s, (unsigned) lnth) : true;
}


// "read" reads a string length and string text from the file.  The
// output string must be deleted when use is finished.
//
char* CmIOFile::read()
{
  int lnth;
  if (!read(lnth)) return NULL;
  char* s = (lnth > 0) ? new char[lnth] : NULL;
  if (s) read(s, (unsigned) lnth);
  return s;
}


// "seek" sets the file position for this file.
//
int CmIOFile::seek(long offset, int origin)
{
  return (_file) ? fseek(_file, offset, origin) : -1;
}


// "tell" returns the current file position for this file.
//
long CmIOFile::tell()
{
  return (_file) ? ftell(_file) : -1L;
}


// "rewind" rewinds the file pointer to the beginning of the file.
//
void CmIOFile::rewind()
{
  if (_file) ::rewind(_file);
}


// "getPos" records the current file position in ptr.
//
int CmIOFile::getPos(fpos_t* ptr)
{
  return (_file) ? fgetpos(_file, ptr) : -1;
}


// "setPos" sets the file position to the position recorded by "getPos"
// in ptr.
//
int CmIOFile::setPos(const fpos_t* ptr)
{
  return (_file) ? fsetpos(_file, ptr) : -1;
}


// "clearError" clears all error indicators for this file.
//
void CmIOFile::clearError()
{
  if (_file) clearerr(_file);
}


// "eof" checks if the end of file indicator is set for this file.
//
bool CmIOFile::eof()
{
  return (!_file) ? -1 : ((feof(_file) != 0) ? true : false);
}


// "error" checks if any error indicators are set for this file.
//
bool CmIOFile::error()
{
  return (!_file) ? -1 : ((ferror(_file) != 0) ? true : false);
}


// "printError" prints the error message associated with the error
// indicator in the file and also the specified message.
//
void CmIOFile::printError(const char* s)
{
  perror(s);
}


// "exists" checks if the specified file exists on the disk.
//
bool CmIOFile::exists(const char* filename)
{
  FILE *fp = fopen(filename, "r");
  if (!fp) return false;
  fclose(fp);
  return true;
}


// "remove" removes the named file from the disk.
//
bool CmIOFile::remove(const char* name)
{
  return (::remove(name) == 0) ? true : false;
}


// "tempName" returns a file name which does not currently exist on the
// disk for use as a temporary file name.
//
char* CmIOFile::tempName(char* buff)
{
  return tmpnam(buff);
}
