// cmlist.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Singly linked list implementation.
// -----------------------------------------------------------------

#include <cm/include/cmlist.h>

// Macro calls initialize the static type info data.
CMOBJECT_INIT(CmLinkedList, CmContainer)
CMOBJECT_INIT(CmLinkedListIterator, CmIterator)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmLinkedList)


// "CmLinkedList" is the list copy constructor.
//
CmLinkedList::CmLinkedList(const CmLinkedList& list)
            : _first(NULL), _last(NULL)
{
  copy(list);
}


// "CmLinkedList" constructs a list from the specified container.
//
CmLinkedList::CmLinkedList(const CmContainer& cont)
            : _first(NULL), _last(NULL)
{
  copy(cont);
}


// "~CmLinkedList" is the list destructor.
//
CmLinkedList::~CmLinkedList()
{
  removeAll();
}


// "=" assignment operator copies the contents of the specified list
// into this list.
//
CmLinkedList& CmLinkedList::operator=(const CmLinkedList& list)
{
  if (&list != this)
  {
    removeAll();
    copy     (list);
  }
  return *this;
}


// "add" adds the specified object to the end of this list.
//
bool CmLinkedList::add(CmObject* pObj)
{
  return append(pObj);
}


// "append" adds the specified object to the end of this list.
//
bool CmLinkedList::append(CmObject* pObj)
{
  if (!pObj) return false;
  CmLink *newlink = new CmLink(pObj);
  if (!newlink) return false;

  if (!_first)
    _first = _last = newlink;
  else
  {
    _last->_next = newlink;
    _last        = newlink;
  }
  _size++;
  return true;
}


// "prepend" adds the specified object to the start of this list.
//
bool CmLinkedList::prepend(CmObject* pObj)
{
  if (!pObj) return false;
  CmLink *newlink = new CmLink(pObj);
  if (!newlink) return false;

  if (!_first)
    _first = _last = newlink;
  else
  {
    newlink->_next = _first;
    _first         = newlink;
  }
  _size++;
  return true;
}


// "insertAt" inserts the specified object at the specified index.
//
bool CmLinkedList::insertAt(int idx, CmObject* pObj)
{
  if (idx < 0 || idx >= _size) return false;
  CmLinkedListIterator iter(*this);
  int                  ii = 0;
  while (!iter.done() && ii < idx)
  {
    iter.next();
    ii++;
  }
  return insertBefore(iter, pObj);
}


// "replaceAt" replaces the object at the specified index with the
// specified object.
//
bool CmLinkedList::replaceAt(int idx, CmObject* pObj)
{
  if (idx < 0 || idx >= _size) return false;

  int     ii    = 0;
  CmLink *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  if (!rover) return false;
  if (ownsObjects()) delete rover->_data;
  rover->_data = pObj;
  return true;
}


// "remove" removes the first occurrence of an object that isEqual
// to the specified object from the list.
//
bool CmLinkedList::remove(const CmObject* pObj)
{
  if (!pObj || !_first) return false;
  CmObject *remObj = lookup(pObj);
  return (remObj) ? removeAbs(remObj) : false;
}


// "remove" removes the object pointed to by the specified iterator
// object.
//
bool CmLinkedList::remove(const CmIterator& iter)
{
  CmObject *remObj = iter.current();
  return (remObj) ? removeAbs(remObj) : false;
}


// "removeAt" removes the object at the specified list index.
//
bool CmLinkedList::removeAt(int idx)
{
  CmObject *remObj = at(idx);
  return (remObj) ? removeAbs(remObj) : false;
}


// "removeAll" removes all the objects from the list.
//
void CmLinkedList::removeAll()
{
  CmLink *temp, *rover = _first;
  while (rover)
  {
    temp = rover->_next;
    if (ownsObjects()) delete rover->_data;
    delete rover;
    rover = temp;
  }
  _size  = 0;
  _first = _last = NULL;
}


// "removeFirst" removes the first object from the list.  The object is
// not deleted regardless of the ownership.
//
CmObject* CmLinkedList::removeFirst()
{
  if (!_first) return NULL;

  CmObject *pObj = _first->_data;
  CmLink   *temp = _first;
  _first = _first->_next;
  if (!_first) _last = NULL;
  delete temp;
  _size--;
  return pObj;
}


// "removeLast" removes the last object from the list.  The object is
// not deleted regardless of the ownership.
//
CmObject* CmLinkedList::removeLast()
{
  if (!_last) return NULL;
  if (_last == _first) return removeFirst();

  CmLink *rover = _first;
  while (rover->_next->_next)
    rover = rover->_next;
  CmObject *pObj = rover->_next->_data;
  delete rover->_next;
  _last = rover;
  _last->_next = NULL;
  _size--;
  return pObj;
}


// "insertAfter" searches for the first occurrence of an object that
// isEqual to the first specified object and inserts the second
// specified object immediately after.
//
bool CmLinkedList::insertAfter(const CmObject* after, CmObject* pObj)
{
  if (!after || !pObj || !_first) return false;

  CmLink *rover = _first;
  bool    found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(after)) found = true;
    else                              rover = rover->_next;
  }

  if (!found) return false;

  CmLink *newlink = new CmLink(pObj);
  if (!newlink) return false;
  newlink->_next = rover->_next;
  rover->_next = newlink;
  if (rover == _last) _last = newlink;
  _size++;
  return true;
}


// "insertAfter" searches for the object pointed to by the specified
// iterator and inserts the specified object immediately after.
//
bool CmLinkedList::insertAfter(const CmIterator& iter, CmObject* pObj)
{
  if (!iter.current() || !pObj || !_first) return false;

  CmObject *after = iter.current();
  CmLink   *rover = _first;
  bool      found = false;
  while (rover && !found)
  {
    if (rover->_data == after) found = true;
    else                       rover = rover->_next;
  }
  if (!found) return false;

  CmLink *newlink = new CmLink(pObj);
  if (!newlink) return false;
  newlink->_next = rover->_next;
  rover->_next = newlink;
  if (rover == _last) _last = newlink;
  _size++;
  return true;
}


// "insertBefore" searches for the first occurrence of an object that
// isEqual to the first specified object and inserts the second
// specified object immediately before.
//
bool CmLinkedList::insertBefore(const CmObject* before, CmObject* pObj)
{
  if (!before || !pObj || !_first) return false;

  CmLink *rover1 = _first;
  CmLink *rover2 = _first;

  while (rover1 != NULL && !(rover1->_data->isEqual(before))) 
  {
    rover2 = rover1;
    rover1 = rover1->_next;
  }

  if (rover1 == NULL) return false;

  CmLink *newlink = new CmLink(pObj);
  if (rover1 == rover2)
  {
    newlink->_next = _first;
    _first         = newlink;
  }
  else
  {
    newlink->_next = rover2->_next;
    rover2->_next  = newlink;
  }
  _size++;
  return true;
}


// "insertBefore" searches for the object pointed to by the specified
// iterator and inserts the specified object immediately before.
//
bool CmLinkedList::insertBefore(const CmIterator& iter, CmObject* pObj)
{
  if (!iter.current() || !pObj || !_first) return false;

  CmObject *before = iter.current();
  CmLink   *rover1 = _first;
  CmLink   *rover2 = _first;

  while (rover1 != NULL && rover1->_data != before)
  {
    rover2 = rover1;
    rover1 = rover1->_next;
  }
  if (rover1 == NULL) return false;

  CmLink *newlink = new CmLink(pObj);
  if (rover1 == rover2)
  {
    newlink->_next = _first;
    _first         = newlink;
  }
  else
  {
    newlink->_next = rover2->_next;
    rover2->_next  = newlink;
  }
  _size++;
  return true;
}


// "index" returns the index in this list of the first occurrence of
// an object that isEqual to the specified object.
//
int CmLinkedList::index(const CmObject* pObj) const
{
  CmLink *rover  = _first;
  int     outidx = -1, idx = 0;
  while (outidx < 0 && rover)
  {
    if (rover->_data->isEqual(pObj)) outidx = idx;
    rover = rover->_next;
    idx++;
  }
  return outidx;
}


// "index" returns the index in the list of the object pointed to by
// the specified iterator.
//
int CmLinkedList::index(const CmIterator& iter) const
{
  CmLink *rover  = _first;
  int     outidx = -1, idx = 0;
  while (outidx < 0 && rover)
  {
    if (rover->_data == iter.current()) outidx = idx;
    rover = rover->_next;
    idx++;
  }
  return outidx;
}


// "at" returns the object at the given list index.
//
CmObject* CmLinkedList::at(int idx) const
{
  if (idx < 0 || idx >= _size) return NULL;

  int     ii    = 0;
  CmLink *rover = _first;
  while (rover && ii < idx)
  {
    rover = rover->_next;
    ii++;
  }
  return (rover) ? rover->_data : NULL;
}


// "lookup" returns the first occurrence of an object which isEqual
// to the specified object.
//
CmObject* CmLinkedList::lookup(const CmObject* pObj) const
{
  if (!pObj || !_first) return NULL;

  CmLink *rover = _first;
  bool    found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(pObj)) found = true;
    else                             rover = rover->_next;
  }
  return (found) ? rover->_data : NULL;
}


// "contains" checks to see if an object that isEqual to the specified
// object exists in the list.
//
bool CmLinkedList::contains(const CmObject* pObj) const
{
  if (!pObj || !_first) return false;

  CmLink *rover = _first;
  bool    found = false;
  while (rover && !found)
  {
    if (rover->_data->isEqual(pObj)) found = true;
    else                             rover = rover->_next;
  }
  return found;
}


// "occurrences" returns the number of objects in the list
// isEqual to the specified object.
//
unsigned CmLinkedList::occurrences(const CmObject* pObj) const
{
  if (!pObj || !_first) return 0;

  CmLink   *rover = _first;
  unsigned  total = 0;
  while (rover)
  {
    if (rover->_data->isEqual(pObj)) total++;
    rover = rover->_next;
  }
  return total;
}


// "newIterator" creates and returns a list iterator.
//
CmIterator* CmLinkedList::newIterator() const
{
  return new CmLinkedListIterator(*this);
}


// "removeAbs" removes the specific object from the list.
//
bool CmLinkedList::removeAbs(CmObject* pObj)
{
  if (!pObj || !_first) return false;   // Return if no objects.

  if (_first->_data == pObj)            // Object to remove is first.
  {
    CmObject* pObj = removeFirst();
    if (!pObj) return false;
    if (ownsObjects()) delete pObj;
    return true;
  }

  CmLink *rover1 = _first;              // Search for object to remove.
  CmLink *rover2 = _first;

  while (rover1 != NULL && rover1->_data != pObj)
  {
    rover2 = rover1;
    rover1 = rover1->_next;
  }

  if (rover1 == NULL) return false;     // Object was not found.

  if (rover1 == _last)                  // Object was last in list.
  {
    CmObject* pObj = removeLast();
    if (!pObj) return false;
    if (ownsObjects()) delete pObj;
    return true;
  }

  rover2->_next = rover2->_next->_next;     // Remove object from list.

  if (ownsObjects()) delete rover1->_data;  // Delete if object if owned.
  delete rover1;                            // Delete link.
  _size--;
  return true;
}


// "next" returns the current object in the list and increments the
// iterator to the next link.
//
CmObject* CmLinkedListIterator::next()
{
  if (!_link) return NULL;
  CmObject *rval = _link->_data;
  _link = _link->_next;
  return rval;
}


// "previous" backs the iterator up one object and returns that object.
//
CmObject* CmLinkedListIterator::previous()
{
  if (!_link) return NULL;
  CmObject *rval = _link->_data;
  if (_link == list()._first)
    _link = NULL;
  else
  {
    CmLink *rover = list()._first;
    while (rover && rover->_next != _link)
      rover = rover->_next;
    _link = rover;
  }
  return rval;
}
