// cmreserv.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// In-memory data base implementation.
// -----------------------------------------------------------------

#include <cm/include/cmreserv.h>
#include <cm/include/cmio.h>
#include <cm/include/cmlist.h>


// "CmReserve" is the reserve constructor.
//
CmReserve::CmReserve(const char* name)
         : _name(name), _identifier("CMRESERVE FILE")
{
  _roots = new CmBTreeDictionary;
}


// "CmReserve" is the reserve copy constructor.
//
CmReserve::CmReserve(const CmReserve& R)
         : _name(R._name), _identifier(R._identifier)
{
  _roots = new CmBTreeDictionary(*(R._roots));
}


// "~CmReserve" is the reserve destructor.
//
CmReserve::~CmReserve()
{
  delete _roots;
}


// "=" assignment operator copies the entire contents of the specified
// reserve into this reserve.
//
CmReserve& CmReserve::operator=(const CmReserve& R)
{
  if (&R != this)
  {
    delete _roots;
    _name       = R._name;
    _identifier = R._identifier;
    _roots      = new CmBTreeDictionary(*(R._roots));
  }
  return *this;
}


// "exists" returns true if the file already exists on disk.
//
bool CmReserve::exists() const
{
  return (CmIOFile::exists((const char*) _name));
}


// "addRoot" adds the specified root label and root to the root list.
//
bool CmReserve::addRoot(const char* name, CmContainer* pCont)
{
  if (!pCont) return false;
  CmString temp(name);
  if (_roots->contains(&temp)) return false;
  return _roots->add(new CmString(temp), pCont);
}


// "createRoot" creates a root container with the specified root name
// and stuffs it into the root list.
//
CmContainer* CmReserve::createRoot(const char* name)
{
  CmContainer *pList = getRoot(name);
  if (!pList)
  {
    pList = new CmLinkedList;
    if (!pList) return NULL;
    if (!_roots->add(new CmString(name), pList))
    {
      delete pList;
      pList = NULL;
    }
  }
  return pList;
}


// "totalRoots" returns the total number of roots.
//
int CmReserve::totalRoots() const
{
  return _roots->size();
}


// "containsRoot" checks if a root with the given name exists in
// the reserve.
//
bool CmReserve::containsRoot(const char* name) const
{
  return (getRoot(name)) ? true : false;
}


// "getRoot" finds the root at the given index in the root list and
// returns a pointer to it.
//
CmContainer* CmReserve::getRoot(const char* name) const
{
  CmString temp(name);
  return (CmContainer*) _roots->lookupObject(&temp);
}


// "removeRoot" removes and destroys the root with the specified
// name in the root list.
//
bool CmReserve::removeRoot(const char* name)
{
  CmString temp(name);
  return _roots->remove(&temp);
}


// "removeAll" removes all roots from the reserve.
//
void CmReserve::removeAll()
{
  _roots->removeAll();
}


// "write" writes all the objects in the reserve to an output file
// name with the reserve name.
//
bool CmReserve::write() const
{
  CmIOFile file((const char*) _name);                   // Open the file.
  if (!file) return false;

  bool success = file.write((const char*) _identifier); // Write identifier.
  if (success) success = _roots->write(file);           // Write roots.
  file.close();                                         // Close the file.
  CmIO::reset();
  return success;
}


// "read" reads the file with the reserve name into the reserve.
//
bool CmReserve::read()
{
  if (!exists()) return false;              // Exit if no disk file.

  CmIOFile file((const char*) _name);       // Open the file.
  if (!file) return false;

  char *ident = file.read();                // Read the identifier.
  bool success = (strcmp(ident, (const char*) _identifier) == 0);
  delete[] ident;

  if (success)
  {
    removeAll();                            // Remove any existing roots.
    success = _roots->read(file);           // Read roots.
  }
  file.close();                             // Close the file.
  CmIO::reset();
  return success;
}


// "forEach" cycles through all objects in the reserve calling the
// specified function for each object passing in the object pointer
// and the application-supplied data.
//
void CmReserve::forEach(CmQueryFunc pFunc, void* pData)
{
  CmIterator *iter = _roots->newIterator();
  while (!iter->done())
  {
    CmAssociation *pAssoc = (CmAssociation*) iter->next();
    CmContainer   *pCont  = (CmContainer*) pAssoc->object();
    pCont->forEach(pFunc, pData);
  }
}


// "forEach" cycles through all objects in the reserve calling the 
// specified object member function for each object and passing in
// the application-supplied data.
//
void CmReserve::forEach(CmQueryMemFunc pFunc, void* pData)
{
  CmIterator *iter = _roots->newIterator();
  while (!iter->done())
  {
    CmAssociation *pAssoc = (CmAssociation*) iter->next();
    CmContainer   *pCont  = (CmContainer*) pAssoc->object();
    pCont->forEach(pFunc, pData);
  }
}


// "firstThat" returns the first object in the reserve that is
// passed into the specified function and returns true.
//
CmObject* CmReserve::firstThat(CmQueryFunc pFunc, void* pData) const
{
  CmObject   *outObj = NULL;
  CmIterator *iter   = _roots->newIterator();
  while (!iter->done() && !outObj)
  {
    CmAssociation *pAssoc = (CmAssociation*) iter->next();
    CmContainer   *pCont  = (CmContainer*) pAssoc->object();
    outObj = pCont->firstThat(pFunc, pData);
  }
  return outObj;
}


// "firstThat" returns the first object in the reserve that is
// passed into the specified function and returns true.
//
CmObject* CmReserve::firstThat(CmQueryMemFunc pFunc, void* pData) const
{
  CmObject   *outObj = NULL;
  CmIterator *iter   = _roots->newIterator();
  while (!iter->done() && !outObj)
  {
    CmAssociation *pAssoc = (CmAssociation*) iter->next();
    CmContainer   *pCont  = (CmContainer*) pAssoc->object();
    outObj = pCont->firstThat(pFunc, pData);
  }
  return outObj;
}


// "lastThat" returns the last object in the reserve that is
// passed into the specified function and returns true.
//
CmObject* CmReserve::lastThat(CmQueryFunc pFunc, void* pData) const
{
  CmObject   *outObj = NULL;
  CmIterator *iter   = _roots->newIterator();
  iter->last();
  while (!iter->done() && !outObj)
  {
    CmAssociation *pAssoc = (CmAssociation*) iter->previous();
    CmContainer   *pCont  = (CmContainer*) pAssoc->object();
    outObj = pCont->lastThat(pFunc, pData);
  }
  return outObj;
}


// "lastThat" returns the last object in the reserve that is
// passed into the specified function and returns true.
//
CmObject* CmReserve::lastThat(CmQueryMemFunc pFunc, void* pData) const
{
  CmObject   *outObj = NULL;
  CmIterator *iter   = _roots->newIterator();
  iter->last();
  while (!iter->done() && !outObj)
  {
    CmAssociation *pAssoc = (CmAssociation*) iter->previous();
    CmContainer   *pCont  = (CmContainer*) pAssoc->object();
    outObj = pCont->lastThat(pFunc, pData);
  }
  return outObj;
}


// "query" cycles through all objects in the reserve calling the
// specified function for each object passing in the object pointer
// and the application-supplied data.  A return value of true from
// the function means to add that object to the output object list.
//
CmContainer* CmReserve::query(CmQueryFunc pFunc, void* pData) const
{
  CmLinkedList *pList   = new CmLinkedList;
  CmIterator   *iter    = _roots->newIterator();
  bool          success = true;
  while (!iter->done() && success)
  {
    CmAssociation *pAssoc = (CmAssociation*) iter->next();
    CmContainer   *pCont  = (CmContainer*) pAssoc->object();
    CmContainer   *temp   = pCont->query(pFunc, pData);
    if (temp)
    {
      temp->ownsObjects(false);
      success = pList->addAllFrom(temp);
      delete temp;
    }
    else
      success = false;
  }
  return pList;
}


// "query" cycles through all objects in the reserve calling the 
// specified object member function for each object and passing in
// the application-supplied data.  A return value of true from the
// function means to add that object to the output object list.
//
CmContainer* CmReserve::query(CmQueryMemFunc pFunc, void* pData) const
{
  CmLinkedList *pList   = new CmLinkedList;
  CmIterator   *iter    = _roots->newIterator();
  bool          success = true;
  while (!iter->done() && success)
  {
    CmAssociation *pAssoc = (CmAssociation*) iter->next();
    CmContainer   *pCont  = (CmContainer*) pAssoc->object();
    CmContainer   *temp   = pCont->query(pFunc, pData);
    if (temp)
    {
      temp->ownsObjects(false);
      success = pList->addAllFrom(temp);
      delete temp;
    }
    else
      success = false;
  }
  return pList;
}


// "fileIdentifier" opens the specified file, reads the identifier
// string, and returns it.  The application must delete the string.
//
char* CmReserve::fileIdentifier(const char* name)
{
  CmIOFile file(name);
  if (!file) return NULL;

  char *file_str = file.read();
  file.close();
  return file_str;
}
