// cmstring.cpp
// -----------------------------------------------------------------
// Compendium - C++ Container Class Library
// Copyright (C) 1992-1995, Glenn M. Poorman, All rights reserved
// -----------------------------------------------------------------
// Character string implementation.
// -----------------------------------------------------------------

#include <cm/include/cmstring.h>
#include <stdlib.h>
#include <ctype.h>

#if defined(__TURBOC__) || defined(_MSC_VER)
#include <strstrea.h>
#else
#include <strstream.h>
#endif

// Define "stricmp" and "strnicmp" for machines that don't support it.
//
#if !defined(__TURBOC__)
  int stricmp(const char* a, const char* b)
  {
    CmString A(a), B(b);
    A.toUpper(); B.toUpper();
    return strcmp(A, B);
  }
  int strnicmp(const char* a, const char* b, int n)
  {
    CmString A(a), B(b);
    A.toUpper(); B.toUpper();
    return strncmp(A, B, n);
  }
#endif

// Macro call initializes the static type info data.
CMOBJECT_INIT(CmString, CmObject)

// Macro registers this class for binary file I/O.
CMOBJECT_REGISTER(CmString)

// Initialize the case sensitive compare flag.
//
bool CmString::_caseSensitive = true;


// "CmString" is the default string constructor.
//
CmString::CmString(const char *str, int n) : _text(NULL)
{
  assign(str, n);
}


// "CmString" constructs a string of size N and fills the string with
// N copies of the specified character.
//
CmString::CmString(int rep, char c) : _text(NULL)
{
  assign(c, rep);
}


// "CmString" is the string copy constructor.
//
CmString::CmString(const CmString &str, int n) : _text(NULL)
{
  assign(str, n);
}


// "CmString" constructs a string from the specified substring.
//
CmString::CmString(const CmSubString& sstr) : _text(NULL)
{
  assign(sstr._sp, sstr._sl);
}


// "~CmString" is the string destructor.
//
CmString::~CmString()
{
  resize(0);
}


// "assign" assigns n characters from the specified string to this
// string.  If n < 0, the string length is used.
//
CmString& CmString::assign(const CmString& str, int n)
{
  if (&str == this) return *this;
  if (n < 0) n = str.length();

  if (n != length())
    if (!resize(n))
      return *this;

  if (n)
  {
    memcpy(_text, str._text, n);
    _text[n] = '\0';
  }
  return *this;
}


// "assign" assigns n characters from the specified string to this
// string.  If n < 0, the string length is used.
//
CmString& CmString::assign(const char* str, int n)
{
  if (str == _text) return *this;
  if (n < 0) n = (str) ? strlen(str) : 0;

  if (n != length())
    if (!resize(n))
      return *this;

  if (n)
  {
    memcpy(_text, str, n);
    _text[n] = '\0';
  }
  return *this;
}


// "assign" assigns rep copies of the character c to this string.
//
CmString& CmString::assign(char c, int rep)
{
  if (rep != length())
    if (!resize(rep))
      return *this;

  if (rep)
  {
    memset(_text, c, rep);
    _text[rep] = '\0';
  }
  return *this;
}


// "append" appends n characters from the specified string to this
// string.  If n < 0, the string length is used.
//
CmString& CmString::append(const CmString& str, int n)
{
  if (&str == this) return *this;
  if (n < 0) n = str.length();

  int thislen = length();
  int newlen  = n + thislen;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  if (n)
  {
    memcpy(&(_text[thislen]), str._text, n);
    _text[newlen] = '\0';
  }
  return *this;
}


// "append" appends n characters from the specified string to this
// string.  If n < 0, the string length is used.
//
CmString& CmString::append(const char* str, int n)
{
  if (str == _text) return *this;
  if (n < 0) n = (str) ? strlen(str) : 0;

  int thislen = length();
  int newlen  = n + thislen;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  if (n)
  {
    memcpy(&(_text[thislen]), str, n);
    _text[newlen] = '\0';
  }
  return *this;
}


// "append" appends rep copies of the character c to this string.
//
CmString& CmString::append(char c, int rep)
{
  int thislen = length();
  int newlen  = thislen + rep;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  if (rep)
  {
    memset(&(_text[thislen]), c, rep);
    _text[newlen] = '\0';
  }
  return *this;
}


// "insert" inserts n characters from the specified string into this
// string at the specified index.
//
CmString& CmString::insert(int idx, const CmString& str, int n)
{
  int thislen = length();
  if (idx < 0 || idx > thislen) return *this;
  if (n < 0) n = str.length();

  int newlen = thislen + n;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  if (n)
  {
    memmove(&(_text[idx+n]), &(_text[idx]), thislen-idx);
    memcpy (&(_text[idx]), str._text, n);
    _text[newlen] = '\0';
  }
  return *this;
}


// "insert" inserts n characters from the specified string into this
// string at the specified index.
//
CmString& CmString::insert(int idx, const char* str, int n)
{
  int thislen = length();
  if (idx < 0 || idx > thislen) return *this;
  if (n < 0) n = (str) ? strlen(str) : 0;

  int newlen = thislen + n;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  if (n)
  {
    memmove(&(_text[idx+n]), &(_text[idx]), thislen-idx);
    memcpy (&(_text[idx]), str, n);  
    _text[newlen] = '\0';
  }
  return *this;
}


// "insert" inserts rep copies of the specified character into this
// string at the specified index.
//
CmString& CmString::insert(int idx, char c, int rep)
{
  int thislen = length();
  if (idx < 0 || idx > thislen) return *this;

  int newlen = thislen + rep;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  if (rep)
  {
    memmove(&(_text[idx+rep]), &(_text[idx]), thislen-idx);
    memset (&(_text[idx]), c, rep);
    _text[newlen] = '\0';
  }
  return *this;
}


// "replace" replaces n1 number of characters at the specified index with
// n2 characters from the specified string.
//
CmString& CmString::replace(int idx, int n1, const CmString& str, int n2)
{
  int thislen = length();
  if (idx < 0 || idx > thislen) return *this;
  if (n2 < 0) n2 = str.length();

  int newlen = thislen - n1 + n2;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  memmove(&(_text[idx+n2]), &(_text[idx+n1]), thislen-(idx+n1));
  memcpy (&(_text[idx]), str._text, n2);
  _text[newlen] = '\0';
  return *this;
}


// "replace" replaces n1 number of characters at the specified index with
// n2 characters from the specified string.
//
CmString& CmString::replace(int idx, int n1, const char* str, int n2)
{
  int thislen = length();
  if (idx < 0 || idx > thislen) return *this;
  if (n2 < 0) n2 = (str) ? strlen(str) : 0;

  int newlen = thislen - n1 + n2;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  memmove(&(_text[idx+n2]), &(_text[idx+n1]), thislen-(idx+n1));
  memcpy (&(_text[idx]), str, n2);
  _text[newlen] = '\0';
  return *this;
}


// "replace" replaces n1 number of characters at the specified index
// with rep copies of the specified character.
//
CmString& CmString::replace(int idx, int n1, char c, int rep)
{
  int thislen = length();
  if (idx < 0 || idx > thislen) return *this;

  int newlen = thislen - n1 + rep;

  if (newlen != thislen)
    if (!resize(newlen))
      return *this;

  memmove(&(_text[idx+rep]), &(_text[idx+n1]), thislen-(idx+n1));
  memset (&(_text[idx]), c, rep);
  _text[newlen] = '\0';
  return *this;
}


// "remove" removes the specified number of characters from the string
// starting at the specified index.
//
CmString& CmString::remove(int idx, int n)
{
  int thislen = length();
  if (n < 0) n = thislen - idx;
  if (idx < 0 || idx+n > thislen || n == 0) return *this;
  memmove(&(_text[idx]), &(_text[idx+n]), thislen-idx-n);
  resize(thislen-n);
  return *this;
}


// "index" returns the index of the first occurrence in this string
// of n characters of the specified string.
//
int CmString::index(const CmString& str, int n) const
{
  int pos = -1;
  if (n < 0) n = str.length();

  if (n)
  {
    int ii = 0, len = length();
    while (ii < len && (ii+n) <= len && pos == -1)
    {
      if (CmString::caseSensitive())
      {
        if (!strncmp(&(_text[ii++]), str._text, n)) pos = ii - 1;
      }
      else
      {
        if (!strnicmp(&(_text[ii++]), str._text, n)) pos = ii - 1;
      }
    }
  }
  return pos;
}


// "index" returns the index of the first occurrence in this string
// of n characters of the specified string.
//
int CmString::index(const char* str, int n) const
{
  int pos = -1;
  if (n < 0) n = (str) ? strlen(str) : 0;

  if (n)
  {
    int ii = 0, len = length();
    while (ii < len && (ii+n) <= len && pos == -1)
    {
      if (CmString::caseSensitive())
      {
        if (!strncmp(&(_text[ii++]), str, n)) pos = ii - 1;
      }
      else
      {
        if (!strnicmp(&(_text[ii++]), str, n)) pos = ii - 1;
      }
    }  
  }
  return pos;
}


// "index" returns the index of the first occurrence in this string
// of the specified character.
//
int CmString::index(char c) const
{
  int pos = -1, len = length(), ii = 0;
  while (ii < len && pos < 0)
  {
    if (CmString::caseSensitive())
    {
      if (c == _text[ii++]) pos = ii - 1;
    }
    else
    {
      if (toupper(c) == toupper(_text[ii++])) pos = ii - 1;
    }
  }
  return pos;
}


// "substr" returns a string containing n characters of this string
// starting at index idx.
//
CmString CmString::substr(int idx, int n) const
{
  int thislen = length();
  if (idx < 0 || idx >= thislen || idx+n >= thislen) return CmString();
  return CmString(&(_text[idx]), n);
}


// "subString" returns a substring containing n characters of this
// string starting at index idx.
//
CmSubString CmString::subString(int idx, int n)
{
  int ln = length();
  int st = (idx < 0) ? 0 : ((idx >= ln) ? ln - 1 : idx);
  int nc = (n < 0) ? 0 : ((n > (ln-st)) ? ln-st : n);
  return CmSubString(*this, st, nc);
}


// "subString" returns a substring containing n characters of this
// string starting at index idx.
//
const CmSubString CmString::subString(int idx, int n) const
{
  int ln = length();
  int st = (idx < 0) ? 0 : ((idx >= ln) ? ln - 1 : idx);
  int nc = (n < 0) ? 0 : ((n > (ln-st)) ? ln-st : n);
  return CmSubString(*this, st, nc);
}


// "compare" compares n characters of the specified string with this
// string returning 0 if they are equal, <0 if this string is less,
// and >0 if this string is greater.
//
int CmString::compare(const CmString& str, int n) const
{
  int cmp, thislen = length();
  if (n < 0) n = str.length();

  if (CmString::caseSensitive())
    cmp = strncmp(_text, str._text, (thislen < n) ? thislen : n);
  else
    cmp = strnicmp(_text, str._text, (thislen < n) ? thislen : n);

  return (cmp != 0) ? cmp : (thislen < n) ? -1 : (thislen == n) ? 0 : 1;
}


// "compare" compares n characters of the specified string with this
// string returning 0 if they are equal, <0 if this string is less,
// and >0 if this string is greater.
//
int CmString::compare(const char* str, int n) const
{
  int cmp, thislen = length();
  if (n < 0) n = (str) ? strlen(str) : 0;

  if (CmString::caseSensitive())
    cmp = strncmp(_text, str, (thislen < n) ? thislen : n);
  else
    cmp = strnicmp(_text, str, (thislen < n) ? thislen : n);

  return (cmp != 0) ? cmp : (thislen < n) ? -1 : (thislen == n) ? 0 : 1;
}


// "toLower" converts all characters in this string to lower case.
//
void CmString::toLower()
{
  int   i = length();
  char *q = _text;
  while (i--) { if (*q >= 'A' && *q <= 'Z') *q = *q - 'A' + 'a'; q++; }
}


// "toUpper" converts all characters in this string to upper case.
//
void CmString::toUpper()
{
  int   i = length();
  char *q = _text;
  while (i--) { if (*q >= 'a' && *q <= 'z') *q = *q - 'a' + 'A'; q++; }
}


// "leftJustify" left justifies the string within itself.
//
void CmString::leftJustify()
{
  int ii  = 0;
  int len = length();
  while (ii < len && isspace(_text[ii])) ii++;

  if (ii == len)
    assign((const char*) NULL);
  else if (ii > 0)
  {
    int nchars = len - ii;
    memmove(_text, &(_text[ii]), nchars);
    resize(nchars);
  }
}


// "rightJustify" right justifies the string within itself.
//
void CmString::rightJustify()
{
  int len = length();
  int ii  = len;
  while (ii >= 0 && isspace(_text[ii])) ii--;

  if (ii < 0)
    assign((const char*) NULL);
  else if (ii < len-1)
  {
    int nchars = len - ii;
    memmove(&(_text[nchars]), _text, ii);
    resize(ii);
  }
}


// "trim" trims the leading and trailing blank characters from the
// string.
//
void CmString::trim()
{
  if (!_text) return;
  char* p = _text;
  while (isspace(*p)) p++;
  if (p != _text) strcpy(_text, p);
  int cc = strlen(_text);
  if (cc) for (p = _text + cc - 1; isspace(*p); p--) *p = '\0';
  int len = strlen(_text);
  resize(len);
}


// "copy" copies n characters from this string starting at the
// specified index into the specified string and returns the number
// of characters copied.
//
int CmString::copy(char* str, int idx, int n) const
{
  int thislen = length();
  if (n < 0) n = thislen;

  if (idx < 0 || (idx+n) > thislen || !str) return 0;

  memcpy(str, &(_text[idx]), n);
  str[n] = '\0';
  return n;
}


// "isEqual" checks to see if the input object is a string and it they
// are equal.
//
bool CmString::isEqual(const CmObject* pObj) const
{
  CmString *str = CmPtrCast(CmString, pObj);
  return (str) ?
         ((compare(*str) == 0) ? true : false) : CmObject::isEqual(pObj);
}


// "compare" does a comparison of the input object and this string and
// returns -1 if this is < object, 0 if equal, 1 if this is greater.
//
int CmString::compare(const CmObject* pObj) const
{
  CmString *str = CmPtrCast(CmString, pObj);
  return (str) ? compare(*str) : CmObject::compare(pObj);
}


// "hash" hashes the character string.
//
unsigned CmString::hash(unsigned m) const
{
  unsigned h;
  char* v = _text;
  for (h = 0; *v != '\0'; v++) h = (64 * h + *v) % m;
  return h;
}


// "printOn" prints the string data to the specified output stream.
//
void CmString::printOn(ostream& strm) const
{
  strm << _text;
}


// "readFrom" reads text into this string from the specified input
// stream operator.
//
void CmString::readFrom(istream& strm)
{
  char       ch;
  ostrstream sos;
  while (strm.get(ch) && ch != '\n') sos << ch;
  sos << ends;
  assign((const char*) sos.str());
  delete[] sos.str();
}


// "write" writes the string length and text to the specified reserve
// binary file.
//
bool CmString::write(CmIOFile& file) const
{
  return file.write(_text);
}


// "read" reads the string length and text from the specified reserve
// binary file.
//
bool CmString::read(CmIOFile& file)
{
  char *str = file.read();
  if (!str) return false;
  assign(str);
  delete[] str;
  return true;
}


// "caseSensitive" sets the case sensitive compare flag.
//
void CmString::caseSensitive(bool f)
{
  CmString::_caseSensitive = f;
}


// "caseSensitive" returns the case sensitive compare flag.
//
bool CmString::caseSensitive()
{
  return CmString::_caseSensitive;
}


// "resize" resizes the current string capacity.
//
bool CmString::resize(unsigned newsize)
{
  if (!newsize)
  {
    free(_text);
    _text = NULL;
  }
  else
  {
    int   thislen = length();
    char *newstr  = (char*) realloc(_text, newsize + 1);
    if (!newstr) return false;
    _text = newstr;
    if (newsize < thislen) _text[newsize] = '\0';
  }
  return true;
}


// "assign" assigns n characters from the specified string into
// this substring.
//
CmString& CmSubString::assign(const CmString& str, int n)
{
  if (n < 0) n = str.length();
  if (_sl == n) strncpy(_sp, str._text, _sl);
  else          replace(str._text, n);
  return *_st;
}


// "assign" assigns n characters from the specified substring into
// this substring.
//
CmString& CmSubString::assign(const CmSubString& str, int n)
{
  if (n < 0) n = str._sl;
  if (_sl == n) strncpy(_sp, str._sp, n);
  else          replace(str._sp, n);
  return *_st;
}


// "assign" assigns n characters from the specified string into
// this substring.
//
CmString& CmSubString::assign(const char* str, int n)
{
  if (n < 0) n = (str) ? strlen(str) : 0;
  if (_sl == n) strncpy(_sp, str, _sl);
  else          replace(str, n);
  return *_st;
}


// "compare" compares this string with the input string returning
// 0 if the strings are equal, <0 if this is less, and >0 if this
// is greater.
//
int CmSubString::compare(const CmString& str) const
{
  if (CmString::caseSensitive())
    return strncmp(_sp, str._text, _sl);
  else
    return strnicmp(_sp, str._text, _sl);
}


// "compare" compares this string with the input substring returning
// 0 if the strings are equal, <0 if this is less, and >0 if this
// is greater.
//
int CmSubString::compare(const CmSubString& str) const
{
  if (CmString::caseSensitive())
    return strncmp(_sp, str._sp, _sl);
  else
    return strnicmp(_sp, str._sp, _sl);
}


// "compare" compares this string with the input string returning
// 0 if the strings are equal, <0 if this is less, and >0 if this
// is greater.
//
int CmSubString::compare(const char* str) const
{
  if (CmString::caseSensitive())
    return strncmp(_sp, str, _sl);
  else
    return strnicmp(_sp, str, _sl);
}


// "<<" outputs this substring to the specified stream.
//
ostream& operator<<(ostream& strm, const CmSubString& str)
{
  char *p = str._sp;
  for (int ii = 0; ii < str._sl; ii++)
    strm << *(p++);
  return strm;
}


// "CmSubString" constructs a substring from the specified string,
// starting position, and ending position.
//
CmSubString::CmSubString(const CmString& S, int start, int nchars)
{
  _sp = &(S._text[start]);
  _sl = nchars;
  _st = &((CmString&)S);
}


// "CmSubString" constructs a substring from the specified substring.
//
CmSubString::CmSubString(const CmSubString& S)
{
  _sp = S._sp;
  _sl = S._sl;
  _st = S._st;
}


// "replace" replaces the contents of this substring with the specified
// string.
//
void CmSubString::replace(const char* src, int srclen)
{
  int tot = (srclen == _sl) ? srclen : ((srclen < _sl) ? srclen : _sl);
  strncpy(_sp, src, tot);
}


// "+" concatenates the specified strings returning the result.
//
CmString operator+(const CmString& s1, const CmString& s2)
{
  return CmString(s1).append(s2);
}

// "+" concatenates the specified strings returning the result.
//
CmString operator+(const CmString& s1, const char* s2)
{
  return CmString(s1).append(s2);
}


// "+" concatenates the specified strings returning the result.
//
CmString operator+(const char* s1, const CmString& s2)
{
  return CmString(s1).append(s2);
}


// "+" concatenates specified string and character returning the result.
//
CmString operator+(const CmString& s, char c)
{
  return CmString(s).append(c);
}


// "+" concatenates specified character and string returning the result.
//
CmString operator+(char c, const CmString& s)
{
  return CmString(1,c).append(s);
}
